/*----------------------------------------------------------------------------
 *  config.c - functions for processing config/<package>.txt files
 *
 *  Copyright (c) 2000-2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Creation:    2000-05-03  fm
 *  Last Update: $Id: config.c 39625 2015-06-10 16:21:12Z kristov $
 *----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <dirent.h>
#include <libmkfli4l/defs.h>
#include <libmkfli4l/var.h>
#include <libmkfli4l/regex.h>
#include <libmkfli4l/check.h>
#include <libmkfli4l/log.h>
#include <libmkfli4l/parse.h>
#include <libmkfli4l/file.h>
#include <libmkfli4l/options.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/file.h>
#include <ziplist/opt.h>
#include "config.h"

#define SUPERSEDES "_fli4l"

static int
read_supersede_config(char const *name)
{
    char *file;
    if (get_file_name(config_dir, name, def_cfg_ext, &file)) {
        log_info(LOG_INFO, "reading supersede configuration file %s\n", file);
        if (read_config_file(file, name, TRUE) != OK) {
            log_error("Error while reading config file %s\n", file);
            free(file);
            return ERR;
        }
    }
    else {
        log_info(LOG_INFO, "ignoring non-existing supersede config file %s\n", file);
    }

    free(file);
    return OK;
}

static int
read_configuration_files(
    char const *config_dir, char const *check_dir, char const *pkg,
    BOOL mandatory
)
{
    char *file;

    if (n_packages == MAX_PACKAGES) {
        log_error("Error: Number of packages exceeded maximum of %d while processing %s!\n", MAX_PACKAGES, pkg);
        return ERR;
    }

    if (strcasecmp(pkg, "config.txt") == 0) {
        log_error("Error: %s/%s is an old style configuration file."
                  " READ DOCUMENTATION!\n", config_dir, pkg);
        return ERR;
    }

    log_info (LOG_INFO, "package %s\n", pkg);
    inc_log_indent_level ();

    if (get_file_name(config_dir, pkg, def_cfg_ext, &file)) {
        log_info(LOG_VERBOSE, "reading %s\n", file);
        if (read_config_file(file, pkg, FALSE) != OK) {
            log_error("Error while reading configuration file %s\n", file);
            free(file);
            return ERR;
        }
        else {
            free(file);
        }

        packages[n_packages] = strsave(pkg);

        if (get_file_name(check_dir, pkg, def_regex_ext, &file)) {
            log_info (LOG_VERBOSE, "reading %s\n", file);
            if (regexp_read_file(file) != OK) {
                log_error("An error found while reading regular "
                          "expressions for package %s, aborting...\n", pkg);
                free(file);
                return ERR;
            }
            else {
                free(file);
            }
        }
        else {
            log_info (LOG_VERBOSE, "no expression file for package %s\n", pkg);
            free(file);
        }

        if (get_file_name(check_dir, pkg, def_check_ext, &file)) {
            log_info (LOG_VERBOSE, "reading %s\n", file);
            if (read_check_file(file, packages[n_packages]) != OK) {
                log_error("An error found while executing checks "
                          "for package %s, aborting...\n", pkg);
                free(file);
                return ERR;
            }
            else {
                free(file);
            }
        }
        else {
            log_error("Error while accessing %s: %s\n", file, strerror(errno));
            free(file);
            return ERR;
        }

        ++n_packages;
    }
    else {
        if (mandatory) {
            log_error("Error while accessing %s: %s\n", file, strerror(errno));
            free(file);
            return ERR;
        }
        else {
            free(file);
        }
    }

    dec_log_indent_level ();
    return OK;
}

struct plist_t {
    char const *packet;
    BOOL mandatory;
};

static struct plist_t plist[] = {
    { SUPERSEDES, FALSE, },
    { "base", TRUE, },
    { "dns_dhcp", FALSE, },
    { NULL, FALSE, },
};

static BOOL
filter_config_file(char const *name)
{
    struct plist_t *p;
    for (p = plist; p->packet; ++p) {
        if (strcmp(name, p->packet) == 0) {
            return TRUE;
        }
    }

    return FALSE;
}

static int
read_preordered_configs(char const *config_dir, char const *check_dir)
{
    struct plist_t const *p;
    for (p = plist + 1; p->packet; ++p) {
        if (read_configuration_files(config_dir, check_dir,
                                     p->packet, p->mandatory) != OK) {
            return ERR;
        }
    }
    return OK;
}

int
read_mkfli4l_config(char const *config_dir, char const *check_dir)
{
    DIR *dirp;
    struct dirent *dp;
    int len;

    var_add_weak_declaration("base", "CONFIG_DIR", config_dir_with_slashes,
                             "internal variable", TYPE_UNKNOWN,
                             __FILE__, __LINE__, 0);

    set_variable("base", "BASE", "yes", "predefined variable", FALSE, NULL, FALSE, 0);
    mark_var_tagged("BASE");
    mark_var_checked("BASE");

    if (read_supersede_config(SUPERSEDES) != OK) {
        return ERR;
    }

    log_info(LOG_INFO, "# parse pre-ordered config files\n");
    if (read_preordered_configs(config_dir, check_dir) != OK) {
        return ERR;
    }
    log_info(LOG_INFO, "# parse other config files\n");

    dirp = opendir(config_dir);
    if (! dirp) {
        log_error("Error opening config dir '%s': %s\n",
                  config_dir, strerror(errno));
        return ERR;
    }

    while ((dp = readdir(dirp)) != NULL) {
        len = strlen(dp->d_name);

        if (len > 4 && strcasecmp(dp->d_name + len - 4, ".txt") == 0) {
            dp->d_name[len - 4] = '\0';
            if (!filter_config_file(dp->d_name)) {
                if (read_configuration_files(config_dir, check_dir,
                                             dp->d_name, TRUE) != OK) {
                    return ERR;
                }
            }
        }
    }

    (void) closedir(dirp);
    return OK;
}

int
read_config(char const *config_dir, char const *check_dir, char **packages)
{
    BOOL read_base_exp = TRUE;
    char **p;

    for (p = packages; *p; p++) {
        if (strcmp("base", *p) == 0) {
            read_base_exp = FALSE;
        }
    }

    if (read_base_exp) {
        char *file;
        if (get_file_name(check_dir, "base", def_regex_ext, &file)) {
            log_info(LOG_VERBOSE, "reading %s\n", file);
            if (regexp_read_file(file) != OK) {
                log_error ("An error found while reading basic regular "
                           "expressions from %s/base.exp\n", check_dir);
                free(file);
                return ERR;
            }
            else {
                free(file);
            }
        }
        else {
            log_error("Missing basic regular expressions (%s/base.exp)\n",
                      check_dir);
            free(file);
            return ERR;
        }
    }

    for (p = packages; *p; p++) {
        if (read_configuration_files(config_dir, check_dir, *p, TRUE) != OK) {
            return ERR;
        }
    }

    return OK;
}
