/*****************************************************************************
 *  @file expression/expr.c
 *  Functions for processing expressions.
 *
 *  Copyright (c) 2012 The fli4l team
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: expr.c 22634 2012-07-06 13:44:32Z kristov $
 *****************************************************************************
 */

#include <expression/expr_internal.h>
#include <expression/expr_literal.h>
#include <libmkfli4l/defs.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>

#define MODULE "expression::expr"

int
expr_get_type(struct expr_t const *expr)
{
    return expr->type;
}

char const *
expr_type_to_string(int type)
{
    switch (type) {
    case EXPR_TYPE_ERROR :
        return "error";
    case EXPR_TYPE_STRING :
        return "string";
    case EXPR_TYPE_INTEGER :
        return "integer";
    case EXPR_TYPE_BOOLEAN :
        return "boolean";
    default :
        fatal_exit(MODULE ": BUG in expr_type_to_string(): Unknown expression type %d.\n", type);
    }
}

int
expr_get_kind(struct expr_t const *expr)
{
    return expr->kind;
}

char const *
expr_kind_to_string(int kind)
{
    switch (kind) {
    case EXPR_KIND_LITERAL :
        return "literal";
    case EXPR_KIND_LOGICAL_NOT :
        return "logical not";
    case EXPR_KIND_LOGICAL_AND :
        return "logical and";
    case EXPR_KIND_LOGICAL_OR :
        return "logical or";
    case EXPR_KIND_VARIABLE :
        return "variable";
    case EXPR_KIND_EQUAL :
        return "equal";
    case EXPR_KIND_MATCH :
        return "match";
    default :
        fatal_exit(MODULE ": BUG in expr_kind_to_string(): Unknown expression kind %d.\n", kind);
    }
}

struct var_scope_t *
expr_get_scope(struct expr_t const *expr)
{
    return expr->scope;
}

void
expr_common_init(
    struct expr_t *expr,
    int type,
    int kind,
    struct var_scope_t *scope,
    expr_evaluate_t evaluate,
    expr_to_string_t toString,
    expr_clone_t clone,
    expr_equal_t equal,
    expr_free_t free
)
{
    expr->type = type;
    expr->kind = kind;
    expr->scope = scope;
    expr->evaluate = evaluate;
    expr->toString = toString;
    expr->clone = clone;
    expr->equal = equal;
    expr->free = free;
}

void
expr_common_free(struct expr_t *expr)
{
    free(expr);
}

struct expr_t *
expr_evaluate(struct expr_t const *expr)
{
    struct expr_t *const result = (*expr->evaluate)(expr);
    if (!result) {
        fatal_exit(
            MODULE ": BUG in expr_evaluate(): Evaluation of expression '%s' yields NULL.\n",
            expr_to_string(expr)
        );
    }
    if (!expr_is_assignable_to(result->type, expr->type)) {
        fatal_exit(
            MODULE ": BUG in expr_evaluate(): Dynamic type '%s' of expression '%s' is not assignable to its static type '%s'.\n",
            expr_type_to_string(result->type),
            expr_to_string(expr),
            expr_type_to_string(expr->type)
        );
    }
    if (result->kind != EXPR_KIND_LITERAL) {
        fatal_exit(
            MODULE ": BUG in expr_evaluate(): Evaluation of expression '%s' yields not a literal, but an expression of kind '%s'.\n",
            expr_to_string(expr),
            expr_kind_to_string(result->kind)
        );
    }
    return result;
}

char *
expr_to_string(struct expr_t const *expr)
{
    return (*expr->toString)(expr);
}

BOOL
expr_is_assignable_to(int sourceType, int targetType)
{
    return sourceType == targetType ||
        sourceType == EXPR_TYPE_ERROR ||
        targetType == EXPR_TYPE_STRING;
}

struct expr_t *
expr_cast(struct expr_t const *expr, int targetType)
{
    if (expr_is_assignable_to(expr->type, targetType)) {
        BOOL is_literal = expr->kind == EXPR_KIND_LITERAL;
        struct expr_t const *e = is_literal ? expr : expr_evaluate(expr);
        struct expr_t *result;

        if (e->type == targetType || e->type == EXPR_TYPE_ERROR) {
            result = expr_clone(e);
        }
        else if (targetType == EXPR_TYPE_STRING) {
            char *value = expr_to_string(e);
            result = expr_literal_create_string(expr->scope, value);
            free(value);
        }
        else {
            fatal_exit(
                MODULE ": BUG in expr_cast(): Cannot cast expression from '%s' to '%s'.\n",
                expr_type_to_string(e->type),
                expr_type_to_string(targetType)
            );
        }

        if (!is_literal) {
            expr_destroy((struct expr_t *) e);
        }

        return result;
    }
    else {
        return NULL;
    }
}

/**
 * Returns the value of an expression. If the expression is not evaluated yet,
 * it will first be evaluated. If the type of the evaluated expression is not
 * identical to the target type, a type cast is performed if possible (i.e. if
 * expr_is_assignable_to(type_of_evaluated_expression, targetType) returns
 * TRUE), otherwise the function fails.
 *
 * @param expr
 *  The expression.
 * @param targetType
 *  The target type.
 * @param result
 *  Points to an object receiving the value.
 * @param error
 *  Receives the error message if evaluation fails due to an error value. The
 *  caller is responsible for free()ing this string after use.
 * @return
 *  TRUE if the evaluation succeeds and the type of the value is assignable to
 *  the target type, FALSE otherwise. In the latter case, "error" will point to
 *  an error message after returning to the caller.
 */
static BOOL
expr_get_value(struct expr_t const *expr, int targetType, void *result, char **error)
{
    struct expr_t *e = expr_cast(expr, targetType);
    if (e && e->type == targetType) {
        expr_literal_get_value(e, result);
        expr_destroy(e);
        return TRUE;
    }
    else if (e) {
        expr_literal_get_value(e, error);
        expr_destroy(e);
        return FALSE;
    }
    else {
        *error = safe_sprintf(
            "invalid type cast from %s to %s",
            expr_type_to_string(expr->type),
            expr_type_to_string(targetType)
        );
        return FALSE;
    }
}

BOOL
expr_get_string_value(struct expr_t const *expr, char **result, char **error)
{
    return expr_get_value(expr, EXPR_TYPE_STRING, result, error);
}

BOOL
expr_get_integer_value(struct expr_t const *expr, int *result, char **error)
{
    return expr_get_value(expr, EXPR_TYPE_INTEGER, result, error);
}

BOOL
expr_get_boolean_value(struct expr_t const *expr, BOOL *result, char **error)
{
    return expr_get_value(expr, EXPR_TYPE_BOOLEAN, result, error);
}

struct expr_t *
expr_clone(struct expr_t const *expr)
{
    return (*expr->clone)(expr);
}

BOOL
expr_equal(struct expr_t const *expr1, struct expr_t const *expr2)
{
    if (expr1->kind != expr2->kind || expr1->type != expr2->type) {
        return FALSE;
    }
    else {
        return (*expr1->equal)(expr1, expr2);
    }
}

void
expr_destroy(struct expr_t *expr)
{
    (*expr->free)(expr);
}
