/*****************************************************************************
 *  @file expression/expr_equal.c
 *  Functions for processing equality comparison expressions.
 *
 *  Copyright (c) 2012 The fli4l team
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: expr_equal.c 22634 2012-07-06 13:44:32Z kristov $
 *****************************************************************************
 */

#include <expression/expr_internal.h>
#include <expression/expr_equal.h>
#include <expression/expr_literal.h>
#include <libmkfli4l/defs.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>
#include <string.h>

#define MODULE "expression::expr_equal"

/**
 * Represents a equality comparison expression.
 */
struct expr_equal_t {
    /**
     * Common part belonging to every expression. This member must be the first
     * one to allow casting between expr_t and expr_variable_t where necessary.
     */
    struct expr_t expr;
    /**
     * The first argument.
     */
    struct expr_t *arg1;
    /**
     * The second argument.
     */
    struct expr_t *arg2;
};

/**
 * Evaluates a equality comparison expression.
 *
 * @param expr
 *  The expression to evaluate. The caller must ensure that it is a equality
 *  comparison expression.
 * @return
 *  A properly typed literal expression representing the result of comparing
 *  the evaluated arguments.
 */
static struct expr_t *
expr_equal_evaluate(struct expr_t const *expr)
{
    char *arg_err = NULL;
    struct expr_t *result = NULL;

    struct expr_equal_t const *const eqexp
        = (struct expr_equal_t const *) expr;

    struct expr_t *arg1_eval = expr_evaluate(eqexp->arg1);
    struct expr_t *arg2_eval = expr_evaluate(eqexp->arg2);

    if (arg1_eval->type != arg2_eval->type) {
        if (expr_is_assignable_to(arg1_eval->type, arg2_eval->type)) {
            struct expr_t *arg = expr_cast(arg1_eval, arg2_eval->type);
            expr_destroy(arg1_eval);
            arg1_eval = arg;
        }
        else if (expr_is_assignable_to(arg2_eval->type, arg1_eval->type)) {
            struct expr_t *arg = expr_cast(arg2_eval, arg1_eval->type);
            expr_destroy(arg2_eval);
            arg2_eval = arg;
        }
        else {
            result = expr_literal_create_boolean(expr->scope, FALSE);
        }
    }

    if (!result) {
        switch (arg1_eval->type) {
        case EXPR_TYPE_STRING :
            {
                char *val1 = NULL;
                char *val2 = NULL;
                if (!expr_get_string_value(arg1_eval, &val1, &arg_err)) {
                    fatal_exit(
                        MODULE ": BUG in expr_equal_evaluate(): Cannot retrieve string value of first expression '%s'.\n",
                        expr_to_string(arg1_eval)
                    );
                }
                else if (!expr_get_string_value(arg2_eval, &val2, &arg_err)) {
                    result = expr_literal_create_error(expr->scope, arg_err);
                    free(arg_err);
                }
                else {
                    result = expr_literal_create_boolean(expr->scope, strcmp(val1, val2) == 0);
                }
                free(val1);
                free(val2);
                break;
            }
        case EXPR_TYPE_INTEGER :
            {
                int val1;
                int val2;
                if (!expr_get_integer_value(arg1_eval, &val1, &arg_err)) {
                    fatal_exit(
                        MODULE ": BUG in expr_equal_evaluate(): Cannot retrieve integer value of first expression '%s'.\n",
                        expr_to_string(arg1_eval)
                    );
                }
                else if (!expr_get_integer_value(arg2_eval, &val2, &arg_err)) {
                    result = expr_literal_create_error(expr->scope, arg_err);
                    free(arg_err);
                }
                else {
                    result = expr_literal_create_boolean(expr->scope, val1 == val2);
                }
                break;
            }
        case EXPR_TYPE_BOOLEAN :
            {
                BOOL val1;
                BOOL val2;
                if (!expr_get_boolean_value(arg1_eval, &val1, &arg_err)) {
                    fatal_exit(
                        MODULE ": BUG in expr_equal_evaluate(): Cannot retrieve boolean value of first expression '%s'.\n",
                        expr_to_string(arg1_eval)
                    );
                }
                else if (!expr_get_boolean_value(arg2_eval, &val2, &arg_err)) {
                    result = expr_literal_create_error(expr->scope, arg_err);
                    free(arg_err);
                }
                else {
                    result = expr_literal_create_boolean(expr->scope, val1 == val2);
                }
                break;
            }
        case EXPR_TYPE_ERROR :
            result = expr_clone(arg1_eval);
            break;
        }
    }

    expr_destroy(arg1_eval);
    expr_destroy(arg2_eval);
    return result;
}

/**
 * Prints a equality comparison expression.
 *
 * @param expr
 *  The expression to print. The caller must ensure that it is a equality
 *  comparison expression.
 * @return
 *  A dynamically allocated string containing a string representation of the
 *  expression.
 */
static char *
expr_equal_toString(struct expr_t const *expr)
{
    struct expr_equal_t const *const eqexp
        = (struct expr_equal_t const *) expr;
    char *arg1_str = expr_to_string(eqexp->arg1);
    char *arg2_str = expr_to_string(eqexp->arg2);
    char *result = safe_sprintf("(%s) == (%s)", arg1_str, arg2_str);
    free(arg1_str);
    free(arg2_str);
    return result;
}

/**
 * Clones a equality comparison expression.
 *
 * @param expr
 *  The expression to clone. The caller must ensure that it is a equality
 *  comparison expression.
 * @return
 *  The cloned expression.
 */
static struct expr_t *
expr_equal_clone(struct expr_t const *expr)
{
    struct expr_equal_t const *const eqexp
        = (struct expr_equal_t const *) expr;
    return expr_equal_create(
        expr->scope,
        expr_clone(eqexp->arg1),
        expr_clone(eqexp->arg2)
    );
}

/**
 * Compares two equality comparison expressions.
 *
 * @param expr1
 *  The first expression. The caller must ensure that it is a equality
 *  comparison expression.
 * @param expr2
 *  The second expression. The caller must ensure that it is a equality
 *  comparison expression.
 * @return
 *  TRUE if both expressions are equal, FALSE otherwise.
 */
static BOOL
expr_equal_equal(struct expr_t const *expr1, struct expr_t const *expr2)
{
    struct expr_equal_t const *eqexp1
        = (struct expr_equal_t const *) expr1;
    struct expr_equal_t const *eqexp2
        = (struct expr_equal_t const *) expr2;
    return expr_equal(eqexp1->arg1, eqexp2->arg1)
        && expr_equal(eqexp1->arg2, eqexp2->arg2);
}

/**
 * Frees a equality comparison expression.
 *
 * @param expr
 *  The expression to free. The caller must ensure that it is a equality
 *  comparison expression.
 */
static void
expr_equal_free(struct expr_t *expr)
{
    struct expr_equal_t const *const eqexp
        = (struct expr_equal_t const *) expr;
    expr_destroy(eqexp->arg1);
    expr_destroy(eqexp->arg2);
    expr_common_free(expr);
}

struct expr_t *
expr_equal_create(
    struct var_scope_t *scope,
    struct expr_t *arg1,
    struct expr_t *arg2
)
{
    struct expr_equal_t *const eqexp
        = (struct expr_equal_t *) safe_malloc(sizeof(struct expr_equal_t));

    expr_common_init(
        &eqexp->expr,
        EXPR_TYPE_BOOLEAN,
        EXPR_KIND_EQUAL,
        scope,
        &expr_equal_evaluate,
        &expr_equal_toString,
        &expr_equal_clone,
        &expr_equal_equal,
        &expr_equal_free
    );

    eqexp->arg1 = arg1;
    eqexp->arg2 = arg2;

    return &eqexp->expr;
}
