/*****************************************************************************
 *  @file expression/expr_variable.c
 *  Functions for processing variable expressions.
 *
 *  Copyright (c) 2012 The fli4l team
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: expr_variable.c 24192 2012-10-29 07:31:29Z kristov $
 *****************************************************************************
 */

#include <expression/expr_internal.h>
#include <expression/expr_variable.h>
#include <expression/expr_literal.h>
#include <var.h>
#include <libmkfli4l/defs.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>
#include <string.h>

#define MODULE "expression::expr_variable"

/**
 * Represents a variable expression.
 */
struct expr_variable_t {
    /**
     * Common part belonging to every expression. This member must be the first
     * one to allow casting between expr_t and expr_variable_t where necessary.
     */
    struct expr_t expr;
    /**
     * The variable instance referenced by this expression.
     */
    struct var_instance_t *instance;
};

/**
 * Creates an error expression due to failed evaluation of a variable
 * expression.
 *
 * @param scope
 *  The scope of the resulting expression.
 * @param instance
 *  The variable instance the evaluation of which has failed.
 * @param msg_tmpl
 *  The message template containing a %s placeholder for the variable name.
 * @return
 *  The error expression.
 */
static struct expr_t *
expr_variable_evaluate_error(
    struct var_scope_t *scope,
    struct var_instance_t *instance,
    char const *msg_tmpl)
{
    char *name = var_instance_to_string(instance);
    char *msg = safe_sprintf(msg_tmpl, name);
    struct expr_t *result = expr_literal_create_error(scope, msg);
    free(msg);
    free(name);
    return result;
}

/**
 * Evaluates a variable expression.
 *
 * @param expr
 *  The expression to evaluate. The caller must ensure that it is a variable
 *  expression.
 * @return
 *  A properly typed literal expression representing the value of the
 *  referenced variable.
 */
static struct expr_t *
expr_variable_evaluate(struct expr_t const *expr)
{
    struct expr_variable_t const *const varexp
        = (struct expr_variable_t const *) expr;
    struct var_value_t *value = var_instance_try_get_value(varexp->instance);

    if (value) {
        struct expr_t *result;
        int int_value;
        BOOL bool_value;

        int const expr_type = vartype_get_expression_type(var_get_type(varexp->instance->var));
        switch (expr_type) {
        case EXPR_TYPE_STRING :
            result = expr_literal_create_string(expr->scope, value->value);
            break;
        case EXPR_TYPE_INTEGER :
            if (var_value_convert_to_integer(value, &int_value)) {
                result = expr_literal_create_integer(expr->scope, int_value);
            }
            else {
                char *msg_tmpl = safe_sprintf(
                    "converting value '%s' of variable '%%s' to type integer failed",
                    value->value
                );
                result = expr_variable_evaluate_error(expr->scope, varexp->instance, msg_tmpl);
                free(msg_tmpl);
            }
            break;
        case EXPR_TYPE_BOOLEAN :
            if (var_value_convert_to_boolean(value, &bool_value)) {
                result = expr_literal_create_boolean(expr->scope, bool_value);
            }
            else {
                char *msg_tmpl = safe_sprintf(
                    "converting value '%s' of variable '%%s' to type boolean failed",
                    value->value
                );
                result = expr_variable_evaluate_error(expr->scope, varexp->instance, msg_tmpl);
                free(msg_tmpl);
            }
            break;
        default :
            fatal_exit(
                MODULE ": BUG in expr_variable_evaluate(): Unsupported variable type '%s' of variable '%s'.\n",
                expr_type_to_string(expr_type),
                var_get_name(varexp->instance->var)
            );
        }

        var_value_destroy(value);
        return result;
    }
    else {
        return expr_variable_evaluate_error(expr->scope, varexp->instance, "reading variable '%s' failed");
    }
}

/**
 * Prints a variable expression.
 *
 * @param expr
 *  The expression to print. The caller must ensure that it is a variable
 *  expression.
 * @return
 *  A dynamically allocated string containing the variable name.
 */
static char *
expr_variable_toString(struct expr_t const *expr)
{
    struct expr_variable_t const *const varexp
        = (struct expr_variable_t const *) expr;
    return var_instance_to_string(varexp->instance);
}

/**
 * Clones a variable expression.
 *
 * @param expr
 *  The expression to clone. The caller must ensure that it is a variable
 *  expression.
 * @return
 *  The cloned expression.
 */
static struct expr_t *
expr_variable_clone(struct expr_t const *expr)
{
    struct expr_variable_t const *varexp = (struct expr_variable_t const *) expr;
    return expr_variable_create(
        expr->scope,
        var_instance_create(
            varexp->instance->var,
            varexp->instance->num_indices,
            varexp->instance->indices
        )
    );
}

/**
 * Compares two variable expressions.
 *
 * @param expr1
 *  The first expression. The caller must ensure that it is a variable
 *  expression.
 * @param expr2
 *  The second expression. The caller must ensure that it is a variable
 *  expression.
 * @return
 *  TRUE if both expressions are equal, FALSE otherwise.
 */
static BOOL
expr_variable_equal(struct expr_t const *expr1, struct expr_t const *expr2)
{
    struct expr_variable_t const *varexp1 = (struct expr_variable_t const *) expr1;
    struct expr_variable_t const *varexp2 = (struct expr_variable_t const *) expr2;

    return varexp1->instance->var == varexp2->instance->var &&
        varexp1->instance->num_indices == varexp2->instance->num_indices &&
        memcmp(
            varexp1->instance->indices,
            varexp2->instance->indices,
            varexp1->instance->num_indices * sizeof(unsigned)
        ) == 0;
}

/**
 * Frees a variable expression.
 *
 * @param expr
 *  The expression to free. The caller must ensure that it is a variable
 *  expression.
 */
static void
expr_variable_free(struct expr_t *expr)
{
    struct expr_variable_t const *const varexp
        = (struct expr_variable_t const *) expr;
    var_instance_destroy(varexp->instance);
    expr_common_free(expr);
}

struct expr_t *
expr_variable_create(
    struct var_scope_t *scope,
    struct var_instance_t *instance
)
{
    struct expr_variable_t *const varexp
        = (struct expr_variable_t *) safe_malloc(sizeof(struct expr_variable_t));

    expr_common_init(
        &varexp->expr,
        vartype_get_expression_type(var_get_type(instance->var)),
        EXPR_KIND_VARIABLE,
        scope,
        &expr_variable_evaluate,
        &expr_variable_toString,
        &expr_variable_clone,
        &expr_variable_equal,
        &expr_variable_free
    );

    varexp->instance = instance;
    return &varexp->expr;
}
