/*****************************************************************************
 *  @file test_expr_variable.c
 *  Functions for testing variable expressions.
 *
 *  Copyright (c) 2012 The fli4l team
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: test_expr_variable.c 24254 2012-11-04 17:47:48Z kristov $
 *****************************************************************************
 */

#include <testing/tests.h>
#include <expression/expr.h>
#include <expression/expr_variable.h>
#include <expression/expr_literal.h>
#include <vartype.h>
#include <var.h>
#include <libmkfli4l/str.h>
#include <stdlib.h>
#include <string.h>

#define MU_TESTSUITE "expression.expr_variable"

/**
 * Tests expr_kind_to_string().
 */
mu_test_begin(test_expr_variable_kindtostring)
    mu_assert_eq_str("variable", expr_kind_to_string(EXPR_KIND_VARIABLE));
mu_test_end()

/**
 * Tests expr_get_type() on a string variable.
 */
mu_test_begin(test_expr_variable_gettype_string)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());

    inst = var_instance_create(var, 0, NULL);
    expr = expr_variable_create(var_get_global_scope(), inst);
    mu_assert_eq_int(EXPR_TYPE_STRING, expr_get_type(expr));
    expr_destroy(expr);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests expr_get_type() on an integer variable.
 */
mu_test_begin(test_expr_variable_gettype_integer)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());

    inst = var_instance_create(var, 0, NULL);
    expr = expr_variable_create(var_get_global_scope(), inst);
    mu_assert_eq_int(EXPR_TYPE_INTEGER, expr_get_type(expr));
    expr_destroy(expr);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests expr_get_type() on a boolean variable.
 */
mu_test_begin(test_expr_variable_gettype_boolean)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_BOOLEAN);
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());

    inst = var_instance_create(var, 0, NULL);
    expr = expr_variable_create(var_get_global_scope(), inst);
    mu_assert_eq_int(EXPR_TYPE_BOOLEAN, expr_get_type(expr));
    expr_destroy(expr);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests expr_get_kind().
 */
mu_test_begin(test_expr_variable_getkind)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());

    inst = var_instance_create(var, 0, NULL);
    expr = expr_variable_create(var_get_global_scope(), inst);
    mu_assert_eq_int(EXPR_KIND_VARIABLE, expr_get_kind(expr));
    expr_destroy(expr);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests evaluation of a variable expression referencing a non-array variable
 * of type string.
 */
mu_test_begin(test_expr_variable_eval_non_array_string)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr;
    char *value = NULL;
    char *error_msg;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    inst = var_instance_create(var, 0, NULL);
    var_instance_set_and_check_value(inst, "2", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION);

    expr = expr_variable_create(var_get_global_scope(), inst);
    mu_assert(expr_get_string_value(expr, &value, &error_msg));
    mu_assert_eq_str("2", value);
    free(value);
    expr_destroy(expr);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests evaluation of a variable expression referencing a non-array variable
 * of type integer.
 */
mu_test_begin(test_expr_variable_eval_non_array_integer)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr;
    int value = 0;
    char *error_msg;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    inst = var_instance_create(var, 0, NULL);
    var_instance_set_and_check_value(inst, "2", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION);

    expr = expr_variable_create(var_get_global_scope(), inst);
    mu_assert(expr_get_integer_value(expr, &value, &error_msg));
    mu_assert_eq_int(2, value);
    expr_destroy(expr);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests evaluation of a variable expression referencing a non-array variable
 * of type integer with invalid contents.
 */
mu_test_begin(test_expr_variable_eval_non_array_integer_invalid)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr;
    int value = 0;
    char *error_msg = NULL;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("NUMEXT", "0|[1-9][0-9]*|NaN", "number expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 1));
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    inst = var_instance_create(var, 0, NULL);
    var_instance_set_and_check_value(inst, "NaN", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION);

    expr = expr_variable_create(var_get_global_scope(), inst);
    mu_assert(!expr_get_integer_value(expr, &value, &error_msg));
    mu_assert_eq_str("converting value 'NaN' of variable 'V' to type integer failed", error_msg);
    free(error_msg);
    expr_destroy(expr);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests evaluation of a variable expression referencing a non-array variable
 * of type boolean.
 */
mu_test_begin(test_expr_variable_eval_non_array_boolean)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr;
    BOOL value = FALSE;
    char *error_msg;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_BOOLEAN);
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    inst = var_instance_create(var, 0, NULL);
    var_instance_set_and_check_value(inst, "yes", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION);

    expr = expr_variable_create(var_get_global_scope(), inst);
    mu_assert(expr_get_boolean_value(expr, &value, &error_msg));
    mu_assert(value);
    expr_destroy(expr);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests evaluation of a variable expression referencing a non-array variable
 * of type boolean with invalid contents.
 */
mu_test_begin(test_expr_variable_eval_non_array_boolean_invalid)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr;
    BOOL value = 0;
    char *error_msg = NULL;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_BOOLEAN);
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    inst = var_instance_create(var, 0, NULL);
    var_instance_set_value(inst, "maybe", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION);

    expr = expr_variable_create(var_get_global_scope(), inst);
    mu_assert(!expr_get_boolean_value(expr, &value, &error_msg));
    mu_assert_eq_str("converting value 'maybe' of variable 'V' to type boolean failed", error_msg);
    free(error_msg);
    expr_destroy(expr);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests evaluation of a variable expression referencing a non-array variable
 * without a value.
 */
mu_test_begin(test_expr_variable_eval_non_array_empty)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr;
    char *value;
    char *error_msg = NULL;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());

    inst = var_instance_create(var, 0, NULL);
    expr = expr_variable_create(var_get_global_scope(), inst);
    mu_assert(!expr_get_string_value(expr, &value, &error_msg));
    mu_assert_eq_str("reading variable 'V' failed", error_msg);
    free(error_msg);
    expr_destroy(expr);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests evaluation of a variable expression referencing an array variable.
 */
mu_test_begin(test_expr_variable_eval_array)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct var_instance_t *inst_expr;
    struct expr_t *expr;
    char *value;
    char *error_msg;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    inst = var_instance_create(var, 1, NULL);
    inst->indices[0] = 0;
    var_instance_set_and_check_value(inst, "1", VAR_ACCESS_WEAK, VAR_PRIORITY_SCRIPT_TRANSIENT);
    inst->indices[0] = 1;
    var_instance_set_and_check_value(inst, "2", VAR_ACCESS_WEAK, VAR_PRIORITY_SCRIPT_TRANSIENT);

    inst_expr = var_instance_create(var, 1, NULL);
    inst_expr->indices[0] = 0;
    expr = expr_variable_create(var_get_global_scope(), inst_expr);
    mu_assert(expr_get_string_value(expr, &value, &error_msg));
    mu_assert_eq_str("1", value);
    free(value);
    inst->indices[0] = 0;
    var_instance_set_and_check_value(inst, "3", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION);
    mu_assert(expr_get_string_value(expr, &value, &error_msg));
    mu_assert_eq_str("3", value);
    free(value);
    expr_destroy(expr);

    inst_expr = var_instance_create(var, 1, NULL);
    inst_expr->indices[0] = 1;
    expr = expr_variable_create(var_get_global_scope(), inst_expr);
    mu_assert(expr_get_string_value(expr, &value, &error_msg));
    mu_assert_eq_str("2", value);
    free(value);
    inst->indices[0] = 1;
    var_instance_set_and_check_value(inst, "0", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION);
    mu_assert(expr_get_string_value(expr, &value, &error_msg));
    mu_assert_eq_str("0", value);
    free(value);
    expr_destroy(expr);

    var_instance_destroy(inst);
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests toString() on a variable expression referencing a non-array variable.
 */
mu_test_begin(test_expr_variable_tostring_non_array)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr;
    char *value;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    inst = var_instance_create(var, 0, NULL);
    var_instance_set_and_check_value(inst, "2", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION);

    expr = expr_variable_create(var_get_global_scope(), inst);
    value = expr_to_string(expr);
    mu_assert_eq_str("V", value);
    free(value);
    expr_destroy(expr);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests toString() on a variable expression referencing an array variable.
 */
mu_test_begin(test_expr_variable_tostring_array)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr;
    char *value;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());

    inst = var_instance_create(var, 1, NULL);
    inst->indices[0] = 0;
    var_instance_set_and_check_value(inst, "1", VAR_ACCESS_WEAK, VAR_PRIORITY_SCRIPT_TRANSIENT);
    inst->indices[0] = 1;
    var_instance_set_and_check_value(inst, "2", VAR_ACCESS_WEAK, VAR_PRIORITY_SCRIPT_TRANSIENT);
    var_instance_destroy(inst);

    inst = var_instance_create(var, 1, NULL);
    inst->indices[0] = 0;
    expr = expr_variable_create(var_get_global_scope(), inst);
    value = expr_to_string(expr);
    mu_assert_eq_str("V_1", value);
    free(value);
    expr_destroy(expr);

    inst = var_instance_create(var, 1, NULL);
    inst->indices[0] = 1;
    expr = expr_variable_create(var_get_global_scope(), inst);
    value = expr_to_string(expr);
    mu_assert_eq_str("V_2", value);
    free(value);
    expr_destroy(expr);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests equal() on two equal variable expressions.
 */
mu_test_begin(test_expr_variable_equal1)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr1;
    struct expr_t *expr2;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());

    inst = var_instance_create(var, 0, NULL);
    expr1 = expr_variable_create(var_get_global_scope(), inst);
    inst = var_instance_create(var, 0, NULL);
    expr2 = expr_variable_create(var_get_global_scope(), inst);
    mu_assert(expr_equal(expr1, expr2));
    expr_destroy(expr1);
    expr_destroy(expr2);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests equal() on two unequal variable expressions that differ in the
 * variable being referenced.
 */
mu_test_begin(test_expr_variable_equal2)
    struct vartype_t *type;
    struct var_t *var1;
    struct var_t *var2;
    struct var_instance_t *inst;
    struct expr_t *expr1;
    struct expr_t *expr2;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();
    var1 = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    var2 = var_add(var_get_global_scope(), "W", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 24), var_properties_create());

    inst = var_instance_create(var1, 0, NULL);
    expr1 = expr_variable_create(var_get_global_scope(), inst);
    inst = var_instance_create(var2, 0, NULL);
    expr2 = expr_variable_create(var_get_global_scope(), inst);
    mu_assert(!expr_equal(expr1, expr2));
    expr_destroy(expr1);
    expr_destroy(expr2);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests equal() on two unequal variable expressions that differ in the indices
 * used.
 */
mu_test_begin(test_expr_variable_equal3)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    unsigned indices[1];
    struct expr_t *expr1;
    struct expr_t *expr2;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());

    indices[0] = 0;
    inst = var_instance_create(var, 1, indices);
    expr1 = expr_variable_create(var_get_global_scope(), inst);
    indices[0] = 1;
    inst = var_instance_create(var, 1, indices);
    expr2 = expr_variable_create(var_get_global_scope(), inst);
    mu_assert(!expr_equal(expr1, expr2));
    expr_destroy(expr1);
    expr_destroy(expr2);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests clone() on a variable expression referencing a non-array variable.
 */
mu_test_begin(test_expr_variable_clone1)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    struct expr_t *expr1;
    struct expr_t *expr2;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());

    inst = var_instance_create(var, 0, NULL);
    expr1 = expr_variable_create(var_get_global_scope(), inst);
    expr2 = expr_clone(expr1);
    mu_assert(expr_equal(expr1, expr2));
    expr_destroy(expr1);
    expr_destroy(expr2);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests clone() on a variable expression referencing an array variable.
 */
mu_test_begin(test_expr_variable_clone2)
    struct vartype_t *type;
    struct var_t *var;
    struct var_instance_t *inst;
    unsigned indices[1];
    struct expr_t *expr1;
    struct expr_t *expr2;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();
    var = var_add(var_get_global_scope(), "V_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());

    indices[0] = 1;
    inst = var_instance_create(var, 1, indices);
    expr1 = expr_variable_create(var_get_global_scope(), inst);
    expr2 = expr_clone(expr1);
    mu_assert(expr_equal(expr1, expr2));
    expr_destroy(expr1);
    expr_destroy(expr2);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

void
test_expr_variable(void)
{
    mu_run_test(test_expr_variable_kindtostring);
    mu_run_test(test_expr_variable_gettype_string);
    mu_run_test(test_expr_variable_gettype_integer);
    mu_run_test(test_expr_variable_gettype_boolean);
    mu_run_test(test_expr_variable_getkind);
    mu_run_test(test_expr_variable_eval_non_array_string);
    mu_run_test(test_expr_variable_eval_non_array_integer);
    mu_run_test(test_expr_variable_eval_non_array_integer_invalid);
    mu_run_test(test_expr_variable_eval_non_array_boolean);
    mu_run_test(test_expr_variable_eval_non_array_boolean_invalid);
    mu_run_test(test_expr_variable_eval_non_array_empty);
    mu_run_test(test_expr_variable_eval_array);
    mu_run_test(test_expr_variable_tostring_non_array);
    mu_run_test(test_expr_variable_tostring_array);
    mu_run_test(test_expr_variable_equal1);
    mu_run_test(test_expr_variable_equal2);
    mu_run_test(test_expr_variable_equal3);
    mu_run_test(test_expr_variable_clone1);
    mu_run_test(test_expr_variable_clone2);
}
