/*****************************************************************************
 *  @file grammar/tests/test_parser_varass.c
 *  Functions for testing the parsing of variable assignment files.
 *
 *  Copyright (c) 2012 The fli4l team
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: test_parser_varass.c 24288 2012-11-10 18:57:33Z kristov $
 *****************************************************************************
 */

#include <testing/tests.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <libmkfli4l/array.h>
#include <stdlib.h>
#include <string.h>
#include "grammar/varass.h"
#include "grammar/varass_scanner.h"
#include "package.h"
#include "var.h"

#define MU_TESTSUITE "grammar.varass.parser"

#define TEST_PACKAGE "test"

/**
 * Runs the scanner and parser for a given input file.
 *
 * @param name
 *  The name of the input file.
 * @param priority
 *  The access priority to use.
 * @param exit_code
 *  Receives the exit code of the parser.
 * @return
 *  A parse tree.
 */
static struct varass_parse_tree_t *
run_scanner_and_parser(char const *name, int priority, int *exit_code)
{
    struct package_t *package = package_get_or_add(TEST_PACKAGE);
    struct package_file_t *file
        = package_set_file(package, PACKAGE_FILE_TYPE_CONFIG, name);
    struct varass_parse_tree_t *tree = varass_parse_tree_create();

    FILE *f = fopen(name, "r");
    if (f) {
        yyscan_t scanner;
        if (varass_lex_init_extra(file, &scanner) != 0) {
            log_error("Error initializing parser for %s\n", name);
        }
        else {
            varass_set_in(f, scanner);
            *exit_code = varass_parse(scanner, file, tree, priority);
            varass_lex_destroy(scanner);
        }
        fclose(f);
    }
    else {
        log_error("Error opening variable assignment file %s\n", name);
    }

    return tree;
}

/**
 * Tests parsing empty file.
 */
mu_test_begin(test_varass_parser_empty_file)
    struct varass_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, ass, struct varass_t);

    package_init_module();
    tree = run_scanner_and_parser("grammar/tests/input/empty.txt",
            VAR_PRIORITY_CONFIGURATION, &exit_code);

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->assignments);
    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(!ass);

    varass_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing only comments (1).
 */
mu_test_begin(test_varass_parser_comment1)
    struct varass_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, ass, struct varass_t);

    package_init_module();
    tree = run_scanner_and_parser("grammar/tests/input/comment_1.txt",
            VAR_PRIORITY_CONFIGURATION, &exit_code);

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->assignments);
    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(!ass);

    varass_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing only comments (2).
 */
mu_test_begin(test_varass_parser_comment2)
    struct varass_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, ass, struct varass_t);

    package_init_module();
    tree = run_scanner_and_parser("grammar/tests/input/comment_2.txt",
            VAR_PRIORITY_CONFIGURATION, &exit_code);

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->assignments);
    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(!ass);

    varass_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing uncommented assignments.
 */
mu_test_begin(test_varass_parser_ass1)
    struct varass_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, ass, struct varass_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser("grammar/tests/input/varass_ass_1.txt",
            VAR_PRIORITY_CONFIGURATION, &exit_code);

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->assignments);
    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(ass);
    mu_assert(ass->valid);
    mu_assert(ass->id->valid);
    mu_assert_eq_str("VAR1", ass->id->name);
    mu_assert_eq_int(VAR_PRIORITY_CONFIGURATION, ass->priority);
    mu_assert(ass->value->valid);
    mu_assert_eq_int('\'', ass->value->delim);
    mu_assert_eq_str("abc\\\"\\", ass->value->value);
    mu_assert_eq_str("", ass->comment);
    locstr = location_toString(ass->location);
    mu_assert_eq_str("[configuration file 'grammar/tests/input/varass_ass_1.txt' (package test) 1:0]", locstr);
    free(locstr);

    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(ass);
    mu_assert(ass->valid);
    mu_assert(ass->id->valid);
    mu_assert_eq_str("VAR2", ass->id->name);
    mu_assert_eq_int(VAR_PRIORITY_CONFIGURATION, ass->priority);
    mu_assert(ass->value->valid);
    mu_assert_eq_int('"', ass->value->delim);
    mu_assert_eq_str("abc\\\"'", ass->value->value);
    mu_assert_eq_str("", ass->comment);
    locstr = location_toString(ass->location);
    mu_assert_eq_str("[configuration file 'grammar/tests/input/varass_ass_1.txt' (package test) 2:0]", locstr);
    free(locstr);

    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(!ass);

    varass_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented assignment.
 */
mu_test_begin(test_varass_parser_ass2)
    struct varass_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, ass, struct varass_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser("grammar/tests/input/varass_ass_2.txt",
            VAR_PRIORITY_CONFIGURATION, &exit_code);

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->assignments);
    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(ass);
    mu_assert(ass->valid);
    mu_assert(ass->id->valid);
    mu_assert_eq_str("VAR1", ass->id->name);
    mu_assert_eq_int(VAR_PRIORITY_CONFIGURATION, ass->priority);
    mu_assert(ass->value->valid);
    mu_assert_eq_int('\'', ass->value->delim);
    mu_assert_eq_str("abc", ass->value->value);
    mu_assert_eq_str("setting VAR1", ass->comment);
    locstr = location_toString(ass->location);
    mu_assert_eq_str("[configuration file 'grammar/tests/input/varass_ass_2.txt' (package test) 1:0]", locstr);
    free(locstr);

    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(!ass);

    varass_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented assignment distributed across
 * many lines.
 */
mu_test_begin(test_varass_parser_ass3)
    struct varass_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, ass, struct varass_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser("grammar/tests/input/varass_ass_3.txt",
            VAR_PRIORITY_CONFIGURATION, &exit_code);

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->assignments);
    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(ass);
    mu_assert(ass->valid);
    mu_assert(ass->id->valid);
    mu_assert_eq_str("VAR1", ass->id->name);
    mu_assert_eq_int(VAR_PRIORITY_CONFIGURATION, ass->priority);
    mu_assert(ass->value->valid);
    mu_assert_eq_int('\'', ass->value->delim);
    mu_assert_eq_str("abc", ass->value->value);
    mu_assert_eq_str("setting VAR1", ass->comment);
    locstr = location_toString(ass->location);
    mu_assert_eq_str("[configuration file 'grammar/tests/input/varass_ass_3.txt' (package test) 2:0]", locstr);
    free(locstr);

    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(!ass);

    varass_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented single definition and comments
 * above and below.
 */
mu_test_begin(test_varass_parser_ass_comment1)
    struct varass_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, ass, struct varass_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser("grammar/tests/input/varass_ass_comment_1.txt",
            VAR_PRIORITY_CONFIGURATION, &exit_code);

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->assignments);
    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(ass);
    mu_assert(ass->valid);
    mu_assert(ass->id->valid);
    mu_assert_eq_str("VAR1", ass->id->name);
    mu_assert_eq_int(VAR_PRIORITY_CONFIGURATION, ass->priority);
    mu_assert(ass->value->valid);
    mu_assert_eq_int('\'', ass->value->delim);
    mu_assert_eq_str("abc", ass->value->value);
    mu_assert_eq_str("setting VAR1", ass->comment);
    locstr = location_toString(ass->location);
    mu_assert_eq_str("[configuration file 'grammar/tests/input/varass_ass_comment_1.txt' (package test) 2:0]", locstr);
    free(locstr);

    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(!ass);

    varass_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a single definition and comments around it.
 */
mu_test_begin(test_varass_parser_ass_comment2)
    struct varass_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, ass, struct varass_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser("grammar/tests/input/varass_ass_comment_2.txt",
            VAR_PRIORITY_CONFIGURATION, &exit_code);

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->assignments);
    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(ass);
    mu_assert(ass->valid);
    mu_assert(ass->id->valid);
    mu_assert_eq_str("VAR1", ass->id->name);
    mu_assert_eq_int(VAR_PRIORITY_CONFIGURATION, ass->priority);
    mu_assert(ass->value->valid);
    mu_assert_eq_int('\'', ass->value->delim);
    mu_assert_eq_str("abc", ass->value->value);
    mu_assert_eq_str("", ass->comment);
    locstr = location_toString(ass->location);
    mu_assert_eq_str("[configuration file 'grammar/tests/input/varass_ass_comment_2.txt' (package test) 2:0]", locstr);
    free(locstr);

    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(!ass);

    varass_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing assignments with a BOM.
 */
mu_test_begin(test_varass_parser_bom)
    struct varass_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, ass, struct varass_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser("grammar/tests/input/varass_bom.txt",
            VAR_PRIORITY_CONFIGURATION, &exit_code);

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->assignments);
    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(ass);
    mu_assert(ass->valid);
    mu_assert(ass->id->valid);
    mu_assert_eq_str("VAR1", ass->id->name);
    mu_assert_eq_int(VAR_PRIORITY_CONFIGURATION, ass->priority);
    mu_assert(ass->value->valid);
    mu_assert_eq_int('\'', ass->value->delim);
    mu_assert_eq_str("abc\\\"\\", ass->value->value);
    mu_assert_eq_str("", ass->comment);
    locstr = location_toString(ass->location);
    mu_assert_eq_str("[configuration file 'grammar/tests/input/varass_bom.txt' (package test) 1:0]", locstr);
    free(locstr);

    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(ass);
    mu_assert(ass->valid);
    mu_assert(ass->id->valid);
    mu_assert_eq_str("VAR2", ass->id->name);
    mu_assert_eq_int(VAR_PRIORITY_CONFIGURATION, ass->priority);
    mu_assert(ass->value->valid);
    mu_assert_eq_int('"', ass->value->delim);
    mu_assert_eq_str("abc\\\"'", ass->value->value);
    mu_assert_eq_str("", ass->comment);
    locstr = location_toString(ass->location);
    mu_assert_eq_str("[configuration file 'grammar/tests/input/varass_bom.txt' (package test) 2:0]", locstr);
    free(locstr);

    ass = (struct varass_t *) get_next_elem(&it);
    mu_assert(!ass);

    varass_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

void
test_varass_parser(void)
{
    mu_run_test(test_varass_parser_empty_file);
    mu_run_test(test_varass_parser_comment1);
    mu_run_test(test_varass_parser_comment2);
    mu_run_test(test_varass_parser_ass1);
    mu_run_test(test_varass_parser_ass2);
    mu_run_test(test_varass_parser_ass3);
    mu_run_test(test_varass_parser_ass_comment1);
    mu_run_test(test_varass_parser_ass_comment2);
    mu_run_test(test_varass_parser_bom);
}
