/*****************************************************************************
 *  @file grammar/tests/test_parser_vardef.c
 *  Functions for testing the parsing of variable definition files.
 *
 *  Copyright (c) 2012 The fli4l team
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: test_parser_vardef.c 23639 2012-09-11 07:50:40Z kristov $
 *****************************************************************************
 */

#include <testing/tests.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>
#include <string.h>
#include "grammar/vardef.h"
#include "grammar/vardef_scanner.h"
#include "package.h"

#define MU_TESTSUITE "grammar.vardef.parser"

#define TEST_PACKAGE "test"

/**
 * Runs the scanner and parser for a given input file.
 *
 * @param name
 *  The name of the input file.
 * @param exit_code
 *  Receives the exit code of the parser.
 * @return
 *  An array of tokens which have been recognized.
 */
static struct vardef_parse_tree_t *
run_scanner_and_parser(char const *name, int *exit_code)
{
    struct package_t *package = package_get_or_add(TEST_PACKAGE);
    struct package_file_t *file
        = package_set_file(package, PACKAGE_FILE_TYPE_VARDEF, name);
    struct vardef_parse_tree_t *tree = vardef_parse_tree_create();

    FILE *f = fopen(name, "r");
    if (f) {
        yyscan_t scanner;
        if (vardef_lex_init_extra(file, &scanner) != 0) {
            log_error("Error initializing scanner for %s\n", name);
        }
        else {
            vardef_set_in(f, scanner);
            *exit_code = vardef_parse(scanner, file, tree);
            vardef_lex_destroy(scanner);
        }
        fclose(f);
    }
    else {
        log_error("Error opening variable definition file %s\n", name);
    }

    return tree;
}

/**
 * Tests parsing empty file.
 */
mu_test_begin(test_vardef_parser_empty_file)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);

    package_init_module();
    tree = run_scanner_and_parser("grammar/tests/input/empty.txt", &exit_code);

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing only comments (1).
 */
mu_test_begin(test_vardef_parser_comment1)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/comment_1.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing only comments (2).
 */
mu_test_begin(test_vardef_parser_comment2)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/comment_2.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a single definition only.
 */
mu_test_begin(test_vardef_parser_def1)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_1.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("abc", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_1.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a single commented definition.
 */
mu_test_begin(test_vardef_parser_def2)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_2.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("abc", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_2.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a single commented definition without a
 * default value.
 */
mu_test_begin(test_vardef_parser_def3)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_3.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(!def->defvalue);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_3.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a single commented definition of an optional
 * variable.
 */
mu_test_begin(test_vardef_parser_def4)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_4.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("123", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_4.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a single commented definition of a really
 * optional array.
 */
mu_test_begin(test_vardef_parser_def5)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_5.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1_%_XY", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("123", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1_%_XY", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_5.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a single commented definition of a really
 * optional array.
 */
mu_test_begin(test_vardef_parser_def6)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_6.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1_N", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(!def->defvalue);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_6.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1_%_XY", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("123", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1_%_XY", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_6.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing two definitions of which one variable depends
 * on the other one.
 */
mu_test_begin(test_vardef_parser_def7a)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_7a.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("YESNO", def->type->u.id->name);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(!def->defvalue);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_7a.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR2", def->id->name);
    mu_assert(def->cond->valid);
    mu_assert(def->cond->var->valid);
    mu_assert_eq_str("VAR1", def->cond->var->name);
    mu_assert_eq_int(FALSE, def->cond->negated);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("123", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_7a.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing two definitions of which one variable depends
 * on the inverse of another one.
 */
mu_test_begin(test_vardef_parser_def7b)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_7b.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("YESNO", def->type->u.id->name);
    mu_assert(!def->defvalue);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_7b.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR2", def->id->name);
    mu_assert(def->cond->valid);
    mu_assert(def->cond->var->valid);
    mu_assert_eq_str("VAR1", def->cond->var->name);
    mu_assert_eq_int(TRUE, def->cond->negated);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("123", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_7b.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing two definitions of which one variable depends
 * on the inverse of another one.
 */
mu_test_begin(test_vardef_parser_def8)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_8.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(FALSE, def->type->named);
    mu_assert_eq_str("abc|(def)+|(e*fg*)?", def->type->u.regex);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("abc\\", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_8.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented single definition and comments
 * above and below.
 */
mu_test_begin(test_vardef_parser_def_comment1)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_comment_1.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("abc", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_comment_1.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a single definition and comments around it.
 */
mu_test_begin(test_vardef_parser_def_comment2)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_comment_2.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("abc", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_comment_2.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a multi-line commented single definition,
 * comments above and below, and LF line endings.
 */
mu_test_begin(test_vardef_parser_def_comment3a)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_comment_3a.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("abc", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_comment_3a.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a multi-line commented single definition,
 * comments above and below, and CR/LF line endings.
 */
mu_test_begin(test_vardef_parser_def_comment3b)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_comment_3b.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("abc", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_comment_3b.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines.
 */
mu_test_begin(test_vardef_parser_def9)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_9.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("123", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_9.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, double-quoted multi-line strings, and LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline1a)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_1a.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("1 23", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_1a.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, double-quoted multi-line strings, and CR/LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline1b)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_1b.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("1 23", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_1b.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, double-quoted continued multi-line strings, and LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline2a)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_2a.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("1  23", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_2a.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, double-quoted continued multi-line strings, and CR/LF line
 * endings.
 */
mu_test_begin(test_vardef_parser_string_multiline2b)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_2b.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("1  23", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_2b.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, single-quoted multi-line strings, and LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline3a)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_3a.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("1 23", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_3a.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, single-quoted multi-line strings, and CR/LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline3b)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_3b.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("1 23", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_3b.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, single-quoted "continued" multi-line strings, and LF line
 * endings.
 */
mu_test_begin(test_vardef_parser_string_multiline4a)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_4a.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("1\\ 23", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_4a.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, single-quoted "continued" multi-line strings, and CR/LF line
 * endings.
 */
mu_test_begin(test_vardef_parser_string_multiline4b)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_4b.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("1\\ 23", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_4b.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, a double-quoted, incorrectly continued multi-line string, and LF
 * line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline5a)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_5a.txt", &exit_code
    );

    mu_assert(!tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(!def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(!def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("1 ", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_5a.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, a double-quoted, incorrectly continued multi-line string, and
 * CR/LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline5b)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_5b.txt", &exit_code
    );

    mu_assert(!tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(!def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(!def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("1 ", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_5b.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, a single-quoted, incorrectly continued multi-line string, and LF
 * line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline6a)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_6a.txt", &exit_code
    );

    mu_assert(!tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(!def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(!def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("1 ", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_6a.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, a single-quoted, incorrectly continued multi-line string, and
 * CR/LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline6b)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_6b.txt", &exit_code
    );

    mu_assert(!tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(!def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(!def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("1 ", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_6b.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, double-quoted multi-line strings contaning empty lines, and LF
 * line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline7a)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_7a.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("1 23", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_7a.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, double-quoted multi-line strings contaning empty lines, and
 * CR/LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline7b)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_7b.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("1 23", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_7b.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, single-quoted multi-line strings contaning empty lines, and LF
 * line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline8a)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_8a.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("1 23", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_8a.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, single-quoted multi-line strings contaning empty lines, and
 * CR/LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline8b)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_8b.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("1 23", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_8b.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, a double-quoted, incorrectly continued multi-line string
 * containing empty lines, and LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline9a)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_9a.txt", &exit_code
    );

    mu_assert(!tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(!def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(!def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("1 ", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_9a.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, a double-quoted, incorrectly continued multi-line string+
 * contaning empty lines, and CR/LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline9b)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_9b.txt", &exit_code
    );

    mu_assert(!tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(!def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(!def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("1 ", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_9b.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, a single-quoted, incorrectly continued multi-line string
 * containing empty lines, and LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline10a)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_10a.txt", &exit_code
    );

    mu_assert(!tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(!def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(!def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("1 ", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_10a.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented definition distributed across
 * many lines, a single-quoted, incorrectly continued multi-line string+
 * contaning empty lines, and CR/LF line endings.
 */
mu_test_begin(test_vardef_parser_string_multiline10b)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_multiline_10b.txt", &exit_code
    );

    mu_assert(!tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(!def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(!def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("1 ", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_multiline_10b.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing strings with escape characters (1).
 */
mu_test_begin(test_vardef_parser_string_escape1)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_escape_1.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("\\abc", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_escape_1.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing strings with escape characters (2).
 */
mu_test_begin(test_vardef_parser_string_escape2)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_escape_2.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("a\\bc", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_escape_2.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing strings with escape characters (3).
 */
mu_test_begin(test_vardef_parser_string_escape3)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_escape_3.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("abc\\", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_escape_3.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing strings with escape characters (4).
 */
mu_test_begin(test_vardef_parser_string_escape4)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_escape_4.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("\\abc", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_escape_4.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing strings with escape characters (5).
 */
mu_test_begin(test_vardef_parser_string_escape5)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_escape_5.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("a\\bc", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_escape_5.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing strings with escape characters (6).
 */
mu_test_begin(test_vardef_parser_string_escape6)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_escape_6.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("abc\\", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_escape_6.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing strings with escape characters (7).
 */
mu_test_begin(test_vardef_parser_string_escape7)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_string_escape_7.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("\\\"abc\"\''\"", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_string_escape_7.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a wrong identifier.
 */
mu_test_begin(test_vardef_parser_def_wrong1)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_wrong_1.txt", &exit_code
    );

    mu_assert(!tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(!def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(!def->id->valid);
    mu_assert_eq_str("VaR1_%_XY", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("123", def->defvalue->value);
    mu_assert_eq_str("definition of VaR1_%_XY", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_wrong_1.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(!def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(!def->id->valid);
    mu_assert_eq_str("VAr1_%_XY", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("123", def->defvalue->value);
    mu_assert_eq_str("definition of VAr1_%_XY", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_wrong_1.txt' (package test) 2:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(!def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(!def->id->valid);
    mu_assert_eq_str("var1_%_xy", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("123", def->defvalue->value);
    mu_assert_eq_str("definition of var1_%_xy", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_wrong_1.txt' (package test) 3:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a wrong condition.
 */
mu_test_begin(test_vardef_parser_def_wrong2)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_wrong_2.txt", &exit_code
    );

    mu_assert(!tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing an unclosed string.
 */
mu_test_begin(test_vardef_parser_def_wrong3)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_def_wrong_3.txt", &exit_code
    );

    mu_assert(!tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(!def->valid);
    mu_assert_eq_int(TRUE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1_%_XY", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NUMERIC", def->type->u.id->name);
    mu_assert(!def->defvalue->valid);
    mu_assert_eq_int('"', def->defvalue->delim);
    mu_assert_eq_str("123 # definition of VAR1_%_XY ", def->defvalue->value);
    mu_assert_eq_str("", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_def_wrong_3.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented single definition with a BOM.
 */
mu_test_begin(test_vardef_parser_bom)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_bom.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("abc", def->defvalue->value);
    mu_assert_eq_str("definition of VAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_bom.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

/**
 * Tests parsing file containing a commented single definition with TABs.
 */
mu_test_begin(test_vardef_parser_tabs)
    struct vardef_parse_tree_t *tree;
    int exit_code;
    DECLARE_ARRAY_ITER(it, def, struct vardef_t);
    char *locstr;

    package_init_module();
    tree = run_scanner_and_parser(
        "grammar/tests/input/vardef_tabs.txt", &exit_code
    );

    mu_assert(tree->ok);

    init_array_iterator(&it, tree->definitions);
    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(def);
    mu_assert(def->valid);
    mu_assert_eq_int(FALSE, def->optional);
    mu_assert(def->id->valid);
    mu_assert_eq_str("VAR1", def->id->name);
    mu_assert(!def->cond);
    mu_assert(def->type->valid);
    mu_assert_eq_int(TRUE, def->type->named);
    mu_assert(def->type->u.id->valid);
    mu_assert_eq_str("NONE", def->type->u.id->name);
    mu_assert(def->defvalue->valid);
    mu_assert_eq_int('\'', def->defvalue->delim);
    mu_assert_eq_str("a\tbc", def->defvalue->value);
    mu_assert_eq_str("definition of\tVAR1", def->comment);
    locstr = location_toString(def->location);
    mu_assert_eq_str("[variable definition file 'grammar/tests/input/vardef_tabs.txt' (package test) 1:0]", locstr);
    free(locstr);

    def = (struct vardef_t *) get_next_elem(&it);
    mu_assert(!def);

    vardef_parse_tree_destroy(tree);
    package_fini_module();
mu_test_end()

void
test_vardef_parser(void)
{
    mu_run_test(test_vardef_parser_empty_file);
    mu_run_test(test_vardef_parser_comment1);
    mu_run_test(test_vardef_parser_comment2);
    mu_run_test(test_vardef_parser_def1);
    mu_run_test(test_vardef_parser_def2);
    mu_run_test(test_vardef_parser_def3);
    mu_run_test(test_vardef_parser_def4);
    mu_run_test(test_vardef_parser_def5);
    mu_run_test(test_vardef_parser_def6);
    mu_run_test(test_vardef_parser_def7a);
    mu_run_test(test_vardef_parser_def7b);
    mu_run_test(test_vardef_parser_def8);
    mu_run_test(test_vardef_parser_def9);
    mu_run_test(test_vardef_parser_def_comment1);
    mu_run_test(test_vardef_parser_def_comment2);
    mu_run_test(test_vardef_parser_def_comment3a);
    mu_run_test(test_vardef_parser_def_comment3b);
    mu_run_test(test_vardef_parser_string_multiline1a);
    mu_run_test(test_vardef_parser_string_multiline1b);
    mu_run_test(test_vardef_parser_string_multiline2a);
    mu_run_test(test_vardef_parser_string_multiline2b);
    mu_run_test(test_vardef_parser_string_multiline3a);
    mu_run_test(test_vardef_parser_string_multiline3b);
    mu_run_test(test_vardef_parser_string_multiline4a);
    mu_run_test(test_vardef_parser_string_multiline4b);
    mu_run_test(test_vardef_parser_string_multiline5a);
    mu_run_test(test_vardef_parser_string_multiline5b);
    mu_run_test(test_vardef_parser_string_multiline6a);
    mu_run_test(test_vardef_parser_string_multiline6b);
    mu_run_test(test_vardef_parser_string_multiline7a);
    mu_run_test(test_vardef_parser_string_multiline7b);
    mu_run_test(test_vardef_parser_string_multiline8a);
    mu_run_test(test_vardef_parser_string_multiline8b);
    mu_run_test(test_vardef_parser_string_multiline9a);
    mu_run_test(test_vardef_parser_string_multiline9b);
    mu_run_test(test_vardef_parser_string_multiline10a);
    mu_run_test(test_vardef_parser_string_multiline10b);
    mu_run_test(test_vardef_parser_string_escape1);
    mu_run_test(test_vardef_parser_string_escape2);
    mu_run_test(test_vardef_parser_string_escape3);
    mu_run_test(test_vardef_parser_string_escape4);
    mu_run_test(test_vardef_parser_string_escape5);
    mu_run_test(test_vardef_parser_string_escape6);
    mu_run_test(test_vardef_parser_string_escape7);
    mu_run_test(test_vardef_parser_def_wrong1);
    mu_run_test(test_vardef_parser_def_wrong2);
    mu_run_test(test_vardef_parser_def_wrong3);
    mu_run_test(test_vardef_parser_bom);
    mu_run_test(test_vardef_parser_tabs);
}
