/*****************************************************************************
 *  @file grammar/varass.h
 *  Functions for scanning/parsing files containing variable assignments.
 *
 *  Copyright (c) 2012 The fli4l team
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: varass.h 23640 2012-09-11 08:21:24Z kristov $
 *****************************************************************************
 */

#ifndef MKFLI4L_GRAMMAR_VARASS_H_
#define MKFLI4L_GRAMMAR_VARASS_H_

#include "libmkfli4l/defs.h"
#include "libmkfli4l/array.h"
#include "grammar/common.h"
#include "package.h"

/**
 * @name Assignments
 * @{
 */

/**
 * Represents an assignment of a variable.
 */
struct varass_t {
    /**
     * The identifier.
     */
    struct identifier_t *id;
    /**
     * The associated value.
     */
    struct qstr_t *value;
    /**
     * The access priority to be used.
     */
    int priority;
    /**
     * The associated comment.
     */
    char *comment;
    /**
     * The location at which this definition is to be found.
     */
    struct location_t *location;
    /**
     * TRUE if assignment is valid, FALSE otherwise.
     */
    BOOL valid;
};

/**
 * Creates a variable assignment.
 *
 * @param id
 *  The identifier of the variable to assign to.
 * @param value
 *  The associated value.
 * @param priority
 *  The access priority to be used.
 * @param comment
 *  The associated comment.
 * @param location
 *  The location of the definition.
 * @return
 *  A varass_t object. Use varass_destroy() to free it after use.
 */
struct varass_t *
varass_create(
    struct identifier_t *id,
    struct qstr_t *value,
    int priority,
    char *comment,
    struct location_t *location
);

/**
 * Destroys a variable assignment.
 *
 * @param ass
 *  The variable assignment to destroy. May be NULL.
 */
void
varass_destroy(struct varass_t *ass);

/**
 * @}
 */

/**
 * @name Parse tree
 * @{
 */

/**
 * Represents the results of parsing the variable assignment files.
 */
struct varass_parse_tree_t {
    /**
     * TRUE if parse was successful, FALSE otherwise.
     */
    BOOL ok;
    /**
     * The variable assignments parsed.
     */
    array_t *assignments;
};

/**
 * Creates a parse tree structure for variable assignment files.
 *
 * @return
 *  A varass_parse_tree_t object. Use varass_parse_tree_destroy() to free it
 *  after use.
 */
struct varass_parse_tree_t *
varass_parse_tree_create(void);

/**
 * Adds a variable assignment to the parse tree.
 *
 * @param tree
 *  The parse tree.
 * @param ass
 *  The variable assignment to add.
 */
void
varass_parse_tree_add_assignment(
    struct varass_parse_tree_t *tree,
    struct varass_t *ass
);

/**
 * Destroys a parse tree structure for variable assignment files.
 *
 * @param tree
 *  The parse tree to destroy.
 */
void
varass_parse_tree_destroy(struct varass_parse_tree_t *tree);

/**
 * @}
 */

/**
 * The parser's entry point.
 *
 * @param scanner
 *  The scanner to use.
 * @param file
 *  The package file that is to be parsed.
 * @param tree
 *  The parse tree to fill.
 * @param priority
 *  The access priority to be used.
 * @return
 *  Zero if successful, a non-zero value otherwise.
 */
int
varass_parse(
    void *scanner,
    struct package_file_t *file,
    struct varass_parse_tree_t *tree,
    int priority
);

#endif
