/*----------------------------------------------------------------------------
 *  check.c - functions for checking variables
 *
 *  Copyright (c) 2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Creation:    2001-08-12  fm
 *  Last Update: $Id: check.c 34688 2014-11-07 12:56:15Z florian $
 *----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include "check.h"
#include "var.h"
#include "regex.h"
#include "log.h"
#include "parse.h"
#include "str.h"
#include "cfg.h"
#include "array.h"

#define CHECK_ARRAY_SIZE        1024

#define SUNDAY              0                       /* 1st day of unix week */
#define HOURS_PER_DAY       24                      /* hours per day        */
#define DAYS_PER_WEEK       7                       /* days per week        */

static int time_table[HOURS_PER_DAY * DAYS_PER_WEEK];  /* 24 * 7 hours */
static int local_time_table[HOURS_PER_DAY * DAYS_PER_WEEK];  /* 24 * 7 hours */
static char **time_variables;
static char **time_values;

static char const *en_wday_strings[DAYS_PER_WEEK] = {
    "Su", "Mo", "Tu", "We", "Th", "Fr", "Sa",
};

static char const *de_wday_strings[DAYS_PER_WEEK] = {
    "So", "Mo", "Di", "Mi", "Do", "Fr", "Sa",
};

struct flags_t {
    BOOL type_numeric;
    BOOL optional;
    BOOL really_optional;
    BOOL weak;
    BOOL neg_opt;
};

struct check {
    char *name;
    char *opt_var;
    char *var_n;
    char *regexp_name;
    char *regexp;
    char *defval;
    char *package;
    struct flags_t flags;
};

static array_t *check_array;

static struct check *
lookup_check_var(char const *name)
{
    ARRAY_ITER(array_iter, check_array, p, struct check) {
        if (strcmp (p->name, name) == 0) {
            return p;
        }
    }
    return NULL;
}

BOOL
check_var_defined(char const *var)
{
    struct check *p = lookup_check_var(var);
    return p ? TRUE : FALSE;
}

BOOL
check_var_numeric(char const *var)
{
    struct check *p = lookup_check_var(var);
    return p ? p->flags.type_numeric : FALSE;
}

BOOL
check_var_weak(char const *var)
{
    struct check *p = lookup_check_var(var);
    return p ? p->flags.weak : FALSE;
}

BOOL
check_var_optional(char const *var)
{
    struct check *p = lookup_check_var(var);
    return p ? (p->flags.optional || p->flags.really_optional) : FALSE;
}

BOOL
check_var_really_optional(char const *var)
{
    struct check *p = lookup_check_var(var);
    return p ? p->flags.really_optional : FALSE;
}

int
check_var_optional_state(char const *var)
{
    struct check *p = lookup_check_var(var);
    if (p) {
        int res = 0;
        if (p->flags.really_optional)
            res |= CHECK_REALLY_OPTIONAL;
        if (p->flags.optional)
            res |= CHECK_OPTIONAL;
        return res;
    }
    else {
        return 0;
    }
}

char *
check_get_opt_var(char const *var)
{
    struct check *p = lookup_check_var(var);
    return p ? p->opt_var : NULL;
}

char *
check_get_var_n(char const *var)
{
    struct check *p = lookup_check_var(var);
    return p ? p->var_n : NULL;
}

static BOOL
check_var_n(char const *var, char const *opt_var, char const *var_n, char const *file, int line)
{
    BOOL result = TRUE;

    if (!check_var_defined(var_n)) {
        log_error("%s:%d: Variable %s depends on var_n %s"
                  " which isn't defined in any check file\n",
                  file, line, var, var_n);
        result = FALSE;
    }
    else {
        char *opt_var_n = check_get_opt_var(var_n);
        if (opt_var && opt_var_n) {
            if (strcmp(opt_var, opt_var_n) != 0 && strchr(opt_var, '%') == NULL) {
                log_error ("%s:%d: var_n depends on different opt_var\n"
                           "\tVariable: '%s' -> '%s'\n"
                           "\tVar_n:    '%s' -> '%s'\n",
                            file, line, var, opt_var, var_n, opt_var_n);
                result = FALSE;
            }
        }
        else if (opt_var_n) {
            log_error ("%s:%d: var_n depends on opt_var while"
                       " var doesn't depend on opt_var\n"
                       "\tVariable: '%s' -> no dependency\n"
                       "\tVar_n:    '%s' -> '%s'\n",
                       file, line, var, var_n, opt_var_n);
            result = FALSE;
        }
    }

    if (!check_var_optional(var)) {
        if (var_n != NULL && check_var_optional(var_n)) {
            log_error ("%s:%d: mandatory variable %s depends on "
                      "optional var_n %s\n", file, line, var, var_n);
            result = FALSE;
        }
    }

    return result;
}

#define CHK_NAME        0
#define CHK_OPT_VAR     1
#define CHK_VAR_N       2
#define CHK_VAL         3
#define CHK_DEFVAL      4

void check_init(void)
{
    check_array = init_array(CHECK_ARRAY_SIZE);
}

int
read_check_file(char const *fname, char const *package) {
    int result = OK;
    int ret;

    struct token_t tokens[] = {
        TOKEN(CFG_ID | CFG_OPT_ID | CFG_REALLY_OPT_ID),
        TOKEN(CFG_ID | CFG_NEG_ID | CFG_HYPHEN),
        TOKEN(CFG_ID | CFG_OPT_ID | CFG_HYPHEN),
        TOKEN(CFG_ID | CFG_REGEXP),
        TOKEN(CFG_STRING|CFG_NL),
        TOKEN(CFG_NONE),
    };

    cfg_fopen(fname);
    while ((ret = get_config_tokens (tokens, 0)) != CFG_EOF) {
        int type;
        struct check *c;

        if (ret == CFG_ERROR) {
            result = ERR;
            continue;
        }

        c = (struct check *) safe_malloc(sizeof(struct check));
        c->package = strsave(package);
        c->flags.type_numeric = FALSE;
        c->flags.optional = FALSE;
        c->flags.really_optional = FALSE;
        c->flags.weak = FALSE;
        c->flags.neg_opt = FALSE;
        append_new_elem(check_array, c);

        switch (tokens[CHK_NAME].token) {
        case CFG_ID:
            break;
        case CFG_OPT_ID:
            c->flags.optional = TRUE;
            log_info(LOG_VERBOSE, "optional variable %s found\n",
                     tokens[CHK_NAME].text);
            break;
        case CFG_REALLY_OPT_ID:
            c->flags.optional = TRUE;
            c->flags.really_optional = TRUE;
            log_info(LOG_VERBOSE, "really optional variable %s found\n",
                     tokens[CHK_NAME].text);
            break;
        default:
            fatal_exit("unexpected value for variable name\n");
        }

        c->name = tokens[CHK_NAME].text;

        switch (tokens[CHK_OPT_VAR].token) {
        case CFG_HYPHEN:
            c->opt_var = NULL;
            break;
        case CFG_ID:
            c->opt_var = tokens[CHK_OPT_VAR].text;
            break;
        case CFG_NEG_ID:
            c->opt_var = tokens[CHK_OPT_VAR].text;
            c->flags.neg_opt = TRUE;
            break;
        default:
            fatal_exit("unexpected value for opt variable\n");
        }

        switch (tokens[CHK_VAR_N].token) {
        case CFG_HYPHEN:
            c->var_n = NULL;
            break;
        case CFG_ID:
            c->var_n = tokens[CHK_VAR_N].text;
            break;
        default:
            fatal_exit("unexpected value for var_n\n");
        }

        if (c->var_n && strchr(c->name, '%') == NULL) {
            log_error("%s:%d  missing '%%' in set variable %s\n",
                      fname, tokens[CHK_NAME].line, c->name);
            result = ERR;
            continue;
        }

        switch (tokens[CHK_VAL].token) {
        case CFG_ID:
            type = regexp_find_type(tokens[CHK_VAL].text);
            if (type >= 0) {
                if (type == TYPE_NUMERIC) {
                    c->flags.type_numeric = TRUE;
                }
            }
            else {
                log_error("%s: line %d: unknown value '%s'\n",
                          fname, tokens[CHK_NAME].line, tokens[CHK_VAL].text);
                result = ERR;
            }
            c->regexp_name = tokens[CHK_VAL].text;
            c->regexp = NULL;
            break;
        case CFG_REGEXP:
            /* user defined regular expression */
            c->regexp_name = NULL;
            c->regexp = tokens[CHK_VAL].text;
            break;
        default:
            fatal_exit("unexpected value for check expression\n");
        }

        switch (tokens[CHK_DEFVAL].token) {
        case CFG_NL:
            c->defval = NULL;
            break;
        case CFG_STRING:
            if (c->var_n != NULL) {
                log_error("You can't specify a default value for an array "
                          "variable: %s='%s'\n",
                          c->name, tokens[CHK_DEFVAL].text);
                c->defval = NULL;
                result = ERR;
            }
            else if (c->flags.optional) {
                log_error("You can't specify a default value for an optional "
                          "variable: %s='%s'\n",
                          c->name, tokens[CHK_DEFVAL].text);
                c->defval = NULL;
                result = ERR;
            }
            else {
                c->defval = tokens[CHK_DEFVAL].text;
                log_info(LOG_VAR, "default: %s='%s'\n", c->name, c->defval);
            }
            break;
        default:
            fatal_exit("unexpected value for default expression\n");
        }

        if (c->opt_var != NULL && !check_var_defined(c->opt_var)) {
            log_error("Variable %s depends on opt_var %s"
                      " which isn't checked yet\n",
                      c->name, c->opt_var);
            result = ERR;
        }

        if (!c->flags.optional &&
                c->opt_var != NULL && check_var_optional(c->opt_var)) {
            log_error("mandatory variable %s depends on "
                      "optional variable %s\n", c->name, c->opt_var);
            result = ERR;
        }

        if (c->var_n != NULL &&
                !check_var_n(c->name, c->opt_var, c->var_n, fname, tokens[CHK_NAME].line)) {
            result = ERR;
        }
    }

    cfg_fclose();
    return result;
}

char *
get_set_var_n(char const *variable)
{
    struct check *c;
    char *ret = NULL;

    char *q = strip_multiple_indices(variable);
    char *p = replace_set_var_indices(q);
    free(q);

    c = lookup_check_var(p);
    if (c) {
        ret = c->var_n;
    }
    free(p);
    return ret;
}

void
check_add_weak_declaration(char const *package, char const *name, char const *value,
                           char **var_n, char const *file, int line,
                           int log_level)
{
    char *weak_var_n;
    char *p;

    /* try to find check rule for % variable */
    struct check * c = lookup_check_var(name);
    if (c) {
        if (var_n) {
            *var_n = NULL;
        }
        return;
    }

    log_info(log_level, "adding weak declaration for %s\n", name);
    /* construct var_n name */
    weak_var_n = strsave(name);
    if ((p = strchr(weak_var_n, '%')) == NULL) {
        fatal_exit("check_add_weak_declaration: can't add non set var %s\n",
                   name);
    }
    *p = 'N';
    *(p + 1) = '\0';

    /* ok, no check rule present, so lets create one */
    c = (struct check *) safe_malloc(sizeof(struct check));
    c->name        = weak_var_n;
    c->opt_var     = NULL;
    c->var_n       = NULL;
    c->regexp_name = NULL;
    c->regexp      = NULL;
    c->defval      = NULL;
    c->flags.type_numeric    = TRUE;
    c->flags.optional        = FALSE;
    c->flags.really_optional = FALSE;
    c->flags.weak            = TRUE;
    c->flags.neg_opt         = FALSE;
    append_new_elem(check_array, c);


    c = (struct check *) safe_malloc(sizeof(struct check));
    c->name        = strsave(name);
    c->opt_var     = NULL;
    c->var_n       = strsave(weak_var_n);
    c->regexp_name = NULL;
    c->regexp      = NULL;
    c->defval      = NULL;
    c->flags.type_numeric    = FALSE;
    c->flags.optional        = TRUE;
    c->flags.really_optional = TRUE;
    c->flags.weak            = TRUE;
    c->flags.neg_opt         = FALSE;
    append_new_elem(check_array, c);

    var_add_weak_declaration("internal", weak_var_n, value,
                             "internal variable used to handle % variables"
                             " declared inside opt/package.txt",
                             TYPE_NUMERIC, file, line, LOG_TREE_EXEC|LOG_VAR);

    mark_var_checked(weak_var_n);
    if (var_n) {
        *var_n = weak_var_n;
    }
    (void) package;
}


/*----------------------------------------------------------------------------
 * convert_week_day_to_day ()               - convert week day to index 0 - 6
 *----------------------------------------------------------------------------
 */
static int
convert_week_day_to_day(char const *week_day)
{
    int i;

    for (i = 0; i < DAYS_PER_WEEK; ++i) {
        if (strcmp(week_day, en_wday_strings[i]) == 0 ||
                strcmp(week_day, de_wday_strings[i]) == 0) {
            return (i);
        }
    }

    return -1;
}

/*----------------------------------------------------------------------------
 * init_local_time_table ()                 - initialize local time table
 *----------------------------------------------------------------------------
 */
static void
init_local_time_table(void)
{
    int i;
    for (i = 0; i < DAYS_PER_WEEK*HOURS_PER_DAY; ++i) {
        local_time_table[i] = -1;
    }
}

/*----------------------------------------------------------------------------
 * check_local_time_table () - check whether time table spans the whole week
 *----------------------------------------------------------------------------
 */
static int
check_local_time_table(int times_idx)
{
    int day;
    int hour;
    int start_day = 0;
    int start_hour = 0;
    BOOL found = FALSE;
    int ret = OK;

    for (day = 0; day < DAYS_PER_WEEK; ++day) {
        for (hour = 0; hour < HOURS_PER_DAY; ++hour) {
            if (!found) {
                if (local_time_table[day * HOURS_PER_DAY + hour] == -1) {
                    found = TRUE;
                    ret = ERR;
                    start_day = day;
                    start_hour = hour;
                }
            }
            else {
                if (local_time_table[day * HOURS_PER_DAY + hour] != -1) {
                    found = FALSE;
                    log_error("Error: undefined time span in variable %s starting at %s:%d ending at %s:%d\n",
                              time_variables[times_idx],
                              de_wday_strings[start_day], start_hour,
                              de_wday_strings[day], hour);
                }
            }
        }
    }

    if (found) {
        log_error("Error: undefined time span in variable %s starting at %s:%d ending at %s:%d\n",
                  time_variables[times_idx],
                  de_wday_strings[start_day], start_hour,
                  de_wday_strings[DAYS_PER_WEEK - 1], HOURS_PER_DAY);
    }

    return ret;
}

/*----------------------------------------------------------------------------
 * fill_time_table ()                       - fill time tables
 *----------------------------------------------------------------------------
 */
static int
fill_global_time_table(int times_idx, int start_time, int end_time)
{
    int i;

    for (i = start_time; i < end_time; i++) {
        if (time_table[i] < 0) {
            time_table[i] = times_idx;
        }
        else {
            if (time_table[i] == times_idx) {
                log_error("Error: overlapping time ranges in variable %s\n",
                          time_variables[times_idx]);
            }
            else {
                log_error("Error: overlapping time ranges in variables %s and %s\n",
                          time_variables[time_table[i]], time_variables[times_idx]);
            }
            return ERR;
        }
    }

    return OK;
}

static int
fill_local_time_table(int times_idx, int start_time, int end_time)
{
    int i;

    for (i = start_time; i < end_time; i++) {
        if (local_time_table[i] < 0) {
            local_time_table[i] = 0;
        }
        else {
            log_error("Error: overlapping time ranges in variable %s\n",
                      time_variables[times_idx]);
            return ERR;
        }
    }
    return OK;
}

static int
fill_time_table(int times_idx, BOOL lcr, int start_time, int end_time)
{
    int res = OK;

    if (lcr) {
        res = fill_global_time_table(times_idx, start_time, end_time);
    }
    if (res == OK) {
        res = fill_local_time_table(times_idx, start_time, end_time);
    }
    return res;
}

/*----------------------------------------------------------------------------
 *  check_a_time_value (int times_idx,
 *----------------------------------------------------------------------------
 */
static int
check_a_time_value(int times_idx, BOOL lcr,
                   char const *start_week_day_str, char const *end_week_day_str,
                   char const *start_hour_str, char const *end_hour_str)
{
    int start_day;
    int end_day;
    int start_hour;
    int end_hour;
    int day;

    start_day = convert_week_day_to_day(start_week_day_str);
    end_day   = convert_week_day_to_day(end_week_day_str);

    if (start_day < 0) {
        log_error("Error: format error in variable %s: no weekday: %s\n",
                  time_variables[times_idx], start_week_day_str);
        return ERR;
    }

    if (end_day < 0) {
        log_error("Error: format error in variable %s: no weekday: %s\n",
                  time_variables[times_idx], end_week_day_str);
        return ERR;
    }

    start_hour = atoi (start_hour_str);
    end_hour   = atoi (end_hour_str);

    if (start_hour < 0 || start_hour >= 24) {
        log_error("Error: format error in variable %s: no valid hour: %s\n",
                  time_variables[times_idx], start_hour_str);
        return ERR;
    }

    if (end_hour <= 0 || end_hour > 24) {
        log_error("Error: format error in variable %s: no valid hour: %s\n",
                  time_variables[times_idx], end_hour_str);
        return ERR;
    }

    day = start_day;

    while (TRUE) {
        int start_time;
        int end_time;

        if (start_hour > end_hour) {
            start_time = day * HOURS_PER_DAY + start_hour;
            end_time   = day * HOURS_PER_DAY + HOURS_PER_DAY;

            if (fill_time_table(times_idx, lcr, start_time, end_time) != OK) {
                return ERR;
            }

            start_time = day * HOURS_PER_DAY + 0;
            end_time   = day * HOURS_PER_DAY + end_hour;

            if (fill_time_table(times_idx, lcr, start_time, end_time) != OK) {
                return ERR;
            }
        }
        else {
            start_time = day * HOURS_PER_DAY + start_hour;
            end_time   = day * HOURS_PER_DAY + end_hour;

            if (fill_time_table(times_idx, lcr, start_time, end_time) != OK) {
                return ERR;
            }
        }

        if (day == end_day) {
            break;
        }

        day = (day + 1) % DAYS_PER_WEEK;
    }

    return OK;
}

/*----------------------------------------------------------------------------
 *  check_time_values (int times_idx, int def_route)    - check time values of variable
 *
 *  Example:
 *
 *  Mo-Fr:08-18:0.032:Y Mo-Fr:18-08:0.025:Y Sa-Su:00-24:0.025:Y
 *----------------------------------------------------------------------------
 */
static int
check_time_values(int times_idx, BOOL def_route)
{
    char *week_range_str;
    char *hour_range_str;
    char *yes_no_str;
    char *start_week_day_str;
    char *end_week_day_str;
    char *start_hour_str;
    char *end_hour_str;
    char *p;
    char *pp;

    char *orig_val;
    BOOL complained = FALSE;

    p = time_values[times_idx];
    orig_val = strsave(p);
    init_local_time_table();

    while (p) {
        char *ppp;

        pp = strchr(p, ' ');
        if (pp != NULL) {
            *pp++ = '\0';

            while (*pp == ' ') {
                ++pp;
            }

            if (!*pp) {
                pp = NULL;
            }
        }

        ppp = strchr(p, ':');
        if (ppp != NULL) {
            week_range_str = p;
            *ppp++ = '\0';
        }
        else {
            log_error("Error: format error in variable %s\n",
                      time_variables[times_idx]);
            return ERR;
        }

        p = ppp;
        ppp = strchr(p, ':');
        if (ppp != NULL) {
            hour_range_str = p;
            *ppp++ = '\0';
        }
        else {
            log_error("Error: format error in variable %s\n",
                      time_variables[times_idx]);
            return ERR;
        }

        p = ppp;
        ppp = strchr(p, ':');
        if (ppp != NULL) {
            *ppp++ = '\0';
        }
        else {
            log_error("Error: format error in variable %s\n",
                      time_variables[times_idx]);
            return ERR;
        }

        yes_no_str = ppp;

        if (strlen(week_range_str) != 5 ||
                *(week_range_str + 2) != '-' ||
                strlen(hour_range_str) != 5 ||
                *(hour_range_str + 2) != '-') {
            log_error("Error: format error in variable %s\n",
                      time_variables[times_idx]);
            return ERR;
        }

        if (strcmp(yes_no_str, "Y") == 0 || strcmp(yes_no_str, "y") == 0 ||
               strcmp(yes_no_str, "N") == 0 || strcmp(yes_no_str, "n") == 0 ||
               strcmp(yes_no_str, "D") == 0 || strcmp(yes_no_str, "d") == 0) {
            BOOL lcr = strcmp(yes_no_str, "Y") == 0 || strcmp(yes_no_str, "y") == 0;
            *(week_range_str + 2) = '\0';
            start_week_day_str    = week_range_str;
            end_week_day_str      = week_range_str + 3;

            *(hour_range_str + 2) = '\0';
            start_hour_str        = hour_range_str;
            end_hour_str          = hour_range_str + 3;

            if (lcr && !def_route) {
                if (!complained) {
                    log_error("ignoring %s in %s='%s' "
                              "since the circuit has no "
                              "default route and therefore can't be used "
                              "for least cost routing\n",
                              yes_no_str,
                              time_variables[times_idx], orig_val);
                    complained = TRUE;
                }
                lcr = FALSE;
            }

            if (check_a_time_value(times_idx, lcr,
                                   start_week_day_str, end_week_day_str,
                                   start_hour_str, end_hour_str) != OK) {
                return ERR;
            }
        }
        else {
            log_error("Error: format error in variable %s\n",
                      time_variables[times_idx]);
            return ERR;
        }

        p = pp;
    }

    free(orig_val);
    return check_local_time_table(times_idx);
}


/*----------------------------------------------------------------------------
 *  check_time_variables (void)
 *----------------------------------------------------------------------------
 */
static int
check_time_variables(void)
{
    char varname[64];
    char const *p;
    int n_time_variables    = 0;
    BOOL pppoe_used          = FALSE;
    BOOL is_pptp             = FALSE;
    int n_isdn_circuits     = 0;
    int times_idx;
    int i;
    int j;
    int ret = OK;

    p = get_variable("OPT_PPPOE");
    if (p && strcmp(p, "yes") == 0) {
        pppoe_used = TRUE;
        ++n_time_variables;
    }
    else {
        p = get_variable("OPT_PPTP");
        if (p && strcmp (p, "yes") == 0) {
            pppoe_used  = TRUE;
            is_pptp     = TRUE;
            ++n_time_variables;
        }
    }

    p = get_variable ("OPT_ISDN");
    if (p && strcmp(p, "yes") == 0) {
        n_isdn_circuits = atoi(get_variable("ISDN_CIRC_N"));
        n_time_variables += n_isdn_circuits;
    }

    if (n_time_variables == 0) {
        return OK;
    }

    time_values    = (char **) malloc(n_time_variables * sizeof(char *));
    time_variables = (char **) malloc(n_time_variables * sizeof(char *));

    for (i = 0; i < n_isdn_circuits; ++i) {
        sprintf(varname, "ISDN_CIRC_%d_TIMES", i + 1);
        time_variables[i] = strsave(varname);
        p = get_variable(varname);
        time_values[i] = strsave(p ? p : "");
    }

    if (pppoe_used) {
        if (is_pptp) {
            time_variables[i] = strsave("PPTP_TIMES");
            p = get_variable("PPTP_TIMES");
            time_values[i] = strsave(p ? p : "");
        }
        else {
            time_variables[i] = strsave("PPPOE_TIMES");
            p = get_variable("PPPOE_TIMES");
            time_values[i] = strsave(p ? p : "");
        }
    }

    for (i = 0; i < DAYS_PER_WEEK; ++i) {
        for (j = 0; j < HOURS_PER_DAY; ++j) {
            time_table[i * HOURS_PER_DAY + j] = -1;
        }
    }

    for (times_idx = 0; times_idx < n_time_variables; ++times_idx) {
        BOOL def_route = TRUE;
        if (times_idx < n_isdn_circuits) {
            sprintf(varname, "ISDN_CIRC_%d_ROUTE", times_idx + 1);
            p = get_variable(varname);

            if (p && strcmp(p, "default") != 0 && strcmp(p, "0.0.0.0") != 0) {
                def_route = FALSE;              /* no lcr circuit, skip */
            }
        }

        if (check_time_values (times_idx, def_route) != OK) {
            ret = ERR;
        }
    }

    return ret;
}

static void
check_external_opt(char const *var, char const *var1)
{
    char *t;
    char const *p, *q;
    char *tmp = strsave(var);

    t = strchr(tmp, '%');
    if (t) {
        *t = '1';
    }

    p = get_variable_package(tmp);
    if (p) {
        q = get_variable_package (var1);
        if (q && strcmp(q, "BASE") != 0 && strcmp(q, "_fli4l") != 0 &&
                strcmp(p, "_fli4l") != 0&& strcmp(p, q) != 0) {
            log_info (LOG_ZIPLIST, "external reference to '%s' ('%s') in '%s' "
                      "(variable '%s)\n",
                      var1, q, p, var);
        }
    }

    free(tmp);
}

int
check_all_variables(void)
{
    int ret = OK;

    ARRAY_ITER(array_iter, check_array, c, struct check) {
        int optional;

        if (c->opt_var && ! c->var_n) {
            char const *opt_var_content;
            check_external_opt(c->name, c->opt_var);
            opt_var_content = get_variable (c->opt_var);

            /*
             * kristov: this is necessary for "deep" dependencies. Take e.g.
             * LPD_PARPORT_N --> OPT_LPD_PARPORT --> OPT_LPD: If OPT_LPD is
             * not set, its value is set to "no" due to its default value, but
             * OPT_LPD_PARPORT (if not set) is _not_ set to "no" in spite of
             * its default value, as OPT_LPD_PARPORT is not enabled (due to
             * OPT_LPD being "no"); see var.c:382 and following lines where the
             * default value is only used if the variable is enabled. So the
             * block containing the log_error call below will be triggered as
             * opt_var_content is NULL. To prevent this, we explicitly check
             * whether the enabling variable has a default value and do not
             * rely on var.c:382ff. to take care of this.
             */
            if (!opt_var_content) {
                struct check *opt_var_check = lookup_check_var(c->opt_var);
                opt_var_content = opt_var_check ? opt_var_check->defval : NULL;
            }

            if (!opt_var_content) {
                log_error("variable %s depends on %s, which is not defined\n",
                          c->name, c->opt_var);
                return ERR;
            }

            if (strcmp(opt_var_content, "no") == 0) {
                char const * content = get_variable(c->name);
                if (strncmp (c->name, "OPT_", 4) == 0 && /* optvar dependent ? */
                        content && strcmp(content, "yes") == 0) {
                    printf ("Warning: %s='%s' ignored, because %s='no'\n",
                            c->name, content, c->opt_var);
                    set_variable_content(c->name, "no");
                }
            }
        }

        if (c->flags.really_optional) {
            optional = CHECK_REALLY_OPTIONAL;
        }
        else if (c->flags.optional) {
            optional = CHECK_OPTIONAL;
        }
        else {
            optional = CHECK_NON_OPTIONAL;
        }

        if (c->var_n) {
            int type = c->flags.type_numeric ? TYPE_NUMERIC : TYPE_UNKNOWN;
            check_external_opt(c->name, c->var_n);
            if (chkvar_idx(c->name, c->var_n, c->opt_var,
                           c->regexp_name, c->regexp,
                           c->package, type, optional, c->flags.neg_opt) != OK) {
                ret = ERR;
            }
        }
        else {
            int type = c->flags.type_numeric ? TYPE_NUMERIC : TYPE_UNKNOWN;
            if (chkvar(c->name,
                       c->regexp_name, c->regexp,  c->package,
                       is_var_enabled(c->name, c->opt_var, c->flags.neg_opt),
                       type, optional, c->defval) != OK) {
                ret =  ERR;
            }
        }
    }

    if (check_time_variables() != OK) {
        return ERR;
    }

    return ret;
}

void
free_check_variables(void) {
    ARRAY_ITER(array_iter, check_array, p, struct check) {
        free(p->name);
        free(p->opt_var);
        free(p->var_n);
        free(p->regexp);
    }
}
