/*----------------------------------------------------------------------------
 *  parse.c - functions for parsing
 *
 *  Copyright (c) 2002-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: parse.c 41037 2015-08-22 20:11:58Z kristov $
 *----------------------------------------------------------------------------
 */

#include "libglob/src/config.h"
#include <glob.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include "libcrypt/libcrypt.h"
#include "var.h"
#include "regex.h"
#include "log.h"
#include "str.h"
#include "check.h"
#include "parse.h"
#include "file.h"
#include "tree_struct.h"

#define ENTRY_OPT_ARRAY_SIZE    4

/**
 * The maxium number of passes made while executing extended check script.
 */
#define MAX_EXT_CHECK_PASS_NUMBER 99

static BOOL exec_failed = FALSE;
static BOOL final_pass = FALSE;
static BOOL package_provider_added = FALSE;
static char *parse_current_package;
static char *parse_current_file;
static char fli4l_version_tag[20];
static char var_buf[VAR_SIZE + 1];

extern FILE *yyin;
void yyrestart(FILE *fp);

static void
parse_set_current(char const *pkg, char const *file)
{
    parse_current_package = strsave(pkg);
    parse_current_file = strsave(file);
}

static BOOL
parse_get_error(void)
{
    return exec_failed;
}

static void
parse_set_error(void)
{
    exec_failed = TRUE;
}

char const *
parse_get_variable(char const *name, char const *package, int line)
{
    char const *content = get_variable(convert_to_upper(name));
    if (!content) {
        log_error("unknown variable '%s' in %s, line %d\n", name, package, line);
        parse_set_error();
        return NULL;
    }
    else {
        return content;
    }
}

static char const *
parse_get_variable_package(char const *name, char const *package, int line)
{
    char const *content = get_variable_package(convert_to_upper(name));
    if (!content) {
        log_error("unknown variable '%s' in %s, line %d\n", name, package, line);
        parse_set_error();
        return NULL;
    }
    else {
        return content;
    }
}

static char const *
parse_get_variable_comment(char const *name, char const *package, int line)
{
    char const *content = get_variable_comment(convert_to_upper(name));
    if (!content) {
        log_error("unknown variable '%s' in %s, line %d\n", name, package, line);
        parse_set_error();
        return NULL;
    }
    else {
        return content;
    }
}

array_t *
parse_entry_options(char const *opts)
{
    array_t *result = init_array(ENTRY_OPT_ARRAY_SIZE);
    char const *p = opts;
    char const *q;

    while ((q = strchr(p, '=')) != NULL) {
        int const name_len = q - p;
        struct opt_t *opt = (struct opt_t *) safe_malloc(sizeof(struct opt_t));
        append_new_elem(result, opt);

        /* read name */
        opt->name = (char *) safe_malloc(name_len + 1);
        strncpy(opt->name, p, name_len);
        opt->name[name_len] = '\0';

        /* read value */
        p = q + 1;
        q = strchr(p, ' ');
        if (q != NULL) {
            int const value_len = q - p;
            opt->value = (char *) safe_malloc(value_len + 1);
            strncpy(opt->value, p, value_len);
            opt->value[value_len] = '\0';
        } else {
            opt->value = strsave(p);
        }
        p += strlen(opt->value);

        /* skip trailing spaces */
        while (*p && isspace((int) *p))
            ++p;
    }

    return result;
}

#define PMAX_MATCH 20
BOOL
str_strmatch(char const *value, char const *expr, char const *package, int line)
{
    int res;
    regmatch_t match[PMAX_MATCH];
    char *match_var_n = "MATCH_N";
    char *match_set_var = "MATCH_%";
    char match_var[10];
    char num[10];

    if (!check_var_defined(match_set_var)) {
        check_add_weak_declaration("internal", match_set_var, "0",
                                   0, package, line, LOG_TREE_EXEC);
    }
    else if (set_variable_content(match_var_n, "0") != OK) {
        log_error("unexpected error while setting value of %s\n", match_var_n);
        parse_set_error();
        return FALSE;
    }

    res = regexp_user(value, expr, PMAX_MATCH, match, 0, package, line);
    log_info(LOG_TREE_EXEC, "checking regular expression '%s' against value '%s', "
             "res = %d\n", expr, value, res);

    if (res == 0) {
        int i;
        for (i = 1; i < PMAX_MATCH && match[i].rm_so != -1; ++i) {
            int const len = match[i].rm_eo - match[i].rm_so;
            char *substr = (char *) safe_malloc(len + 1);
            strncpy(substr, value + match[i].rm_so, len);
            substr[len] = '\0';

            sprintf(match_var, "MATCH_%d", i);
            sprintf(num, "%d", i);
            var_add_weak_declaration("internal variable - match_%",
                                     match_var, substr,
                                     num, TYPE_UNKNOWN,
                                     package, line, LOG_TREE_EXEC);
            free(substr);
        }

        set_variable_content(match_var_n, num);
        return TRUE;
    }
    else if (res == REG_NOMATCH) {
        return FALSE;
    }
    else {
        log_error("wrong regular expression '%s' in %s, line %d\n",
                  expr, package, line);
        parse_set_error();
        return FALSE;
    }
}

int
val_numcmp(char const *val, char const *num, char const *file, int line)
{
    unsigned long x, y;

    log_info(LOG_TREE_EXEC, "numcmp ('%s' rel %s') ?\n", val, num);

    x = convert_to_long(val, file, line);
    y = convert_to_long(num, file, line);

    if (x == y) {
        return CMP_EQUAL;
    }
    else if (x < y) {
        return CMP_LESS;
    }
    else {
        return CMP_GREATER;
    }
}

static int
get_ver_num(char const *ver, char const *file, int line)
{
    int ver_major, ver_minor;
    unsigned long sub = 0;

    if (!ver) {
        fatal_exit("%s %d: Null pointer passed to get_ver_num\n",
                   __FILE__, __LINE__);
    }
    if (strlen(ver) < 3) {
        fatal_exit("%s %d: Short version passed to get_ver_num\n",
                   __FILE__, __LINE__);
    }

    /* kristov TODO: this algorithm only works with major and minor < 10! */
    ver_major = ver[0] - '0';
    ver_minor = ver[2] - '0';

    if (ver[3] != '\0') {
        sub = convert_to_long(&ver[4], file, line);
    }
    return sub + ver_minor * 1000 + ver_major * 10000 ;
}

int
val_vercmp(char const *val, char const *version, char const *file, int line)
{
    int x, y;

    log_info(LOG_TREE_EXEC, "vercmp ('%s'=='%s') ?\n", val, version);

    x = get_ver_num(val, file, line);
    y = get_ver_num(version, file, line);

    if (x == y) {
        return CMP_EQUAL;
    }
    else if (x < y) {
        return CMP_LESS;
    }
    else {
        return CMP_GREATER;
    }
}

char *
parse_rewrite_string(char const *msg, char const *package, int line)
{
#define MSG_SIZE 6144
    char msg_buf[MSG_SIZE];
    char *p=msg_buf;

    while (*msg) {
        switch (*msg) {
        case '$':
        case '%':
        case '@':
            if (isalpha((int) *(msg + 1)) || *(msg + 1) == '{') {
                /* get variable name */
                char *var = var_buf;
                char prefix = *msg++;
                char const *value = NULL;

                if (*msg == '{') {
                    ++msg;
                    while ((isalnum((int) *msg) || *msg == '_') && *msg != '}') {
                        *var++ = *msg++;
                    }
                    ++msg;
                }
                else {
                    while (isalnum((int) *msg) || *msg == '_') {
                        *var++ = *msg++;
                    }
                }
                *var = '\0';
                if (strlen(var_buf) >= VAR_SIZE) {
                    fatal_exit("%s %d: buffer overflow,"
                               " variable name too long\n",
                               __FILE__, __LINE__);
                }

                switch (prefix) {
                case '$':
                    value = parse_get_variable(var_buf, package, line);
                    break;
                case '%':
                    value = parse_get_variable_package(var_buf, package, line);
                    break;
                case '@':
                    value = parse_get_variable_comment(var_buf, package, line);
                    break;
                }
                /* append to msg text */
                *p = '\0';
                if (value) {
                    strcat(msg_buf, value);
                    p += strlen(value);
                }

                break;
            }
            else {
                switch (*(msg + 1)) {
                case '$':
                case '%':
                case '@':
                    ++msg;
                }
                /* fall through */
            }
        default:
            *p++ = *msg++;
            break;
        }
    }
    *p = '\0';
    if (strlen(msg_buf) >= MSG_SIZE) {
        fatal_exit("%s %d: buffer overflow, msg too long\n",
                   __FILE__, __LINE__);
    }
    return strsave(msg_buf);
}

void
parse_warning(char const *warning, char const *package, int line)
{
    char *p;
    if (!final_pass) {
        return;
    }
    p = parse_rewrite_string(warning, package, line);
    log_error("Warning: %s\n", p);
    free(p);

}
void
parse_error(char const *warning, char const *package, int line)
{
    char *p;
    if (!final_pass) {
        return;
    }
    p = parse_rewrite_string(warning, package, line);
    log_error("Error: %s\n", p);
    free(p);
    parse_set_error();
}

void
parse_fatal_error(char const *warning, char const *package, int line)
{
    char *p = parse_rewrite_string(warning, package, line);
    log_error("Fatal Error: %s\n", p);
    free(p);
    fatal_exit ("fatal error, aborting...\n");;
}

void
parse_add_to_opt(char const *file, char const *options, char const *package, int line)
{
    if (!final_pass) {
        return;
    }
    if (add_to_zip_list("BASE", "yes", file,
                        options ? parse_entry_options(parse_rewrite_string(options, package, line)) : NULL,
                        package, line) != OK) {
        parse_set_error();
    }
}

void
parse_crypt(char const *id, char const *package, int line)
{
    char const *key;

    if (!final_pass) {
        return;
    }

    key = parse_get_variable(id, package, line);
    if (key) {
        char *crypted_passwd;
        char salt[8 + 1];
        unsigned i;
        static char const rndChar[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789./";

        for (i = 0; i < sizeof(salt) - 1; ++i) {
            salt[i] = rndChar[rand() % 64];
        }
        salt[i] = '\0';
        crypted_passwd = __sha256_crypt(key, salt);
        set_variable_content(id, crypted_passwd);
    }
}

void
parse_stat(char const* file, char const *id, char const *package, int line)
{
    char *p;
    struct stat stat_buf;

    p = parse_rewrite_string(file, package, line);
    log_info (LOG_TREE_EXEC, "executing stat on %s, results go to %s_*\n",
              p, id);

    strcpy(var_buf, id);
    if (stat(p, &stat_buf) != 0) {
        strcat(var_buf, "_RES");
        var_add_weak_declaration(package, var_buf,
                                 strsave(strerror(errno)), NULL,
                                 TYPE_UNKNOWN, package, line, LOG_TREE_EXEC);
    }
    else {
        char buf[128];
        char *end = var_buf + strlen(var_buf);
        strcpy(end, "_RES");
        var_add_weak_declaration(package, var_buf,
                                 strsave("OK"), NULL, TYPE_UNKNOWN,
                                 package, line, LOG_TREE_EXEC);
        sprintf(buf, "%ld", (unsigned long) stat_buf.st_size);
        strcpy(end, "_SIZE");
        var_add_weak_declaration(package, var_buf,
                                 strsave(buf), NULL, TYPE_NUMERIC,
                                 package, line, LOG_TREE_EXEC);
    }
    free(p);
}

void
parse_fgrep(char const *file, char const *_expr, char const *package, int line)
{
    char *p;
    char *expr;
    int res;
    char *match_var_n = "FGREP_MATCH_N";
    char *match_set_var = "FGREP_MATCH_%";
    char match_var[25];
    char num[10];
    FILE *fh;
    glob_t glob_res;
    int matches = 0;

    p = parse_rewrite_string(file, package, line);
    expr = parse_rewrite_string(_expr, package, line);

    log_info(LOG_TREE_EXEC, "checking regular expression '%s' against file '%s'\n", expr, p);

    if (!check_var_defined(match_set_var)) {
        check_add_weak_declaration("internal", match_set_var, "0",
                                   0, package, line, LOG_TREE_EXEC);
    }
    else if (set_variable_content (match_var_n, "0") != OK) {
        fatal_exit("(%s:%d) unexpected error while setting value of %s\n",
                   __FILE__, __LINE__, match_var_n);
    }

    if (glob(p, 0, NULL, &glob_res) == 0) {
        size_t i;
        for (i = 0; i < glob_res.gl_pathc; ++i) {
            char *found = glob_res.gl_pathv[i];
            fh = fopen(found, "r");
            if (fh != NULL) {
                char buff[1024];

                while(fgets(buff, sizeof(buff), fh) != NULL) {
                    regmatch_t match[PMAX_MATCH];
                    int len;

                    len = strlen(buff);
                    if (len > 1) {
                         --len;
                    }
                    if (buff[len] == '\n' || buff[len] == '\r') {
                        buff[len] = '\0';
                    }
                    if (len > 1) {
                        --len;
                    }
                    if (buff[len] == '\n' || buff[len] == '\r') {
                        buff[len] = '\0';
                    }

                    res = regexp_user(buff, expr, PMAX_MATCH, match, 0, package, line);
                    if (res == 0) {
                        int i;

                        for (i=0; i<PMAX_MATCH && match[i].rm_so != -1; i++)
                          {
                            char * end = buff+match[i].rm_eo;
                            char tmp = *end;
                            *end = '\0';

                            matches++;
                            sprintf (match_var, "FGREP_MATCH_%d", matches);
                            sprintf (num, "%d", matches);
                            var_add_weak_declaration ("internal variable - fgrep_match_%",
                                                      match_var, buff+match[i].rm_so,
                                                      num, TYPE_UNKNOWN,
                                                      package, line, LOG_TREE_EXEC);
                            *end = tmp;
                          }
                        set_variable_content (match_var_n, num);
                    }
                    else if (res != REG_NOMATCH) {
                        fatal_exit("wrong regular expression in fgrep '%s' in %s, line %d, terminating\n",
                                   expr, package, line);
                    }
                }

                fclose(fh);
            }
        }
    }

    free(p);
    free(expr);
}

void
parse_provides(char const *id, char const *version, char const *package, int line)
{
    char const *p;
    char internal_id[128];

    strcpy(internal_id, "provides_");
    strcat(internal_id, convert_to_upper(id));
    p = get_variable_package(internal_id);
    if (p) {
        char const *q = get_variable_package(internal_id);
        if (strcmp(package, q) != 0) {
            fatal_exit("Conflicting packages, package %s already"
                       " provides '%s' with version '%s'\n",
                       package, id, get_variable (internal_id));
        }
        return;
    }
    else {
        log_info(LOG_TREE_EXEC, "package %s provides %s version %s\n",
                 package, id, version);
        var_add_weak_declaration(package, internal_id, version, version,
                                 TYPE_NUMERIC, package, line,
                                 LOG_TREE_EXEC);
        mark_var_generated(internal_id);
        mark_var_provide(internal_id);

        package_provider_added = TRUE;
    }
}

void
parse_depends(char const *id, elem_t *version, char const *package,
              char const *file, int line)
{
    char const *p;
    char internal_id[128];

    if (!final_pass) {
        return;
    }

    strcpy(internal_id, "provides_");
    strcat(internal_id, convert_to_upper(id));
    p = get_variable(internal_id);
    if (p) {
        size_t const p_len = strlen(p);
        BOOL err = version != NULL;
        char *verstr = NULL;

        while (version) {
            char const *const value = version->ARG[0]->VAL;
            size_t const value_len = strlen(value);

            if (!verstr) {
                verstr = strsave(value);
            }
            else {
                verstr = strcat_save(verstr, "/");
                verstr = strcat_save(verstr, value);
            }

            if (value_len > p_len) {
                continue;
            }
            if (strncmp(p, value, value_len) == 0) {
                if (p[value_len] == '\0' || p[value_len] == '.') {
                    err = FALSE;
                    break;
                }
            }

            version = version->ARG[1];
        }

        if (err) {
            fatal_exit("(%s:%d) Version mismatch, package '%s' depends on "
                       "'%s', version %s, but '%s' has version %s\n",
                       file, line, package, id, verstr, id, p);
        }
    }
    else {
        fatal_exit("(%s:%d) Package missing, %s depends on %s, "
                   "which is either not present or not enabled\n",
                   file, line, package, id);
    }
}

/**
 * Parses an extended check script.
 *
 * @param name
 *  The file name of the extended check script without path and extension.
 * @param check_dir
 *  The directory containing extended check scripts.
 * @param ext
 *  The file extension used for extended check scripts.
 * @return
 *  OK on success, ERR otherwise.
 */
static int
parse_check_file(char const *name, char const *check_dir, char const *ext)
{
    char *file;
    FILE *fp;
    int ret;

    if (!get_file_name(check_dir, name, ext, &file)) {
        log_info(LOG_VERBOSE, "no extended check file for package %s\n",
                 name);
        free(file);
        return OK;
    }

    fp = fopen(file, "r");
    if (!fp) {
        log_error("Error opening extended check file %s: %s\n",
                  file, strerror (errno));
        free(file);
        return ERR;
    }

    log_info(LOG_TREE_EXEC|LOG_INFO,
             "reading extended check file %s\n", file);

    yyrestart(fp);
    yyline = 1;
    inc_log_indent_level();
    parse_set_current(name, file);
    ret = yyparse();
    dec_log_indent_level();
    fclose (fp);

    if (ret != OK) {
        log_error("error while parsing check file %s\n", file);
        free(file);
        return ERR;
    }

    free(file);
    return OK;
}

/**
 * Parses all extended check scripts and build parse tree.
 *
 * @param check_dir
 *  The directory containing extended check scripts.
 * @param ext
 *  The file extension used for extended check scripts.
 * @return
 *  OK on success, ERR otherwise.
 */
static int
parse_all_extended_checks(char const *check_dir, char const *ext, char **packages, int n_packages)
{
    int i;

    log_info(LOG_INFO, "parsing extended check files\n\n");
    inc_log_indent_level();

    /* prepare version string, we can't have it in the source because
     * it would be replaced by th svn version tag */
    strcpy(fli4l_version_tag, "__FLI4L");
    strcat(fli4l_version_tag, "VER__");

    for (i = 0; i < n_packages; ++i) {
        int res = parse_check_file(packages[i], check_dir, ext);
        if (res != OK) {
            return res;
        }
    }

    dec_log_indent_level();
    log_info(LOG_INFO, "\nfinished parsing extended check files\n");
    return OK;
}

char const *
parse_get_current_package(void)
{
    return parse_current_package;
}

char const *
parse_get_current_file(void)
{
    return parse_current_file;
}

int
execute_all_extended_checks(char const *check_dir, char const *ext, char **packages, int n_packages)
{
    int error;
    int parse_pass;

    if ((error = parse_all_extended_checks(check_dir, ext, packages, n_packages)) != OK) {
        return error;
    }

    parse_pass = 1;
    while (TRUE)
    {
        if (final_pass) {
            log_info(LOG_INFO, "executing extended checks, final pass\n\n");
        }
        else {
            log_info(LOG_INFO, "executing extended checks, pass %d\n\n",
                     parse_pass);
        }

        inc_log_indent_level();
        walk_tree();
        if (parse_get_error()) {
            return ERR;
        }
        dec_log_indent_level();

        if (final_pass) {
            if (package_provider_added) {
                log_error("package provided during final check pass, "
                          "bad extended check file\n");
                return ERR;
            }
            log_info(LOG_INFO, "\nfinished final pass\n");
            break;
        }
        else {
            log_info (LOG_INFO, "\nfinished pass %d\n", parse_pass);
        }

        if (package_provider_added) {
            package_provider_added = FALSE;
        }
        else {
            final_pass = TRUE;
        }

        ++parse_pass;
        if (parse_pass > MAX_EXT_CHECK_PASS_NUMBER) {
            log_error("bad extended check files, aborting after pass %d", MAX_EXT_CHECK_PASS_NUMBER);
            return ERR;
        }
    }

    return OK;
}
