    /*----------------------------------------------------------------------------
 *  regex.c - functions for processing regular expressions
 *
 *  Copyright (c) 2002-2005 - Jean Wolter
 *  Copyright (c) 2005-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Creation:    2002-03-01  jw5
 *  Last Update: $Id: regex.c 34688 2014-11-07 12:56:15Z florian $
 *----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <regex.h>
#include <stdarg.h>
#include <ctype.h>

#include <dirent.h>

#include "regex.h"
#include "var.h"
#include "check.h"
#include "log.h"
#include "str.h"
#include "cfg.h"
#include "array.h"

/* #define DEBUG_REGEX */

#define REGEXP_ARRAY_SIZE       256

struct exp_cache_t {
    char *name;
    char *expr;
    char *error_msg;
    char *file;
    int line;
    int type;
    BOOL extended;
    BOOL compiled;
    char *complete_expr;
    regex_t comp_expr;
};

#define SIZE_ERR_BUFFER (128 + 1)
#define SIZE_EXPR_BUFFER (20480)

static char expr_buffer[SIZE_EXPR_BUFFER + 1];
static char err_buf[SIZE_ERR_BUFFER];
static regex_t reg;
static array_t *regexp_array;
static char const *user_regexp = "user defined regular expression";

#define REGEXP_WARN_PREFIX "WARN_"
static int
has_warn_prefix (const char *name)
{
    return name &&
        *name == 'W' &&
        strlen(name) >= sizeof(REGEXP_WARN_PREFIX) &&
        strncmp(name, REGEXP_WARN_PREFIX, sizeof(REGEXP_WARN_PREFIX) - 1) == 0;
}

static struct exp_cache_t *
find_expr(char const *name)
{
    DECLARE_ARRAY_ITER(array_iter, p, struct exp_cache_t);

    if ((p = get_first_elem(regexp_array)) == NULL) {
        return NULL;
    }

    if (has_warn_prefix(name)) {
        name += sizeof(REGEXP_WARN_PREFIX) - 1;
    }

    ARRAY_ITER_LOOP(array_iter, regexp_array, p) {
        if (strcmp (name, p->name) == 0) {
            return p;
        }
    }

    return NULL;
}

static void
copy_substring(char **dest, char const *src, int size,
               char const *expr_file, int line)
{
    if (!size) {
        size = strlen (src);
    }

    if (size >= SIZE_EXPR_BUFFER - (*dest - expr_buffer)) {
        fatal_exit ("expression too long in %s line %d (max len after "
                    "sub expression replacement: %d)\n",
                    expr_file, line, SIZE_EXPR_BUFFER);
    }

    strncpy(*dest, src, size);
    *dest += size;
    **dest = '\0';
}

static void
compile_expression(struct exp_cache_t *e, char const *expr, BOOL final,
                   BOOL modify_expr, char const *expr_file, int line)
{
    int res = 0;
    regmatch_t pmatch[4];
    char *sub_name;
    char *subs = NULL;

    char expr_src[SIZE_EXPR_BUFFER + 1];
    struct exp_cache_t *p;
    char *dst;
    char *src;

    strcpy(expr_src, expr);

    while (res == 0) {
        dst = expr_buffer;
        src = expr_src;

        while (*src) {
            /* look for subexpressions (RE:.*) */
            res = regexec(&reg, src, 2, pmatch, 0);
            if (res != 0) {
                /* no sub expressions */
                copy_substring(&dst, src, 0, expr_file, line);
                break;
            }

            /* copy substring including '(' and skip substring
             * including '(RE:' */
            copy_substring(&dst, src, pmatch[1].rm_so + 1, expr_file, line);
            sub_name = src + pmatch[1].rm_so + 4;
            src[pmatch[1].rm_eo - 1] = '\0';
            src += pmatch[1].rm_eo;

            if (!subs) {
                subs = strsave(e->name);
            }

            subs = strcat_save(subs, " -> ");
            subs = strcat_save(subs, sub_name);

            if (strcmp(sub_name, e->name) == 0) {
                fatal_exit("recursive definition of %s (%s) detected in %s line %d'\n",
                           sub_name, subs, expr_file, line);
            }

            p = find_expr(sub_name);
            if (!p) {
                fatal_exit("unknown subexpression %s in %s line %d'\n",
                           sub_name, expr_file, line);
            }

            copy_substring(&dst, p->expr, 0, expr_file, line);
            copy_substring(&dst, ")", 0, expr_file, line);
        }

        /* look for subexpressions (RE:.*) */
        res = regexec(&reg, expr_buffer, 2, pmatch, 0);
        if (res == 0) {
            /* there are still sub expressions */
            strcpy(expr_src, expr_buffer);
        }
    }

    strcpy (expr_src, expr_buffer);
    dst = expr_buffer;
    src = expr_src;

    /* add leading '^(' and trailing ')$' */
    if (modify_expr) {
        copy_substring(&dst, "^(", 0, expr_file, line);
    }
    copy_substring(&dst, src, 0, expr_file, line);
    if (modify_expr) {
        copy_substring(&dst, ")$", 0, expr_file, line);
    }

    /* compile regexp for regexps */
    res = regcomp(&e->comp_expr, expr_buffer, REG_EXTENDED | REG_NEWLINE);
    if (res != 0) {
        regerror(res, &reg, err_buf, SIZE_ERR_BUFFER);
        fatal_exit("regex error %d (%s) for expr '%s' in %s line %d\n",
                   res, err_buf, expr_buffer, expr_file, line);
    }

    if (final) {
        e->complete_expr = strsave(expr_buffer);
        if (e->name != user_regexp) {
            log_info (LOG_REGEXP, "adding %s='%s' ('%s')\n",
                      e->name, expr, e->complete_expr);
        }
    }
    else {
        regfree(&e->comp_expr);
    }

    free(subs);
}

int
regexp_exists(char const *name)
{
    return find_expr(name) != NULL;
}

int
regexp_user(char const *value, char const *expr, size_t nmatch,
            regmatch_t *pmatch, BOOL modify_expr, char const *file, int line)
{
    struct exp_cache_t e;
    int res;

    e.name = strsave(user_regexp);
    compile_expression(&e, expr, TRUE, modify_expr, file, line);
    log_info(LOG_REGEXP, "checking '%s' against regexp '%s' ('%s')\n",
             value, expr, e.complete_expr);

    res = regexec (&e.comp_expr, value, nmatch, pmatch, 0);
    if (res != 0) {
        regerror(res, &e.comp_expr, err_buf, SIZE_ERR_BUFFER);
        log_info(LOG_REGEXP|LOG_VERBOSE, "regex error %d (%s) for value '%s' "
                 "and regexp '%s' ('%s')\n",
                 res, err_buf, value, expr, e.complete_expr);
    }

    regfree(&e.comp_expr);
    return res;
}

int
regexp_chkvar(char const *name, char const *value,
              char const *regexp_name, char const *user_regexp)
{
    int res;
    char *err_msg;

    if (!regexp_name) {
        if (strlen(user_regexp) > SIZE_EXPR_BUFFER - 4) {
            log_error("user supplied regular expression for variable"
                      " %s too long (max len: %d)\n",
                      name, SIZE_EXPR_BUFFER - 4);
            return ERR;
        }
        err_msg = "user supplied regular expression";
        res = regexp_user(value, user_regexp, 0, NULL, TRUE, __FILE__, __LINE__);
    }
    else {
        struct exp_cache_t *p = find_expr(regexp_name);
        if (!p) {
            log_error("unknown regular expression %s\n", regexp_name);
            return ERR;
        }

        if (!p->compiled) {
            compile_expression (p, p->expr, TRUE, TRUE, p->file, p->line);
            p->compiled = TRUE;
        }

        err_msg = p->error_msg;
        res = regexec(&p->comp_expr, value, 0, NULL, 0);
        if (res != 0) {
            regerror(res, &p->comp_expr, err_buf, SIZE_ERR_BUFFER);
            log_info(LOG_VERBOSE, "regex error %d (%s) for value '%s' "
                     "and regexp '%s'\n",
                     res, err_buf, value, p->expr);
        }
    }

    if (res == 0) {
        return OK;
    }
    else if (res == REG_NOMATCH) {
        char *config_file;
        int line;

        get_variable_src(name, &config_file, &line);
        if (has_warn_prefix(regexp_name)) {
            log_error("(%s:%d) problematical value of variable %s: '%s' (%s)\n",
                      config_file, line, name, value, err_msg);
            return OK;
        }
        else {
            log_error("(%s:%d) wrong value of variable %s: '%s' (%s)\n",
                      config_file, line, name, value, err_msg);
        }
    }

    return ERR;
}

int
regexp_find_type(char const *regexp_name)
{
    struct exp_cache_t *p = find_expr(regexp_name);
    return p ? p->type : -1;
}

int
extend_expression(char const *name, char const *expr, char const *error_msg,
                  char const *expr_file, int line)
{
    char *tmp;
    struct exp_cache_t *p;

    /* name already existing ? */
    if ((p = find_expr (name)) == NULL) {
        log_error ("%s: line %d: tries to extend expression '%s' which "
                   "doesn't exist\n",
                   expr_file, line, name);
        return ERR;
    }

    /* create new regexp and compile it */
    tmp = (char *) safe_malloc(strlen(expr) + strlen(p->expr) + 6);
    strcpy(tmp, "(");
    strcat(tmp, p->expr);
    strcat(tmp, ")|(");
    strcat(tmp, expr);
    strcat(tmp, ")");
    compile_expression(p, tmp, FALSE, FALSE, expr_file, line);
    free(p->expr);
    p->expr = tmp;

    /* extend error message */
    tmp = (char *) safe_malloc(strlen(p->error_msg) + strlen(error_msg) + 1);
    strcpy(tmp, p->error_msg);
    strcat(tmp, error_msg);
    free(p->error_msg);
    p->error_msg = strsave_ws(tmp);
    free(tmp);
    return OK;
}

int
check_for_dangling_extensions(void)
{
    ARRAY_ITER(array_iter, regexp_array, p, struct exp_cache_t) {
        if (p->extended) {
            log_info(LOG_INFO|LOG_REGEXP,
                     "Tried to extend undefined regular expression '%s'.\n",
                      p->name);
            *p->name = '#';
            /* kristov TODO: check whether ERR should be returned */
        }
    }
    return OK;
}

void
compile_all_expressions(void)
{
    DECLARE_ARRAY_ITER(array_iter, p, struct exp_cache_t);

    if ((p = get_first_elem(regexp_array)) == NULL) {
        fatal_exit ("No regular expression defined, unable to check configuration.\n");
    }

    ARRAY_ITER_LOOP(array_iter, regexp_array, p) {
        if (p->expr) {
            compile_expression (p, p->expr, TRUE, TRUE, __FILE__, __LINE__);
        }
    }
}

void
regexp_init(void)
{
    int res;
    static const char *regexpr = "(\\(RE:[^\\)]*\\))";

    /* compile regexp for regexps */
    res = regcomp(&reg, regexpr, REG_EXTENDED | REG_NEWLINE);
    if (res != 0) {
        regerror(res, &reg, err_buf, SIZE_ERR_BUFFER);
        fatal_exit("regex error %d (%s) in %s:%d\n",
                   res, err_buf, __FILE__, __LINE__);
    }
    regexp_array = init_array(REGEXP_ARRAY_SIZE);
}

int
add_expression(char const *name, char const *expr, char const *error_msg,
               BOOL extend_expr, char const *expr_file, int line)
{
    char *saved_expr = NULL;
    char *saved_error_msg = NULL;

    struct exp_cache_t *p;

    /* name already existing ? */
    if ((p = find_expr(name)) != NULL) {
        if (extend_expr) {
            return extend_expression(name, expr, error_msg, expr_file, line);
        }
        else if (p->extended) {
            saved_expr = p->expr;
            saved_error_msg = p->error_msg;
            p->extended = FALSE;
            p->expr = p->error_msg = NULL;
        }
        else {
            log_error("%s: line %d: redefinition of expression %s,"
                      " already defined in %s:%d\n",
                      expr_file, line, p->name, p->file, p->line);
            return ERR;
        }
    }
    else {
        p = (struct exp_cache_t *) safe_malloc(sizeof(struct exp_cache_t));
        memset(p, 0, sizeof(*p));
        p->extended = extend_expr;
        append_new_elem(regexp_array, p);
    }

    /* add name, error msg and regexp */
    p->name = strsave(name);
    p->expr = strsave(expr);
    p->error_msg = strsave_ws(error_msg);
    p->file = strsave(expr_file);
    p->line = line;
    p->compiled = FALSE;
    p->complete_expr = NULL;

    if (strncmp(name, "NUM", 3) == 0) {
        log_info(LOG_VERBOSE|LOG_TREE_EXEC, "found numeric mode '%s'\n", name);
        p->type = TYPE_NUMERIC;
    }
    else {
        p->type = TYPE_UNKNOWN;
    }
    compile_expression(p, p->expr, FALSE, FALSE, expr_file, line);

    if (saved_expr) {
        int ret = extend_expression(name, saved_expr, saved_error_msg,
                                    expr_file, line);
        free(saved_expr);
        free(saved_error_msg);
        p->extended = FALSE;
        return ret;
    }

    return OK;
}

/*----------------------------------------------------------------------------
 * read file with regular expressions
 *----------------------------------------------------------------------------
 */
#define EXPR_NAME       0
#define EXPR_VAL        2
#define EXPR_ERRMSG     4

int
regexp_read_file(char const *expr_file)
{
    int result = OK;
    int ret;

    struct token_t tokens[6] = {
        TOKEN(CFG_ID|CFG_OPT_ID|CFG_DEP_ID),
        TOKEN(CFG_EQUAL),
        TOKEN(CFG_STRING),
        TOKEN(CFG_COLON),
        TOKEN(CFG_STRING|CFG_ML_STRING),
        TOKEN(CFG_NONE),
    };

    cfg_fopen(expr_file);
    while ((ret = get_config_tokens (tokens, 1)) != CFG_EOF) {
        char *name;
        BOOL expand;

        if (ret == CFG_ERROR) {
            result = ERR;
            continue;
        }

        name = tokens[EXPR_NAME].text;
        expand = tokens[EXPR_NAME].token != CFG_ID;

        if (tokens[EXPR_NAME].token == CFG_DEP_ID) {
            char *dep;
            char *expr;
            char const *res;
            int len;

            if (*name != '+') {
                log_error("%s:%d: %s: fatal scanner error: CFG_DEP_ID does not start with '+'\n",
                           expr_file, tokens[EXPR_NAME].line, name);
                result = ERR;
                continue;
            }
            ++name;

            len = strlen(name);
            dep = strchr(name, '(');
            if (dep && name[len - 1] != ')') {
                log_error("%s:%d: %s: fatal scanner error: CFG_DEP_ID does not end with ')'\n",
                           expr_file, tokens[EXPR_NAME].line, name);
                result = ERR;
                continue;
            }

            name[len - 1] = '\0';
            *dep++ = '\0';

            expr = strstr(dep, "=~'");
            if (expr) {
                *expr = 0;
                expr += 3;
                if (name[len - 2] != '\'') {
                    log_error("%s:%d: %s: fatal scanner error: CFG_DEP_ID does not end with \"'\"\n",
                               expr_file, tokens[EXPR_NAME].line, name);
                    result = ERR;
                    continue;
                }

                name[len - 2] = '\0';
            }
            else {
                expr = "^yes$";
            }

            if ((res = get_variable(dep)) == NULL) {
                log_info(LOG_EXP, "%s:%d: %s depends on undefined variable %s\n",
                         expr_file, tokens[EXPR_NAME].line, name, dep);
#if 0 /* kristov: does not work as NET_DRV depends on KERNEL_VERSION which is undefined when using mkfli4l for checking mkfli4l.conf */
                result = ERR;
#endif
                continue;
            }

            if (regexp_user(res, expr, 0, NULL, FALSE,
                            expr_file, tokens[EXPR_NAME].line) != 0) {
                log_info(LOG_EXP, "%s:%d: ignoring expansion of %s - '%s' =~ '%s'\n",
                         expr_file, tokens[EXPR_NAME].line, name, res, expr);
                continue;
            }
        }

        if (expand) {
            log_info(LOG_EXP, "%s:%d: expanding definition of %s\n",
                     expr_file, tokens[EXPR_NAME].line, name);
        }

        if (add_expression(name,
                           tokens[EXPR_VAL].text,
                           tokens[EXPR_ERRMSG].text,
                           expand,
                           expr_file, tokens[EXPR_NAME].line) != OK) {
            result = ERR;
        }
    }

    cfg_fclose ();

#if 0
    compile_all_expressions();
#endif

    return result;
}

BOOL
regexp_get_expr(char const *name, regex_t **preg, char **err_msg)
{
    struct exp_cache_t *e = find_expr(name);
    if (e) {
        if (!e->compiled) {
            compile_expression(e, e->expr, TRUE, TRUE, e->file, e->line);
            e->compiled = TRUE;
        }
        *preg = &e->comp_expr;
        *err_msg = e->error_msg;
    }

    return e != NULL;
}
