/*----------------------------------------------------------------------------
 *  set_var.c - functions for processing set-like variables
 *
 *  Copyright (c) 2002-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: set_var.c 34688 2014-11-07 12:56:15Z florian $
 *----------------------------------------------------------------------------
 */

#include "libmkfli4l/check.h"
#include "libmkfli4l/var.h"
#include "libmkfli4l/log.h"
#include "libmkfli4l/str.h"
#include "libmkfli4l/array.h"
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define SET_VAR_ARRAY_SIZE      128

#define _free(x) do { free(x) ; (x) = 0; } while (0)
#define VAR_N_STACK_SIZE 32

struct var_n_stack_t {
    char *var_n_fmt;
    char *var_n;
    int max_index;
    int index;
    int state;
};

struct set_var_t {
    char *set_var;
    int enabled;
};

struct iter_t {
    struct var_n_stack_t var_n_stack[VAR_N_STACK_SIZE];
    array_t *set_var_array;
    array_iterator_t current;

    char *set_var_fmt;
    char *opt_var_fmt;
    int set_var_state;
    int opt_var_state;

    int set_var_level;
};

static void
set_var_init(struct iter_t *iter)
{
    iter->set_var_array = init_array(SET_VAR_ARRAY_SIZE);
}

static int
get_level(char const *set_var)
{
    int level;
    char const *p;

    for (level = 0, p = set_var; (p = strchr(p, '%')) != NULL; ++level, ++p) {
        ; /* do nothing */
    }
    return level;
}

static char *
get_var_fmt(char const *set_var)
{
    int level = get_level(set_var);
    char *fmt = (char *) safe_malloc(strlen(set_var) + level + 1);
    char *p = fmt;
    while (*set_var) {
        if (*set_var == '%') {
            *p++ = '%';
            *p++ = 'd';
        }
        else {
            *p++ = *set_var;
        }
        ++set_var;
    }
    *p = '\0';
    return fmt;
}

static char *
gen_set_var(struct iter_t * iter, char const *set_var_fmt)
{
    char buf[1024];
    if (set_var_fmt != NULL) {
        sprintf(buf,
                set_var_fmt,
                iter->var_n_stack[0].index,
                iter->var_n_stack[1].index,
                iter->var_n_stack[2].index);
        assert(strlen(buf) < 1024);
        return strsave(buf);
    }
    else {
        return NULL;
    }
}

int
get_last_index(struct iter_t * iter)
{
    return iter->current.index + 1;
}

void
end_set_var_iteration(struct iter_t *iter)
{
    int i;

    ARRAY_ITER(it, iter->set_var_array, p, struct set_var_t) {
        _free(p->set_var);
    }
    free_array(iter->set_var_array);

    for (i = 0; i < iter->set_var_level; ++i) {
        _free(iter->var_n_stack[i].var_n_fmt);
    }

    _free(iter->set_var_fmt);
    _free(iter->opt_var_fmt);
    _free(iter);
}

static void
iterate(struct iter_t *iter, int level, int neg)
{
    if (level == iter->set_var_level) {
        char *set_var = gen_set_var(iter, iter->set_var_fmt);
        char *opt_var = gen_set_var(iter, iter->opt_var_fmt);
        struct set_var_t *p = (struct set_var_t *) safe_malloc(sizeof(struct set_var_t));

        p->set_var = set_var;
        p->enabled = is_var_enabled(set_var, opt_var, neg);
        append_new_elem(iter->set_var_array, p);

        _free(opt_var);
    }
    else {
        struct var_n_stack_t *s = &iter->var_n_stack[level];
        char *var_n = gen_set_var(iter, s->var_n_fmt);
        char const *val = get_variable(var_n);
        if (val) {
            int max = atoi(val);
            for (s->index = 1; s->index <= max; ++s->index) {
                iterate(iter, level + 1, neg);
            }
        }
        init_array_iterator(&iter->current, iter->set_var_array);
    }
}

struct iter_t *
init_set_var_iteration(char const *set_var, char const *opt_var, int neg)
{
    struct iter_t *iter = (struct iter_t *) safe_malloc(sizeof(struct iter_t));
    int level = get_level(set_var);
    char *var_n;
    char const *p;

    int l;
    if (!level) {
        fatal_exit ("BUG: init_set_var_iteration called "
                    "for non set variable %s\n", set_var);
    }
    set_var_init(iter);

    p = set_var;
    for (l = level; l--; ) {
        var_n = check_get_var_n(p);
        if (!var_n) {
            log_error("Error: array variable '%s' has no associated index variable\n", p);
            free(iter);
            return NULL;
        }
        iter->var_n_stack[l].state = check_var_optional_state(var_n);
        iter->var_n_stack[l].var_n_fmt = get_var_fmt(var_n);
        iter->var_n_stack[l].index = 0;
        iter->var_n_stack[l].max_index = 0;
        p = var_n;
    }

    iter->set_var_level = level;
    iter->set_var_fmt = get_var_fmt(set_var);
    iter->opt_var_fmt = opt_var ? get_var_fmt(opt_var) : 0;
    iter->set_var_state = check_var_optional_state(set_var);
    iter->opt_var_state = opt_var ? check_var_optional_state(opt_var) : CHECK_NON_OPTIONAL;

    iterate(iter, 0, neg);

    return iter;
}

char *
get_next_set_var(struct iter_t *iter, int *enabled)
{
    struct set_var_t *entry = (struct set_var_t *) get_next_elem(&iter->current);
    if (entry) {
        char *name = entry->set_var;
        *enabled = entry->enabled;
        log_info(LOG_VAR, "get_next_set_var: %s (%s)\n", name,
                 *enabled ? "enabled" : "disabled");

        return name;
    }
    return 0;
}
