/*----------------------------------------------------------------------------
 *  var.c - functions for processing variables
 *
 *  Copyright (c) 2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Creation:    2001-08-12  fm
 *  Last Update: $Id: var.c 34688 2014-11-07 12:56:15Z florian $
 *----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#if !defined(PLATFORM_mingw32)
#include <sys/utsname.h>
#endif

#include "defs.h"
#include "var.h"
#include "regex.h"
#include "check.h"
#include "log.h"
#include "str.h"
#include "cfg.h"
#include "array.h"
#include "options.h"

#define VAR_ARRAY_SIZE          1024

typedef struct {
    unsigned dq:1;
    unsigned supersede:1;
    unsigned checked:1;
    unsigned tagged:1;
    unsigned requested:1;
    unsigned copied:1;
    unsigned weak:1;
    unsigned numeric:1;
    unsigned generated:1;
    unsigned provide:1;
} flags_t;

typedef struct variables {
    char *package;
    char *name;
    char *content;
    char *comment;
    char *file;
    int line;
    flags_t flags;
} VARIABLES;

static array_t *var_array;

static int
var_dump_package_header(FILE *fp, char const *package)
{
    int error = 0;
    BOOL found = FALSE;
    DECLARE_ARRAY_ITER(array_iter, p, VARIABLES);

    if ((p = get_first_elem(var_array)) == NULL) {
        return 0;
    }

    error = fprintf(fp, "#\n# package '%s'", package) < 0 ? ERR : OK;

    ARRAY_ITER_LOOP(array_iter, var_array, p) {
        if (p->flags.provide) {
            if (strcmp(package, p->package) == 0) {
                if (!found) {
                    error = fprintf(fp, ", provides:") < 0 ? ERR : OK;
                    found = TRUE;
                }
                error = fprintf(fp, "\n#\t- %s version %s",
                                p->name, p->comment) < 0 ? ERR : OK;
            }
        }
    }

    error = fprintf(fp, "\n#\n") < 0 ? ERR : OK;

    return error;
}

int
check_variables(void)
{
    int ret = OK;

    ARRAY_ITER(array_iter, var_array, p, VARIABLES) {
        if (p->flags.weak) {
            /* skip weak variables */
            continue;
        }

        if (!p->flags.tagged && !p->flags.checked) {
            /* check wether there is a corrosponding entry
             * in check/package.txt
             */
            char *set_var = replace_set_var_indices(p->name);
            if (!check_var_defined(set_var)) {
                if (p->flags.supersede) {
                    log_error ("Unchecked variable %s='%s' in supersede config"
                               " file %s:%d. Please check whether you forgot "
                               "to install the package this variable belongs "
                               "to.\n",
                               p->name, p->content,
                               p->file, p->line);
                    ret = ERR;
                }
                else if (strong_consistency) {
                    log_error ("Unchecked variable %s='%s' in package %s. This "
                               "may be an obsolete variable from an earlier "
                               "version (check %s/%s%s and "
                               "documentation) or an error in %s/%s%s. "
                               "If there is an error in %s/%s%s, ask the "
                               "maintainer of the package to fix this.\n",
                               p->name, p->content, p->package,
                               config_dir, p->package,  def_cfg_ext ? def_cfg_ext : "",
                               check_dir, p->package,  def_check_ext ? def_check_ext : "",
                               check_dir, p->package,  def_check_ext ? def_check_ext : "");
                    ret = ERR;
                }
            }
#if 0 /* kristov: does not work for disabled set variables */
            else {
                while (strchr(set_var, '%') != NULL) {
                    char *s;
                    char *var_n = check_get_var_n(set_var);
                    if (var_n) {
                        char *val = get_variable(var_n);
                        if (val == NULL && check_var_optional(var_n)) {
                            log_info (LOG_INFO|LOG_VAR,
                                      "skipping optional variable '%s' "
                                      "(depends on undefined optional "
                                      "variable '%s')\n",
                                      p->name, var_n);
                            break;
                        }
                    }
                    else {
                        log_error ("undefined var_n for '%s'\n",
                                   set_var);
                        ret = ERR;
                        break;
                    }

                    s = set_var;
                    set_var = get_set_var_name_int(set_var, 1, __FILE__, __LINE__);
                    free(s);
                }
            }
#endif
            free(set_var);
        }
    }
    return ret;
}

int
dump_variables(char const *fname, char const *fname_full_cfg)
{
    FILE *fp;
    FILE *fp_full;
    char *last_package;

    int num;
    int total_num;
    int ret = OK;
    BOOL next_package;

#if !defined(PLATFORM_mingw32)
    struct utsname sys;
#endif

    DECLARE_ARRAY_ITER(array_iter, p, VARIABLES);

    log_info(LOG_INFO, "generating %s\n", fname);

    num = total_num = 0;

    fp = fopen(fname, "wb");
    if (!fp) {
        log_error("Error opening '%s': %s\n", fname, strerror (errno));
        return ERR;
    }

    fp_full = fopen(fname_full_cfg, "wb");
    if (!fp_full) {
        log_error("Error opening '%s': %s\n",
                  fname_full_cfg, strerror (errno));
        return ERR;
    }

#if !defined(PLATFORM_mingw32)
    if (uname(&sys) != -1) {
        if (fprintf(fp, "#\n# generated by mkfli4l (%s) running under "
                    "%s Version %s\n", BUILD_DATE, sys.sysname, sys.release) < 0) {
            log_error("Error while writing to %s: %s\n", fname, strerror(errno));
            ret = ERR;
            goto cleanup;
        }
    }
    else
#endif
    if (fprintf(fp, "#\n# generated by mkfli4l (%s)\n", BUILD_DATE) < 0) {
        log_error("Error while writing to %s: %s\n", fname, strerror(errno));
        ret = ERR;
        goto cleanup;
    }

    last_package = "";

    ARRAY_ITER_LOOP(array_iter, var_array, p) {
        BOOL ok = TRUE;
        char quote;
        ++total_num;

        if (p->flags.weak || p->flags.generated) {
            /* skip weak variables */
            continue;
        }

        next_package = strcmp(last_package, p->package) != 0;
        last_package = p->package;
        if (next_package) {
            ok = var_dump_package_header(fp_full, p->package) == OK && ok;
        }

        ok = fprintf(fp_full, "%s='%s'\n", p->name, p->content) >= 0 && ok;
        if (!ok) {
            log_error("Error while writing to %s: %s\n",
                      fname_full_cfg, strerror (errno));
            ret = ERR;
            break;
        }

        if (p->flags.tagged) {
            quote = p->flags.dq ? '"' : '\'';
            ++num;
            if (next_package) {
                ok = var_dump_package_header(fp, p->package) == OK && ok;
            }
            ok = (fprintf(fp, "%s=%c%s%c\n", p->name, quote, p->content, quote) >= 0) && ok;

            if (!ok) {
                log_error("Error while writing to %s: %s\n", fname, strerror (errno));
                ret = ERR;
                break;
            }
        }
    }

cleanup:
    if (ret == OK) {
        inc_log_indent_level();
        log_info(LOG_INFO, "total number of variables : %d, \twritten: %d\n",
                 total_num, num);
        log_info(LOG_INFO, "total size of variables   : %ld, \twritten: %ld\n",
                 ftell(fp_full), ftell(fp));
        dec_log_indent_level();
    }

    fclose(fp);
    fclose(fp_full);

    if (ret != OK) {
        unlink(fname);
        unlink(fname_full_cfg);
    }

    return ret;
}

static VARIABLES *last_lookup = NULL;

static VARIABLES *
lookup_variable(char const *name)
{
    DECLARE_ARRAY_ITER(array_iter, p, VARIABLES);

    if (last_lookup) {
        if (strcmp (name, last_lookup->name) == 0) {
            return last_lookup;
        }
        last_lookup = NULL;
    }

    ARRAY_ITER_LOOP(array_iter, var_array, p) {
        if (strcmp (name, p->name) == 0) {
            last_lookup = p;
            return last_lookup;
        }
    }

    return NULL;
}

char const *
get_variable(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    return p ? p->content : NULL;
}

char const *
get_variable_package(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    return p ? p->package : NULL;
}

char const *
get_variable_comment(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    return p ? p->comment : NULL;
}

void
get_variable_src(char const *name, char **src, int *line)
{
    static char *unknown = "unknown";
    VARIABLES *p = lookup_variable(name);

    if (p) {
        *src = p->file;
        *line = p->line;
    }
    else {
        *src = unknown;
        *line = -1;
    }
}

static void
chkvar_add_empty_instance(char const *var, int type, char const *package)
{
    /* add an empty instance of this variable */
    if (set_variable(package, var, type == TYPE_NUMERIC ? "0" : "", NULL, FALSE, "internal", FALSE, 0) != OK) {
        fatal_exit("(%s:%d) internal error while generating "
                   "entry for %s\n", __FILE__, __LINE__, var);
    }
    mark_var_generated(var);
    mark_var_checked(var);
    mark_var_tagged(var);
    if (type == TYPE_NUMERIC) {
        mark_var_numeric(var);
    }
}

int
chkvar(char const *name, char const *regexp_name, char const *user_regexp,
       char const *package, BOOL enabled, int type, int opt, char const *defval)
{
    char const *p = get_variable(name);

    log_info(LOG_VAR, "checking %s='%s'\n", name, p ? p : "(undefined)");

    if (!p) {
        if (opt == CHECK_NON_OPTIONAL) {
            if (enabled) {
                if (!defval) {
                    log_error("Error: missing variable '%s' in config "
                              "file of package '%s'\n", name, package);
                    return ERR;
                }
                else {
                    set_variable(package, name, defval, "default value", FALSE,
                                 "check file", FALSE, 0);
                    log_info(LOG_INFO, "using default value for %s='%s'\n",
                             name, defval);
                    p = defval;
                }
            }
            else {
                return OK;
            }
        }
        else {
            chkvar_add_empty_instance(name, type, package);
            return OK;
        }
    }

    mark_var_checked(name);
    if (type == TYPE_NUMERIC) {
        mark_var_numeric(name);
    }
    if (enabled) {
        mark_var_tagged(name);
        return regexp_chkvar(name, p, regexp_name, user_regexp);
    }
    else {
        return OK;
    }
}

int
chkvar_idx(char const *var, char const *max_idx_name, char const *opt_name,
           char const *regexp_name, char const *user_regexp, char const *package,
           int type, int opt, BOOL neg)
{
    char *tmp_var;
    int error = OK;
    BOOL enabled;
    struct iter_t *iter;

    log_info(LOG_VAR, "checking index var %s, var_n %s\n", var, max_idx_name);

    iter = init_set_var_iteration(var, opt_name, neg);
    if (!iter) {
        return ERR;
    }

    while ((tmp_var = get_next_set_var(iter, &enabled)) != NULL) {
        int ret;
        if (get_variable(tmp_var)) {
            ret = chkvar(tmp_var, regexp_name, user_regexp, package,
                         enabled, type, opt, NULL);
            if (ret != OK) {
                error = ret;
            }
        }
        else {
            if (enabled) {
                if (opt != CHECK_REALLY_OPTIONAL) {
                    log_error("Error: missing variable '%s' in config "
                              "file of package '%s'\n", tmp_var, package);
                    error = ERR;
                }
                else {
                    chkvar_add_empty_instance(tmp_var, type, package);
                }
            }
        }
    }
    end_set_var_iteration(iter);

    return error;
}

#define VAR_NAME        0
#define VAR_VAL         2
#define VAR_COMMENT     3

void
var_init(void)
{
    var_array = init_array(VAR_ARRAY_SIZE);
}

int
read_config_file(char const *config_file, char const *package, BOOL def)
{
    int ret;
    int result = OK;

    struct token_t tokens[] = {
        TOKEN(CFG_ID),
        TOKEN(CFG_EQUAL),
        TOKEN(CFG_STRING),
        TOKEN(CFG_COMMENT|CFG_NL),
        TOKEN(CFG_NONE),
    };

    cfg_fopen(config_file);
    while ((ret = get_config_tokens(tokens, 0)) != CFG_EOF) {
        if (ret == CFG_ERROR) {
            result = ERR;
            continue;
        }

        if (set_variable(package, tokens[VAR_NAME].text,
                         tokens[VAR_VAL].text, tokens[VAR_COMMENT].text,
                         def, config_file, tokens[VAR_VAL].dq,
                         tokens[VAR_NAME].line) == ERR) {
            log_error("%s: line %d: variable %s already defined\n",
                      config_file, tokens[VAR_NAME].line,
                      tokens[VAR_NAME].text);
            result = ERR;
        }
    }
    cfg_fclose();

    return result;
}

int
set_variable(char const *package, char const *name, char const *content, char const *comment,
             BOOL supersede, char const *config_file, BOOL dq, int line)
{
    VARIABLES *p;
    static char *do_debug = "_DO_DEBUG";
    int offset;

    log_info(LOG_VERBOSE, "adding %s:%s='%s'%s\n",
             package, name, content, comment ? comment : "");

    p = lookup_variable(name);
    if (p) {
        if (!p->flags.supersede || supersede) {
            return ERR;
        }

        p->flags.supersede = 0;
        log_info(LOG_INFO, "using supersede value %s='%s' "
                 "instead of %s='%s' in %s:%d\n",
                 name, p->content, name, content, config_file, line);
        return OK;
    }

    last_lookup = NULL;
    p = (VARIABLES *) safe_malloc(sizeof(VARIABLES));
    offset = strlen(name) - strlen(do_debug);

    p->package   = strsave(package);
    p->name      = strsave(name);
    p->content   = strsave(content);
    p->comment   = strsave(comment);
    p->file      = strsave(config_file);
    p->line      = line;
    p->flags.dq        = dq ? 1 : 0;
    p->flags.supersede = supersede ? 1 : 0;
    p->flags.checked   = 0;
    p->flags.tagged    = 0;
    p->flags.requested = 0;
    p->flags.copied    = 0;
    p->flags.weak      = 0;
    p->flags.numeric   = 0;
    p->flags.generated = 0;
    p->flags.provide   = 0;

    if (offset > 0 && strcmp(name + offset, do_debug) == 0) {
        /* copy .*_DO_DEBUG variables without further checking */
        log_info(LOG_VERBOSE, "copying debug macro %s\n", name);
        p->flags.tagged = 1;
    }

    append_new_elem(var_array, p);
    return OK;
}

int
set_variable_content(char const *name, char const *content)
{
    VARIABLES *p = lookup_variable(name);
    if (!p) {
        return ERR;
    }

    free(p->content);
    p->content = strsave(content);
    p->flags.generated = 0;
    return OK;
}

void
var_add_weak_declaration(char const *package, char const *name, char const *value,
                         char const *comment, int type, char const *file,
                         int line, int log_level)
{
    VARIABLES *p = lookup_variable (name);

    if (!p) {
        log_info(log_level, "adding weak declaration for %s:%s='%s'\n",
                 package, name, value);

        set_variable(package, name, value, comment, FALSE, NULL, FALSE, 0);
        p = lookup_variable(name);
        if (!p) {
            fatal_exit("%s %d: Error adding weak deklaration\n",
                       __FILE__, __LINE__);
        }
    }
    else {
        if (p->flags.weak) {
            log_info(log_level, "overwriting weak declaration for %s:%s='%s'"
                     " with %s:%s='%s'\n",
                     p->package, name, p->content, package, name, value);
            free (p->content);
            free (p->package);
            free (p->comment);
            p->content = strsave(value);
            p->package = strsave(package);
            p->comment = strsave(comment);
        }
        else {
            fatal_exit("(%s:%d) trying to overwrite config variable '%s', "
                       "aborting...\n", file, line, p->name);
        }
    }
    p->flags.checked = 1;
    p->flags.weak    = 1;
    switch (type) {
    case TYPE_NUMERIC:
        p->flags.numeric = 1;
        break;
    case TYPE_UNKNOWN:
        p->flags.numeric = 0;
        break;
    default:
        fatal_exit("(%s:%d) unknown type %x\n",
                   __FILE__, __LINE__, type);
    }
}

int
mark_var_tagged(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    if (!p) {
        return ERR;
    }
    else {
        p->flags.tagged = 1;
        return OK;
    }
}

int
mark_var_checked (char const *name)
{
    VARIABLES *p = lookup_variable(name);
    if (!p) {
        log_error("undefined variable %s\n", name);
        return ERR;
    }
    else {
        p->flags.checked = 1;
        return OK;
    }
}

int
mark_var_requested(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    if (!p) {
        return ERR;
    }
    else {
        p->flags.requested = 1;
        return OK;
    }
}

int
mark_var_copied(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    if (!p) {
        return ERR;
    }
    else {
        p->flags.copied = 1;
        return OK;
    }
}

int
mark_var_numeric(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    if (!p) {
        return ERR;
    }
    else {
        log_info(LOG_TREE_EXEC, "setting %s to numeric type\n", name);
        p->flags.numeric = 1;
        return OK;
    }
}

int
mark_var_generated(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    if (!p) {
        return ERR;
    }
    else {
        p->flags.generated = 1;
        return OK;
    }
}

int
mark_var_provide(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    if (!p) {
        return ERR;
    }
    else {
        p->flags.provide = 1;
        return OK;
    }
}

BOOL
is_var_checked(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    return p ? p->flags.checked : FALSE;
}

BOOL
is_var_tagged(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    return p ? p->flags.tagged : FALSE;
}

BOOL
is_var_enabled(char const *name, char const *opt_var, int neg)
{
    char const *val;

    if (!opt_var) {
        return TRUE;
    }

    if ((val = get_variable (opt_var)) != NULL) {
        if (!is_var_checked(opt_var)) {
            fatal_exit("%s depends on unchecked variable %s\n",
                       name, opt_var);
        }
        if (neg) {
            /* kristov TODO: return strcmp(val, "yes") != 0 */
            return strcmp(val, "no") == 0;
        }
        else {
            return is_var_tagged(opt_var) && strcmp(val, "yes") == 0;
        }
    }
    else {
        return FALSE;
    }
}

BOOL
is_var_numeric(char const *name)
{
    VARIABLES *p = lookup_variable(name);

    if (p && p->flags.numeric) {
        log_info(LOG_TREE_EXEC, "(%s:%d) %s numeric type (%d)\n",
                 __FILE__, __LINE__, name, (int) p->flags.numeric);
        return TRUE;
    }
    else {
        return FALSE;
    }
}

BOOL
is_var_weak(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    return p ? p->flags.weak : FALSE;
}

BOOL
is_var_generated(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    return p ? p->flags.generated : FALSE;
}

BOOL
is_var_provide(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    return p ? p->flags.provide : FALSE;
}

BOOL
is_var_copy_pending(char const *name)
{
    VARIABLES *p = lookup_variable(name);
    return p ? p->flags.tagged && p->flags.requested && !p->flags.copied : FALSE;
}

BOOL
is_var_unique(char const *name, char const *file, int line)
{
    char *val;

    DECLARE_ARRAY_ITER(array_iter, p, VARIABLES);
    VARIABLES *q = lookup_variable(name);

    if (!q) {
        return ERR;
    }

    val = q->content;

    ARRAY_ITER_LOOP(array_iter, var_array, p) {
        if (q->flags.tagged && strcmp(val, p->content) == 0 && p != q) {
#if 0
            char buffer[1024];
            snprintf(buffer, sizeof (buffer), "Variable '%s' (%s:%d) contains the same value as '%s'\n",  p->name, p->file, p->line, name);
            parse_warning(buffer, file, line);
#else
            (void) file;
            (void) line;
#endif
            return FALSE;
        }
    }

    return TRUE;
}

void
free_variables(void) {
    ARRAY_ITER(array_iter, var_array, p, VARIABLES) {
        free(p->name);
        free(p->package);
        free(p->content);
        free (p->comment);
    }
}
