/*----------------------------------------------------------------------------
 *  mkfli4l.c - creates archives for fli4l
 *
 *  Copyright (c) 2000-2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Creation:    2000-05-03  fm
 *  Last Update: $Id: mkfli4l.c 41037 2015-08-22 20:11:58Z kristov $
 *----------------------------------------------------------------------------
 */

#include <stdlib.h>
#include <signal.h>
#include <time.h>

#ifdef MALLOC_DEBUG
#include <mcheck.h>
#endif

#include <libmkfli4l/defs.h>
#include <libmkfli4l/var.h>
#include <libmkfli4l/regex.h>
#include <libmkfli4l/check.h>
#include <libmkfli4l/log.h>
#include <libmkfli4l/parse.h>
#include <libmkfli4l/options.h>
#include <ziplist/opt.h>
#include <ziplist/optlibs.h>
#include <ziplist/kernmod.h>
#include <boottype/boottype.h>
#include "config.h"
#include "buildinfo.h"
#include "mk_syslinux.h"

#define VAR_BOOT_TYPE "BOOT_TYPE"

static void
catch(int sig)
{
  abort();
  (void) sig;
}

/*----------------------------------------------------------------------------
 *  main (int argc, char ** argv)
 *----------------------------------------------------------------------------
 */
int
main(int argc, char ** argv)
{
    int   res = OK;
    char const *kernel_version = NULL;
    BOOL  check_only = FALSE;
#ifdef MALLOC_DEBUG
    mtrace();
#endif

    signal(SIGTERM, catch);
    signal(SIGINT, catch);

    srand((unsigned int) time(NULL));
    get_options(argc, argv);
    open_logfile(logfile);

    var_init();
    check_init();
    regexp_init();
    zip_list_init();
    boot_type_init();

    log_info(LOG_INFO, "reading configuration files\n\n");
    inc_log_indent_level();

    set_variable("base", "AUTO", "99", "predefined variable", 0, NULL, 0, 0);
    mark_var_tagged("AUTO");
    mark_var_checked("AUTO");

    if (!*opt_packages) {
        if (read_mkfli4l_config(config_dir, check_dir) != OK) {
            fatal_exit("Error in configuration, aborting...\n");
        }
    }
    else {
        check_only = TRUE;
        if (read_config(config_dir, check_dir, opt_packages) != OK) {
            fatal_exit("Error in configuration, aborting...\n");
        }
    }
    if (check_for_dangling_extensions() != OK) {
        fatal_exit ("Error in configuration, aborting...\n");
    }
    dec_log_indent_level();
    log_info(LOG_INFO, "\nfinished reading configuration files\n");

    if (set_arch() != OK) {
        fatal_exit("unable to read fli4l architecture\n");
    }
    else {
        log_info (LOG_INFO, "architecture: %s\n", get_variable("ARCH"));
    }

    log_info(LOG_INFO, "checking variables\n");
    if (check_all_variables() != OK) {
        fatal_exit("Error in configuration, aborting...\n");
    }

    if (check_variables() != OK) {
        fatal_exit("Error in configuration or packet structure, aborting...\n");
    }

    if (is_mkfli4l && !check_only) {
        /* ignore package list if there is no KERNEL_VERSION defined */
        if ((kernel_version = get_variable ("KERNEL_VERSION")) != NULL) {
            log_info (LOG_INFO, "generating zip list\n\n");
            inc_log_indent_level ();
            res = make_zip_list (kernel_version);
            dec_log_indent_level ();
            log_info (LOG_INFO, "\nfinished generating zip list\n");
        }
        else {
            fatal_exit("undefined KERNEL_VERSION, "
                       "unable to generate of package list\n");
        }
    }

    if (execute_all_extended_checks(check_dir, def_extcheck_ext, packages, n_packages) != OK || res != OK) {
        fatal_exit ("Error in configuration, aborting...\n");
    }

    if (is_mkfli4l && !check_only) {
        char const *boottype = get_variable(VAR_BOOT_TYPE);
        if (boottype == NULL) {
            fatal_exit(VAR_BOOT_TYPE " must be defined\n");
        }

        if (resolve_dependencies(kernel_version) != OK) {
            fatal_exit("unable to resolve dependecies\n");
        }

        if (resolve_library_dependencies() != OK) {
            fatal_exit("unable to resolve library dependecies\n");
        }

        if (create_library_symlinks() != OK) {
            fatal_exit("unable to create library symlinks\n");
        }

        if (create_build_info() != OK) {
            fatal_exit("unable to create build information\n");
        }

        if (determine_rootfs_compression(kernel_version) != OK) {
            fatal_exit("unable to determine ROOTFS compression\n");
        }

        if (dump_variables(rc_file, full_rc_file) != OK) {
            fatal_exit("error while writing %s (error in variables)\n", rc_file);
        }

        if (add_boot_type_specific_entries(boottype) != OK) {
            fatal_exit("unable to perform actions dependent on the boot type\n");
        }

        if (prepare_zip_lists() != OK) {
            fatal_exit("error while preparing zip lists\n");
        }

        if (create_archives(boottype) != OK) {
            fatal_exit("error while creating archives\n");
        }

#ifdef SKIP_SYSLINUX
        if (!access(syslinux_template_file, R_OK)) {
#endif
            log_info(LOG_INFO, "generating syslinux.cfg...\n");
            inc_log_indent_level();
            if (mk_syslinux(syslinux_template_file, syslinux_cfg_file) != OK) {
                fatal_exit("Error building syslinux.cfg, aborting...\n");
            }
            dec_log_indent_level();
            log_info(LOG_INFO, "\nfinished generating syslinux.cfg\n");
#ifdef SKIP_SYSLINUX
        } else {
            log_error("-> skipping generation of syslinux.cfg due to missing template file\n");
        }
#endif
    }

    close_logfile();

    free_check_variables();
    free_variables();

    return 0;
}
