/*****************************************************************************
 *  @file var.h
 *  Functions for processing variables.
 *
 *  Copyright (c) 2000-2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: var.h 34688 2014-11-07 12:56:15Z florian $
 *****************************************************************************
 */

#ifndef MKFLI4L_VAR_H_
#define MKFLI4L_VAR_H_

#include <vartype.h>
#include <package.h>
#include <container/hashmap.h>

struct var_properties_t;
struct var_t;
struct var_scope_t;
struct var_value_iterator_t;
struct expr_t;

/**
 * @name Access modes.
 *
 * The access modes are used by the var_set_value() and
 * var_set_and_check_value() functions and describe how to cope with the
 * overriding of existing values. Note that the higher the access mode, the
 * higher the requirements that have to be met to make the access successful.
 * @{
 */
/**
 * Specifies tentative access. Access priority must be at least as high as the
 * slot priority in order to perform a value change. If the access priority is
 * lower, though, the change request is ignored without generating an error.
 *
 * This access mode is used to change array bound variables as these variables
 * may be explicitly in the configuration, such that internal changes have to
 * be ignored in this case.
 */
#define VAR_ACCESS_TENTATIVE 1
/**
 * Specifies weak access. Access priority must be at least as high as the slot
 * priority in order to perform a value change. If the access priority is
 * lower, an error is generated and the slot is not changed.
 *
 * This access mode is used to change local script variables as this mode
 * allows overriding variables previously set.
 */
#define VAR_ACCESS_WEAK 2
/**
 * Specifies strong access. Access priority must be higher than the slot
 * priority in order to perform a value change. If the access priority is
 * equal or lower, an error is generated and the slot is not changed.
 *
 * This access mode is used to set variable values coming from the
 * configuration, as multiple changes to the same variable are not allowed.
 */
#define VAR_ACCESS_STRONG 3
/**
 * @}
 */

/**
 * @name Access priorities.
 *
 * The access priorities are used by the var_set_value() and
 * var_set_and_check_value() functions and describe the "overridability" of
 * values.
 * @{
 */
/**
 * This priority is used for undefined slots.
 */
#define VAR_PRIORITY_UNDEFINED (-1)
/**
 * This priority is used for updating array bound variables when resizing an
 * array.
 */
#define VAR_PRIORITY_ARRAY_BOUND 0
/**
 * This priority is used for updating local transient variables within an
 * extended check script.
 */
#define VAR_PRIORITY_SCRIPT_TRANSIENT 0
/**
 * This priority is used for updating persistent variables with values
 * originating from the normal configuration.
 */
#define VAR_PRIORITY_CONFIGURATION 1
/**
 * This priority is used for updating persistent variables with values
 * originating from the superseding configuration.
 */
#define VAR_PRIORITY_SUPERSEDE 2
/**
 * This priority is used for updating persistent variables from within an
 * extended check script.
 */
#define VAR_PRIORITY_SCRIPT_PERSISTENT 3
/**
 * @}
 */

/**
 * @name Values.
 * @{
 */

/**
 * Describes a value of a variable slot.
 */
struct var_value_t {
    /**
     * The value.
     */
    char const *value;
    /**
     * The slot priority.
     */
    int priority;
};

/**
 * Converts a variable value to integer type.
 *
 * @param THIS
 *  The variable value as returned by var_get_value().
 * @param result
 *  Receives the integer value.
 * @return
 *  TRUE if the type conversion was successful, FALSE otherwise.
 */
BOOL
var_value_convert_to_integer(struct var_value_t const *THIS, int *result);

/**
 * Converts a variable value to boolean type.
 *
 * @param THIS
 *  The variable value as returned by var_get_value().
 * @param result
 *  Receives the integer value.
 * @return
 *  TRUE if the type conversion was successful, FALSE otherwise.
 */
BOOL
var_value_convert_to_boolean(struct var_value_t const *THIS, BOOL *result);

/**
 * Destroys a value.
 *
 * @param THIS
 *  The value to destroy.
 */
void
var_value_destroy(struct var_value_t *THIS);

/**
 * @}
 */

/**
 * @name Variable creation, lookup, and scopes.
 * @{
 */

/**
 * Returns the global variable scope.
 *
 * @return
 *  The global variable scope.
 */
struct var_scope_t *
var_get_global_scope(void);

/**
 * Creates a variable scope.
 *
 * @param location
 *  The location pointing to the beginning of the scope. Ownership is passed to
 *  the scope.
 * @param parent
 *  The parent scope.
 * @return
 *  The new scope being an inner scope of "parent".
 */
struct var_scope_t *
var_scope_create(
    struct location_t *location,
    struct var_scope_t *parent
);

/**
 * Looks up a variable in a given scope hierarchy.
 *
 * @param THIS
 *  The scope to search.
 * @param name
 *  The name of the variable.
 * @param recursive
 *  If TRUE, parent scopes are also searched.
 * @return
 *  A pointer to the corresponding var_t object or NULL if not found.
 */
struct var_t *
var_scope_try_get(
    struct var_scope_t *THIS,
    char const *name,
    BOOL recursive
);

/**
 * Looks up a variable in a given scope hierarchy as var_try_get(). If the
 * variable cannot be found, an error message is logged.
 *
 * @param THIS
 *  The scope to search.
 * @param name
 *  The name of the variable.
 * @param recursive
 *  If TRUE, parent scopes are also searched.
 * @return
 *  A pointer to the corresponding var_t object or NULL if not found.
 */
struct var_t *
var_scope_get(
    struct var_scope_t *THIS,
    char const *name,
    BOOL recursive
);

/**
 * Iterates over all variables in a scope and calls a visitor on them. If the
 * visitor returns FALSE at some point, the iteration is aborted and FALSE is
 * returned.
 *
 * @param THIS
 *  The scope.
 * @param recursive
 *  If TRUE, parent scopes are also visited.
 * @param visitor
 *  The visitor.
 * @param data
 *  The data for the visitor.
 * @return
 *  TRUE if all visitor calls returned TRUE, FALSE otherwise.
 */
BOOL
var_scope_foreach(
    struct var_scope_t *THIS,
    BOOL recursive,
    hashmap_visitor_t visitor,
    void *data
);

/**
 * Destroys a variable scope.
 *
 * @param THIS
 *  The scope to destroy.
 */
void
var_scope_destroy(struct var_scope_t *THIS);

/**
 * Looks up a variable recursively in the global scope.
 *
 * @param name
 *  The name of the variable.
 * @return
 *  A pointer to the corresponding var_t object or NULL if not found.
 */
struct var_t *
var_try_get(char const *name);

/**
 * Looks up a variable recursively in the global scope. If the variable cannot
 * be found, an error message is logged.
 *
 * @param name
 *  The name of the variable.
 * @return
 *  A pointer to the corresponding var_t object or NULL if not found.
 */
struct var_t *
var_get(char const *name);

/**
 * Adds a new variable.
 *
 * @param scope
 *  The scope where to add the variable to.
 * @param name
 *  The name of the variable.
 * @param type
 *  The type of the variable.
 * @param location
 *  The location of the variable definition.
 * @param props
 *  The properties to use. Ownership is passed to the variable.
 * @return
 *  A pointer to the new var_t object or NULL if the variable could not be
 *  created (e.g. because a variable with the same name already exists).
 */
struct var_t *
var_add(
    struct var_scope_t *scope,
    char const *name,
    struct vartype_t *type,
    struct location_t *location,
    struct var_properties_t *props
);

/**
 * @}
 */

/**
 * @name General properties.
 * @{
 */

/**
 * Returns the name of a variable. If the variable name contains placeholders,
 * they are left unchanged.
 *
 * @param THIS
 *  The variable.
 * @return
 *  The name of the variable.
 */
char const *
var_get_name(struct var_t const *THIS);

/**
 * Returns the type of a variable.
 *
 * @param THIS
 *  The variable.
 * @return
 *  The type of the variable.
 */
struct vartype_t *
var_get_type(struct var_t const *THIS);

/**
 * Returns the level of a variable. The level of a variable denotes the number
 * of indices needed to access a variable slot and corresponds to the number of
 * '%' placeholders in the variable's name.
 *
 * @param THIS
 *  The variable.
 * @return
 *  The level of the variable.
 */
int
var_get_level(struct var_t const *THIS);

/**
 * Returns the properties associated with a variable.
 *
 * @param THIS
 *  The variable.
 * @return
 *  The properties associated with the variable.
 */
struct var_properties_t const *
var_get_properties(struct var_t const *THIS);

/**
 * Updates the properties of a variable. Not all properties can be changed,
 * however. Currently, the only possible changes are to modify the variable's
 * default value and to specify a new condition.
 *
 * @param THIS
 *  The variable.
 * @param props
 *  The new properties to use. Ownership is passed to the variable if
 *  successful.
 * @return
 *  TRUE if the properties have been changed successfully, FALSE otherwise.
 */
BOOL
var_set_properties(struct var_t *THIS, struct var_properties_t *props);

/**
 * Returns a variable's condition. Note that this is not the same as the
 * condition stored in the variable's properties, as a variable's condition may
 * depend on the conditions of other variables. This is especially true for
 * array bound variables whose condition is a logical OR of its own condition
 * and the conditions of all array variables referencing it.
 *
 * @param THIS
 *  The variable.
 * @return
 *  The condition.
 */
struct expr_t *
var_get_condition(struct var_t const *THIS);

/**
 * Returns a variable's location.
 *
 * @param THIS
 *  The variable.
 * @return
 *  The location.
 */
struct location_t const *
var_get_location(struct var_t const *THIS);

/**
 * Returns TRUE if the variable is enabled. A variable is enabled if it either
 * does not have a condition associated or if its condition evaluates to TRUE.
 * If the evaluation of a condition leads to a cycle (e.g. variable A depends
 * on variable B containing "yes" which itself depends on variable A containing
 * "yes"), the variable is taken to be disabled and FALSE is returned.
 */
BOOL
var_is_enabled(struct var_t const *THIS);

/**
 * @}
 */

/**
 * @name Additional properties.
 * @{
 */

/**
 * Describes additional properties of a variable that possess a sensible
 * default value.
 */
struct var_properties_t {
    /**
     * A comment associated with the variable.
     */
    char *comment;
    /**
     * TRUE if the variable is persistent, i.e. will be written to the image's
     * configuration. Transient (non-persistent) variables are e.g. used as
     * local variables in extended check scripts. Defaults to TRUE.
     */
    BOOL persistent;
    /**
     * TRUE if the variable is optional, i.e. it may be missing some or all
     * values. Defaults to FALSE.
     */
    BOOL optional;
    /**
     * The default value for this variable. This member is only used if
     * "optional" is set to FALSE. It may be NULL if no default value exists.
     * Defaults to NULL.
     */
    char *def_value;
    /**
     * The condition enabling this variable. May be NULL. If a condition is set
     * and evaluates to FALSE, the associated variable is not written to the
     * image's configuration. Defaults to NULL.
     */
    struct expr_t *condition;
};

/**
 * Creates a set of default properties for a new variable. The default
 * properties describe a persistent, mandatory (non-optional) variable without
 * a default value and without an associated condition.
 *
 * @return
 *  A property set containing default properties for a new variable.
 */
struct var_properties_t *
var_properties_create(void);

/**
 * Clones a property set.
 *
 * @param THIS
 *  The property set to clone.
 * @return
 *  The cloned properties.
 */
struct var_properties_t *
var_properties_clone(struct var_properties_t const *THIS);

/**
 * Changes the comment of a property set.
 *
 * @param THIS
 *  The property set to modify.
 * @param comment
 *  The comment. Note that the old comment (if any) is free()ed before, so
 *  <code>
 *    var_properties_set_comment(props, props->comment)
 *  </code> will fail.
 */
void
var_properties_set_comment(
    struct var_properties_t *THIS,
    char const *comment
);

/**
 * Changes the default value of a property set.
 *
 * @param THIS
 *  The property set to modify.
 * @param def_value
 *  The new default value. Note that the old default value (if any) is free()ed
 *  before, so <code>
 *    var_properties_set_default_value(props, props->def_value)
 *  </code> will fail.
 */
void
var_properties_set_default_value(
    struct var_properties_t *THIS,
    char const *def_value
);

/**
 * Changes the condition of a property set.
 *
 * @param THIS
 *  The property set to modify.
 * @param condition
 *  The new condition. Note that the old condition (if any) is freeed before,
 *  so <code>
 *    var_properties_set_condition(props, props->condition)
 *  </code> will fail.
 */
void
var_properties_set_condition(
    struct var_properties_t *THIS,
    struct expr_t *condition
);

/**
 * Checks if a property set is self-consistent. For example, an optional
 * variable may not have a default value.
 *
 * @param THIS
 *  The property set to check.
 * @param name
 *  The name of the variable this set is to be associated with. It is used in
 *  error messages if the property set is incorrect.
 * @return
 *  TRUE if the property set is self-consistent, FALSE otherwise.
 */
BOOL
var_properties_check(struct var_properties_t const *THIS, char const *name);

/**
 * Destroys a property set.
 *
 * @param THIS
 *  The property set to destroy.
 */
void
var_properties_destroy(struct var_properties_t *THIS);

/**
 * @}
 */

/**
 * @name Instances.
 * @{
 */

/**
 * Describes a variable instance. A variable instance is basically a slot
 * address and consists of a variable together with a list of indices. If all
 * indices the variable requires are given, the instance is said to be
 * "complete" (and points to a scalar slot), otherwise it is said to be
 * "partial" (and points to an array slot).
 */
struct var_instance_t {
    /**
     * The variable.
     */
    struct var_t *var;
    /**
     * The number of indices in the array.
     */
    int num_indices;
    /**
     * The indices.
     */
    unsigned *indices;
};

/**
 * Creates a variable instance.
 *
 * @param THIS
 *  The variable.
 * @param num_indices
 *  The number of indices to be found in the "indices" array.
 * @param indices
 *  The indices to use. If you provide less entries than the level of the
 *  variable requires, a partial instance is created, otherwise the instance
 *  is complete. May be NULL if to initialize all indices with zero.
 * @return
 *  A variable instance.
 */
struct var_instance_t *
var_instance_create(
    struct var_t *THIS,
    int num_indices,
    unsigned const *indices
);

/**
 * Tries to deduce variable and indices used from the name of a variable
 * instance. For example, V_1_2 returns the variable named V_%_% and the
 * indices 0 and 1 (remember that indices internally count from zero).
 *
 * This function also works for non-array variables; in such a case, the
 * index array returned is NULL.
 *
 * @param scope
 *  The scope where to perform the lookup.
 * @param name
 *  The name of the variable slot.
 * @return
 *  A variable instance or NULL if deduction failed.
 */
struct var_instance_t *
var_instance_create_from_string(struct var_scope_t *scope, char const *name);

/**
 * Returns the string value of a variable instance by replacing the index
 * placeholders by index values. Note that internally, indices are counted from
 * zero, but in variable names visible to the user, they are counted from one.
 *
 * @param THIS
 *  The variable instance.
 * @return
 *  A dynamically allocated string containing the substituted variable name.
 */
char *
var_instance_to_string(struct var_instance_t const *THIS);

/**
 * Returns the value of a variable slot. Note that if the variable is not set,
 * an error is logged. To prevent this, use var_is_value_set() to check whether
 * there is a value that can be read.
 *
 * @param THIS
 *  The variable instance addressing the scalar slot to be read. If the
 *  variable instance addresses an array slot, NULL is returned as one cannot
 *  read a whole array at once.
 * @return
 *  The value of the variable or NULL if the accessed variable slot has not
 *  been set or another error occurs. Use var_value_destroy() to destroy it if
 *  you don't need it anymore.
 */
struct var_value_t *
var_instance_get_value(struct var_instance_t *THIS);

/**
 * Returns the value of a variable slot. If the variable is not set, NULL is
 * returned without generating an error message.
 *
 * @param THIS
 *  The variable instance addressing the scalar slot to be read. If the
 *  variable instance addresses an array slot, NULL is returned as one cannot
 *  read a whole array at once.
 * @return
 *  The value of the variable or NULL if the accessed variable slot has not
 *  been set. Use var_value_destroy() to destroy it if you don't need it
 *  anymore.
 */
struct var_value_t *
var_instance_try_get_value(struct var_instance_t *THIS);

/**
 * Sets the value of a variable slot and checks whether it conforms to the
 * variable's type.
 *
 * @param THIS
 *  The variable instance addressing the scalar slot to be written to. If the
 *  variable instance addresses an array slot, NULL is returned as one cannot
 *  write to a whole array at once.
 * @param value
 *  The value.
 * @param mode
 *  The access mode, expressed by some VAR_ACCESS_* constant.
 * @param priority
 *  The priority of the value being set. This is some arbitrary value that is
 *  stored together with the variable slots and that, together with the access
 *  mode, controls whether a slot can be changed or not.
 * @return
 *  TRUE if successful, FALSE otherwise.
 */
BOOL
var_instance_set_and_check_value(
    struct var_instance_t *THIS,
    char const *value,
    int mode,
    int priority
);

/**
 * Sets the value of a variable slot. The value is not checked against the
 * variable's type. This function is meant to be used while reading the
 * configuration values as at this time the variable type system is not
 * completely ready (there may exist type extensions which refer to variables
 * being read in).
 *
 * @param THIS
 *  The variable instance addressing the scalar slot to be written to. If the
 *  variable instance addresses an array slot, NULL is returned as one cannot
 *  write to a whole array at once.
 * @param value
 *  The value.
 * @param mode
 *  The access mode, expressed by some VAR_ACCESS_* constant.
 * @param priority
 *  The priority of the value being set. This is some arbitrary value that is
 *  stored together with the variable slots and that, together with the access
 *  mode, controls whether a slot can be changed or not.
 * @return
 *  TRUE if successful, FALSE otherwise.
 */
BOOL
var_instance_set_value(
    struct var_instance_t *THIS,
    char const *value,
    int mode,
    int priority
);

/**
 * Clears an array (or sub-array), i.e. makes the array contain no elements
 * afterwards.
 *
 * @param THIS
 *  The variable instance addressing the array slot. If the variable instance
 *  addresses a scalar slot, FALSE is returned as one cannot reset non-array
 *  variables to NULL.
 * @param mode
 *  The access mode, expressed by some VAR_ACCESS_* constant.
 * @param priority
 *  The priority of the value being set. This is some arbitrary value that is
 *  stored together with the variable slots and that, together with the access
 *  mode, controls whether a slot can be changed or not.
 * @return
 *  TRUE if clearing was successful, FALSE otherwise.
 */
BOOL
var_instance_clear(struct var_instance_t *THIS, int mode, int priority);

/**
 * Destroys a variable instance.
 *
 * @param THIS
 *  The variable instance to destroy.
 */
void
var_instance_destroy(struct var_instance_t *THIS);

/**
 * @}
 */

/**
 * @name Iterators.
 * @{
 */

/**
 * Creates an iterator iterating over the values of an array variable. This
 * iterator points to the first element of the array or is the END iterator if
 * the array does not contain any values.
 *
 * @param THIS
 *  The variable instance addressing the array to iterate over. If this points
 *  to a scalar slot, the iterator returned will be an END iterator.
 * @return
 *  A var_value_iterator_t object. Use var_value_iterator_destroy() to destroy
 *  it if you don't use it anymore.
 */
struct var_value_iterator_t *
var_value_iterator_create(struct var_instance_t *THIS);

/**
 * Returns the variable associated with an iterator.
 *
 * @param THIS
 *  The iterator.
 * @return
 *  The associated variable.
 */
struct var_t *
var_value_iterator_get_var(struct var_value_iterator_t *THIS);

/**
 * Returns the indices associated with the array element the iterator points
 * to. If the iterator is the END iterator, NULL is returned.
 *
 * @param THIS
 *  The iterator.
 * @return
 *  The indices to use in order to select the value the iterator is pointing
 *  to.
 */
unsigned *
var_value_iterator_get_var_indices(struct var_value_iterator_t *THIS);

/**
 * Returns the value the iterator points to without advancing the iterator to
 * the next element. If the iterator is the END iterator, NULL is returned.
 *
 * @param THIS
 *  The iterator.
 * @return
 *  The value the iterator points to.
 */
struct var_value_t *
var_value_iterator_get(struct var_value_iterator_t *THIS);

/**
 * Advances the iterator to the next element. If there is no next element, the
 * iterator becomes the END iterator. If the iterator is already the END
 * iterator, nothing happens.
 *
 * @param THIS
 *  The iterator.
 * @return
 *  TRUE if the iterator is not the END iterator and var_value_iterator_get()
 *  will succeed the next time it is called on this iterator, FALSE otherwise.
 */
BOOL
var_value_iterator_next(struct var_value_iterator_t *THIS);

/**
 * Destroys an iterator. The iterator need not be the END iterator.
 *
 * @param THIS
 *  The iterator to destroy.
 */
void
var_value_iterator_destroy(struct var_value_iterator_t *THIS);

/**
 * @}
 */

/**
 * @name Type checking.
 * @{
 */

/**
 * Checks the values of all variables against their types.
 *
 * @return
 *  TRUE if checking was successful, FALSE otherwise.
 */
BOOL
var_check_values(void);

/**
 * @}
 */

/**
 * @name Module initialization and finalization.
 * @{
 */

/**
 * Initializes this module.
 */
void
var_init_module(void);

/**
 * Finalizes this module.
 */
void
var_fini_module(void);

/**
 * @}
 */

/**
 * @name Wrapper functions.
 * @{
 */

/**
 * Convenience function for reading from a global variable slot. Uses
 * var_instance_create_from_string() and var_instance_get_value().
 *
 * @param name
 *  The name of the variable slot.
 * @return
 *  The value of the variable slot or NULL if the accessed variable slot has not
 *  been set or another error occurs.
 */
char const *
var_read(char const *name);

/**
 * Convenience function for reading from a global variable slot. Uses
 * var_instance_create_from_string() and var_instance_try_get_value().
 *
 * @param name
 *  The name of the variable slot.
 * @return
 *  The value of the variable slot or NULL if the accessed variable slot has not
 *  been set or another error occurs.
 */
char const *
var_try_read(char const *name);

/**
 * Convenience function for writing to a global variable slot. Uses
 * var_instance_create_from_string() and var_instance_set_value() or
 * var_instance_set_and_check_value(), respectively
 *
 * @param name
 *  The name of the variable slot.
 * @param value
 *  The value.
 * @param mode
 *  The access mode, expressed by some VAR_ACCESS_* constant.
 * @param priority
 *  The priority of the value being set. This is some arbitrary value that is
 *  stored together with the variable slots and that, together with the access
 *  mode, controls whether a slot can be changed or not.
 * @param check
 *  TRUE if the value is to be checked, FALSE otherwise.
 * @return
 *  TRUE if successful, FALSE otherwise.
 */
BOOL
var_write(
    char const *name,
    char const *value,
    int mode,
    int priority,
    BOOL check
);

/**
 * @}
 */

#endif
