/*****************************************************************************
 *  @file var/core.c
 *  Functions for processing variables.
 *
 *  Copyright (c) 2000-2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: core.c 34688 2014-11-07 12:56:15Z florian $
 *****************************************************************************
 */

#include <libmkfli4l/defs.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <expression/expr_literal.h>
#include <expression/expr_logical_or.h>
#include <stdlib.h>
#include <string.h>
#include <regex.h>
#include "core.h"
#include "scope.h"

#define MODULE "var::core"

/**
 * Name of the variable type used for array bound variables.
 */
#define VAR_TYPE_ARRAY_BOUND VARTYPE_PREDEFINED_UNSIGNED_INTEGER
/**
 * The suffix used for names of array bound variables.
 */
#define VAR_N_SUFFIX 'N'
/**
 * The default value for array bound variables.
 */
#define VAR_N_DEF_VALUE "0"
/**
 * Forbidden characters in a variable name.
 */
#define VAR_NAME_FORBIDDEN_CHARACTERS "()[]{}.?*+^$"

/**
 * Type used for typing array bounds.
 *
 * @note
 *  Don't use it directly, call var_get_array_bound_type() instead which performs
 *  lazy initialization.
 */
static struct vartype_t *var_array_bound_type = NULL;

/**
 * Returns the type used for typing array bounds.
 */
static struct vartype_t *
var_get_array_bound_type(void)
{
    if (!var_array_bound_type) {
        var_array_bound_type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
        if (!var_array_bound_type) {
            fatal_exit(
                MODULE ": BUG in var_get_array_bound_type(): Type '%s' needed for array bound variables is missing.\n",
                VARTYPE_PREDEFINED_UNSIGNED_INTEGER
            );
        }
    }
    return var_array_bound_type;
}

/**
 * Registers a new array variable at an array bound variable which the array
 * depends upon.
 *
 * @param THIS
 *  The array bound variable.
 * @param client
 *  The new array variable.
 */
static void
var_register_client(struct var_t *THIS, struct var_t *client)
{
    THIS->var_n_clients = (struct var_t **) safe_realloc(THIS->var_n_clients,
            (THIS->num_var_n_clients + 1) * sizeof(struct var_t *));
    THIS->var_n_clients[THIS->num_var_n_clients] = client;
    ++THIS->num_var_n_clients;
}

/**
 * Unregisters an array variable being destroyed at an array bound variable
 * which the array depends upon.
 *
 * @param THIS
 *  The array bound variable.
 * @param client
 *  The array variable being destroyed.
 */
static void
var_unregister_client(struct var_t *THIS, struct var_t *client)
{
    int i;
    for (i = 0; i < THIS->num_var_n_clients; ++i) {
        if (THIS->var_n_clients[i] == client) {
            --THIS->num_var_n_clients;
            memcpy(
                &THIS->var_n_clients[i],
                &THIS->var_n_clients[i + 1],
                (THIS->num_var_n_clients - i) * sizeof(struct var_t *)
            );
            THIS->var_n_clients = (struct var_t **) safe_realloc(
                THIS->var_n_clients,
                THIS->num_var_n_clients * sizeof(struct var_t *)
            );
            break;
        }
    }
}

/**
 * Unregisters an array bound variable being destroyed at an array variable
 * depending on it. This happens at module finalization only.
 *
 * @param THIS
 *  The array variable.
 * @param var_n
 *  The array bound variable being destroyed.
 */
static void
var_unregister_var_n(struct var_t *THIS, struct var_t *var_n)
{
    int i;
    for (i = 0; i < THIS->level; ++i) {
        if (THIS->var_n[i] == var_n) {
            THIS->var_n[i] = NULL;
        }
    }
}

/**
 * Finds or creates an array bound variable storing one or more array bounds
 * (according to its level). A newly and implicitly created array bound
 * variable has the following properties: It is persistent iff the referencing
 * array variable is, it is not optional, it has the default value
 * VAR_N_DEF_VALUE (which corresponds to zero), and its enabling condition is
 * FALSE. The latter ensures that an implicitly created array bound variable is
 * enabled only if at least one array using it is enabled.
 *
 * @param scope
 *  The scope where to add the array bound variable to.
 * @param name
 *  The name of the array bound variable.
 * @param location
 *  The location of the variable definition.
 * @param props
 *  The properties to use. Parts of it are used when creating a new array bound
 *  variable.
 * @param created
 *  Is set to TRUE if the array bound has been created.
 * @return
 *  A pointer to the associated var_t object or NULL if the variable could
 *  neither be found nor created (e.g. because a variable with the same name
 *  but different properties already exists).
 */
static struct var_t *
var_get_or_add_array_bound(
    struct var_scope_t *scope,
    char const *name,
    struct location_t const *location,
    struct var_properties_t const *props,
    BOOL *created
)
{
    struct var_t *var_n = var_scope_try_get(scope, name, TRUE);
    if (var_n) {
        *created = FALSE;
        if (var_n->type != var_get_array_bound_type()) {
            log_error(
                MODULE ": Array bound variable '%s' has wrong type '%s', must be '%s'.\n",
                name, vartype_get_name(var_n->type),
                vartype_get_name(var_get_array_bound_type())
            );
            return NULL;
        }
        else {
            if (!var_n->props->def_value) {
                var_properties_set_default_value(var_n->props, VAR_N_DEF_VALUE);
            }
        }
    }
    else {
        struct var_properties_t *new_props = var_properties_create();
        new_props->persistent = props->persistent;
        new_props->optional = FALSE;
        var_properties_set_default_value(new_props, VAR_N_DEF_VALUE);
        var_properties_set_condition(new_props,
            expr_literal_create_boolean(scope, FALSE));
        *created = TRUE;

        var_n = var_add(
            scope,
            name,
            var_get_array_bound_type(),
            location_clone(location),
            new_props
        );
    }

    return var_n;
}

/**
 * Creates a variable. If an array variable is to be created, the associated
 * array bound variables are also created if not already done so. For example,
 * creating the variable VBOX_USER_%_SECTION_%_SECDEF_% also creates the
 * array bound variables VBOX_USER_N, VBOX_USER_%_SECTION_N, and
 * VBOX_USER_%_SECTION_%_SECDEF_N (of which the last two ones are array
 * variables, too).
 *
 * @param scope
 *  The scope of the variable. If needed array bound variables are missing,
 *  they are added to this scope.
 * @param name
 *  The name of the variable.
 * @param type
 *  The type of the variable.
 * @param location
 *  The location of the variable definition.
 * @param props
 *  The properties to use. Ownership is passed to the variable.
 * @return
 *  A pointer to the new var_t object or NULL if some dependent array bound
 *  variable could neither be found nor added.
 */
static struct var_t *
var_create(
    struct var_scope_t *scope,
    char const *name,
    struct vartype_t *type,
    struct location_t *location,
    struct var_properties_t *props
)
{
    char const *p;
    char const *q;
    struct var_t *var;
    char *regex = NULL;
    BOOL comp_result;
    BOOL *created = NULL;
    int level;

    var = (struct var_t *) safe_malloc(sizeof(struct var_t));
    var->type = type;
    var->location = location;
    var->level = 0;

    var->var_n = NULL;
    var->num_var_n_clients = 0;
    var->var_n_clients = NULL;

    p = name;
    regex = strcat_save(regex, "^");
    while ((q = strchr(p, VAR_PLACEHOLDER)) != NULL) {
        char *var_n_name = substrsave(name, q + 1);
        var_n_name[q - name] = VAR_N_SUFFIX;
        var->var_n = (struct var_t **)
            safe_realloc(var->var_n, (var->level + 1) * sizeof(struct var_t *));
        created = (BOOL *)
            safe_realloc(created, (var->level + 1) * sizeof(BOOL));

        if ((var->var_n[var->level] = var_get_or_add_array_bound(scope,
                var_n_name, location, props, &created[var->level])) == NULL) {
            int i;
            free(regex);
            free(var_n_name);
            for (i = 0; i < var->level; ++i) {
                if (created[i]) {
                    var_destroy(var->var_n[i]);
                }
            }
            free(created);
            free(var->var_n);
            free(var);
            return NULL;
        }

        var_n_name[q - name] = '\0';
        regex = strcat_save(regex, var_n_name + (p - name));
        regex = strcat_save(regex, VAR_PLACEHOLDER_REGEX);
        free(var_n_name);

        ++var->level;
        p = q + 1;
    }
    if (*p) {
        regex = strcat_save(regex, p);
    }
    regex = strcat_save(regex, "$");

    if (var->level > 0) {
        comp_result = regcomp(&var->tmpl_regex, regex, REG_EXTENDED | REG_NEWLINE);
        if (comp_result != 0) {
            char *const locstr = location_toString(location);
            size_t const errbuf_size
                = regerror(comp_result, &var->tmpl_regex, NULL, 0);
            char *const errbuf = (char *) safe_malloc(errbuf_size);
            regerror(comp_result, &var->tmpl_regex, errbuf, errbuf_size);

            fatal_exit(
                MODULE ": BUG in var_create(): Compiling regular expression '%s' for matching instances of variable '%s' at %s failed: %s.\n",
                regex,
                name,
                locstr,
                errbuf
            );
        }
    }
    free(regex);
    free(created);

    var_slot_init(&var->slot, 0, var->level);
    var->name = strsave(name);
    var->props = props;

    for (level = 0; level < var->level; ++level) {
        var_register_client(var->var_n[level], var);
    }

    return var;
}

struct var_t *
var_try_get(char const *name)
{
    return var_scope_try_get(var_get_global_scope(), name, TRUE);
}

struct var_t *
var_get(char const *name)
{
    struct var_t *var = var_try_get(name);
    if (var) {
        return var;
    }
    else {
        log_error(MODULE ": Variable '%s' does not exist.\n", name);
        return NULL;
    }
}

void
var_destroy(struct var_t *THIS)
{
    int i;
    for (i = 0; i < THIS->level; ++i) {
        if (THIS->var_n[i]) {
            var_unregister_client(THIS->var_n[i], THIS);
        }
    }

    for (i = 0; i < THIS->num_var_n_clients; ++i) {
        var_unregister_var_n(THIS->var_n_clients[i], THIS);
    }

    var_slot_destroy(&THIS->slot, 0, THIS->level);
    var_properties_destroy(THIS->props);
    location_destroy(THIS->location);
    if (THIS->level > 0) {
        regfree(&THIS->tmpl_regex);
    }
    free(THIS->var_n);
    free(THIS->var_n_clients);
    free(THIS->name);
    free(THIS);
}

struct var_t *
var_add(
    struct var_scope_t *scope,
    char const *name,
    struct vartype_t *type,
    struct location_t *location,
    struct var_properties_t *props
)
{
    struct var_t *var;

    if (!name) {
        char *locstr = location_toString(location);
        fatal_exit(
            MODULE ": BUG in var_add(): Name of variable at %s may not be NULL.\n",
            locstr
        );
    }

    if (!type) {
        char *locstr = location_toString(location);
        fatal_exit(
            MODULE ": BUG in var_add(): Type for variable '%s' at %s may not be NULL.\n",
            name, locstr
        );
    }

    if (!props) {
        char *locstr = location_toString(location);
        fatal_exit(
            MODULE ": BUG in var_add(): Properties for variable '%s' at %s may not be NULL.\n",
            name, locstr
        );
    }

    if (!var_properties_check(props, name)) {
        return NULL;
    }

    if (strcspn(name, VAR_NAME_FORBIDDEN_CHARACTERS) != strlen(name)) {
        char *locstr = location_toString(location);
        log_error(
            MODULE ": Variable '%s' at %s may not contain one of the following characters in its name: '%s'.\n",
            name,
            locstr,
            VAR_NAME_FORBIDDEN_CHARACTERS
        );
        free(locstr);
        return NULL;
    }

    var = var_create(scope, name, type, location, props);
    if (!var) {
        char *locstr = location_toString(location);
        log_error(
            MODULE ": Variable '%s' at %s could not be created.\n",
            name, locstr
        );
        free(locstr);
        return NULL;
    }

    if (!var_scope_add(scope, var)) {
        return NULL;
    }
    else {
        return var;
    }
}

char const *
var_get_name(struct var_t const *THIS)
{
    return THIS->name;
}

struct vartype_t *
var_get_type(struct var_t const *THIS)
{
    return THIS->type;
}

int
var_get_level(struct var_t const *THIS)
{
    return THIS->level;
}

struct var_properties_t const *
var_get_properties(struct var_t const *THIS)
{
    return THIS->props;
}

BOOL
var_set_properties(struct var_t *THIS, struct var_properties_t *props)
{
#define PROP_ERRMSG MODULE ": Change of property '%s' of variable '%s' at %s is not allowed.\n"

    if (THIS->props->persistent != props->persistent) {
        char *locstr = location_toString(THIS->location);
        log_error(PROP_ERRMSG, "persistent", THIS->name, locstr);
        free(locstr);
        return FALSE;
    }
    if (THIS->props->optional != props->optional) {
        char *locstr = location_toString(THIS->location);
        log_error(PROP_ERRMSG, "optional", THIS->name, locstr);
        free(locstr);
        return FALSE;
    }

    var_properties_destroy(THIS->props);
    THIS->props = props;
    return TRUE;

#undef PROP_ERRMSG
}

struct expr_t *
var_get_condition(struct var_t const *THIS)
{
    int i;
    struct expr_t *expr_false
        = expr_literal_create_boolean(var_get_global_scope(), FALSE);
    struct expr_t *expr_true
        = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    struct expr_t *result = expr_clone(THIS->props->condition);
    struct var_scope_t *global_scope = var_get_global_scope();

    for (i = 0; i < THIS->num_var_n_clients; ++i) {
        struct expr_t *cond = var_get_condition(THIS->var_n_clients[i]);
        if (expr_equal(result, expr_false) || expr_equal(cond, expr_true)) {
            expr_destroy(result);
            result = cond;
        }
        else
        if (expr_equal(result, expr_true) || expr_equal(cond, expr_false)) {
            expr_destroy(cond);
        }
        else {
            result = expr_logical_or_create(global_scope, result, cond);
        }
    }

    expr_destroy(expr_true);
    expr_destroy(expr_false);
    return result;
}

struct location_t const *
var_get_location(struct var_t const *THIS)
{
    return THIS->location;
}

BOOL
var_is_enabled(struct var_t const *THIS)
{
    BOOL enabled;
    char *error_msg;
    struct expr_t *cond = var_get_condition(THIS);

    if (!expr_get_boolean_value(cond, &enabled, &error_msg)) {
        free(error_msg);
        enabled = FALSE;
    }

    expr_destroy(cond);
    return enabled;
}
