/*****************************************************************************
 *  @file var/core.h
 *  Functions for processing variables (private interface).
 *
 *  Copyright (c) 2000-2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: core.h 34688 2014-11-07 12:56:15Z florian $
 *****************************************************************************
 */

#ifndef MKFLI4L_VAR_CORE_H_
#define MKFLI4L_VAR_CORE_H_

#include <libmkfli4l/defs.h>
#include <vartype.h>
#include <var.h>
#include <regex.h>
#include "slot.h"

/**
 * The character used as a placeholder in names of array variables.
 */
#define VAR_PLACEHOLDER '%'
/**
 * The regular expression used to match instantiated array slots with the
 * array definition.
 */
#define VAR_PLACEHOLDER_REGEX "([1-9][0-9]*)"

/**
 * Represents a variable definition.
 */
struct var_t {
    /**
     * The variable name.
     */
    char *name;
    /**
     * The variable type.
     */
    struct vartype_t *type;
    /**
     * Number of necessary indices. Level zero denotes a simple variable, level
     * "n" (with n > 0) denotes an n-dimensional array.
     */
    int level;
    /**
     * Array of references to array bound variables for each variable level.
     * Each array bound variable contains the number of entries for the
     * corresponding level. This member is NULL if the variable level is zero.
     * If e.g. a variable A_%_% (level 2) is given, the first reference points
     * to A_N (level 0) and the second points to A_%_N (level 1).
     */
    struct var_t **var_n;
    /**
     * Number of array variables referencing this array bound variable.
     */
    int num_var_n_clients;
    /**
     * Array of references to arrays referencing this array bound variable.
     */
    struct var_t **var_n_clients;
    /**
     * Slot of level zero. Contains the variable's value if the variable's
     * level is zero. Otherwise, it points to an array of slots containing
     * slots of the next higher level (level 1) and so on.
     */
    union varslot_t slot;
    /**
     * The location of the variable definition.
     */
    struct location_t *location;
    /**
     * The properties of the variable.
     */
    struct var_properties_t *props;
    /**
     * The compiled regular expression for template matching.
     */
    regex_t tmpl_regex;
};

/**
 * Destroys a variable.
 *
 * @param THIS
 *  The variable to destroy.
 */
void
var_destroy(struct var_t *THIS);

#endif
