/*****************************************************************************
 *  @file var/inst.c
 *  Functions for processing variable instantiations.
 *
 *  Copyright (c) 2000-2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: inst.c 34688 2014-11-07 12:56:15Z florian $
 *****************************************************************************
 */

#include <libmkfli4l/defs.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <container/hashmap.h>
#include <stdlib.h>
#include <string.h>
#include <regex.h>
#include "inst.h"
#include "core.h"
#include "scope.h"

#define MODULE "var::inst"

struct var_instance_t *
var_instance_create(
    struct var_t *THIS,
    int num_indices,
    unsigned const *indices
)
{
    struct var_instance_t *instance = (struct var_instance_t *)
        safe_malloc(sizeof(struct var_instance_t));
    instance->var = THIS;
    instance->num_indices = num_indices;
    instance->indices
        = (unsigned *) safe_malloc(num_indices * sizeof(unsigned));
    if (indices) {
        memcpy(instance->indices, indices, num_indices * sizeof(unsigned));
    }
    else {
        memset(instance->indices, 0, num_indices * sizeof(unsigned));
    }
    return instance;
}

void
var_instance_destroy(struct var_instance_t *THIS)
{
    free(THIS->indices);
    free(THIS);
}

char *
var_instance_to_string(struct var_instance_t const *THIS)
{
    char *result = strsave("");
    char const *p = THIS->var->name;
    char const *q;
    int level = 0;
    unsigned const *indices = THIS->indices;

    while ((q = strchr(p, VAR_PLACEHOLDER)) != NULL) {
        char *s;

        s = substrsave(p, q);
        result = strcat_save(result, s);
        free(s);

        if (level < THIS->num_indices) {
            s = safe_sprintf("%u", *indices + 1);
            result = strcat_save(result, s);
            free(s);
            ++indices;
        }
        else {
            char placeholder[2];
            placeholder[0] = VAR_PLACEHOLDER;
            placeholder[1] = '\0';
            result = strcat_save(result, placeholder);
        }

        ++level;
        p = q + 1;
    }

    return strcat_save(result, p);
}

/**
 * Tries to deduce instantiation arguments.
 *
 * @param entry
 *  The variable to check.
 * @param data
 *  The instantiation request.
 * @return
 *  TRUE if no instantiation possible with given variable, FALSE otherwise. In
 *  the latter case, the "result" member in the instantiation request is filled
 *  with the variable and the deduced indices.
 */
static BOOL
var_try_deduce_instantiation(void *entry, void *data)
{
    struct var_t *var = (struct var_t *) entry;
    struct var_instantiation_request_t *req
        = (struct var_instantiation_request_t *) data;
    regmatch_t *matches;

    if (var->level == 0) {
        if (strcmp(var->name, req->name) == 0) {
            req->instance = var_instance_create(var, 0, NULL);
            return FALSE;
        }
        else {
            return TRUE;
        }
    }

    matches = (regmatch_t *)
        safe_malloc((var->level + 1) * sizeof(regmatch_t));

    if (regexec(&var->tmpl_regex, req->name, var->level + 1, matches, 0) == 0) {
        int level;
        regmatch_t *match;
        unsigned *indices
            = (unsigned *) safe_malloc(var->level * sizeof(unsigned));
        unsigned *index;

        for (
            match = matches + 1, index = indices, level = 0;
            level < var->level;
            ++match, ++index, ++level
        ) {
            if (match->rm_so != -1 && match->rm_eo != -1 && match->rm_so != match->rm_eo) {
                char *i = substrsave(req->name + match->rm_so, req->name + match->rm_eo);
                *index = atoi(i) - 1;
                free(i);
            }
            else {
                fatal_exit(
                    MODULE ": BUG in var_try_deduce_instantiation(): Matching '%s' with '%s' failed.\n",
                    req->name,
                    var->name
                );
            }
        }

        req->instance = var_instance_create(var, var->level, indices);
        free(indices);
        free(matches);
        return FALSE;
    }
    else {
        free(matches);
        return TRUE;
    }
}

struct var_instance_t *
var_instance_create_from_string(struct var_scope_t *scope, char const *name)
{
    struct var_t *var = var_scope_try_get(scope, name, TRUE);
    if (var) {
        return var_instance_create(var, 0, NULL);
    }
    else {
        struct var_instantiation_request_t *req
            = (struct var_instantiation_request_t *)
                safe_malloc(sizeof(struct var_instantiation_request_t));

        req->name = name;
        req->instance = NULL;

        if (var_scope_foreach(scope, TRUE, &var_try_deduce_instantiation, req)) {
            free(req);
            return NULL;
        }
        else {
            struct var_instance_t *const instance = req->instance;
            free(req);
            return instance;
        }
    }
}
