/*****************************************************************************
 *  @file var/slot_iter.c
 *  Functions for processing variable slot iterators.
 *
 *  Copyright (c) 2000-2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: iter.c 34688 2014-11-07 12:56:15Z florian $
 *****************************************************************************
 */

#include <libmkfli4l/defs.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>
#include "core.h"
#include "slot.h"
#include "slot_read.h"
#include "iter.h"

#define MODULE "var::iter"

static struct varslot_iterator_t *
var_slot_iterator_create(union varslot_t *THIS, struct var_t *var, int level)
{
    struct varslot_iterator_t *const iter = (struct varslot_iterator_t *)
        safe_malloc(sizeof(struct varslot_iterator_t));
    iter->parent = NULL;
    iter->var = var;
    iter->slot = THIS;
    iter->level = level;
    iter->i = 0;
    return iter;
}

static void
var_slot_iterator_destroy(struct varslot_iterator_t *THIS)
{
    free(THIS);
}

static struct varslot_iterator_t *
var_slot_iterator_next(struct varslot_iterator_t *THIS)
{
    if (THIS->level == THIS->var->level
            || THIS->i == THIS->slot->array.num_slots) {
        struct varslot_iterator_t *const parent = THIS->parent;
        var_slot_iterator_destroy(THIS);
        if (parent) {
            ++parent->i;
        }
        return parent;
    }
    else {
        struct varslot_iterator_t *const next = (struct varslot_iterator_t *)
            safe_malloc(sizeof(struct varslot_iterator_t));
        next->parent = THIS;
        next->slot = &THIS->slot->array.slots[THIS->i];
        next->var = THIS->var;
        next->level = THIS->level + 1;
        next->i = 0;
        return next;
    }
}

struct var_value_iterator_t *
var_value_iterator_create(struct var_instance_t *THIS)
{
    union varslot_t *const slot = var_instance_get_slot_for_reading(THIS, TRUE);
    struct var_value_iterator_t *const iter = (struct var_value_iterator_t *)
        safe_malloc(sizeof(struct var_value_iterator_t));
    iter->var = THIS->var;
    if (slot) {
        iter->iter = var_slot_iterator_create(slot, THIS->var, THIS->num_indices);
        iter->level = THIS->num_indices;
        var_value_iterator_next(iter);
    }
    else {
        iter->iter = NULL;
    }
    return iter;
}

void
var_value_iterator_destroy(struct var_value_iterator_t *THIS)
{
    struct varslot_iterator_t *iter = THIS->iter;
    while (iter) {
        struct varslot_iterator_t *const parent = iter->parent;
        var_slot_iterator_destroy(iter);
        iter = parent;
    }

    free(THIS);
}

struct var_t *
var_value_iterator_get_var(struct var_value_iterator_t *THIS)
{
    return THIS->var;
}

unsigned *
var_value_iterator_get_var_indices(struct var_value_iterator_t *THIS)
{
    unsigned *indices;
    struct varslot_iterator_t *iter;

    if (!THIS->iter) {
        return NULL;
    }

    iter = THIS->iter->parent;
    if (!iter) {
        return NULL;
    }

    indices = (unsigned *)
        safe_malloc((iter->level - THIS->level + 1) * sizeof(unsigned));

    while (iter) {
        indices[iter->level - THIS->level] = iter->i;
        iter = iter->parent;
    }

    return indices;
}

struct var_value_t *
var_value_iterator_get(struct var_value_iterator_t *THIS)
{
    if (THIS->iter) {
        struct var_value_t *result = (struct var_value_t *)
            safe_malloc(sizeof(struct var_value_t));
        result->value = var_slot_get_value(THIS->iter->slot, THIS->iter->var);
        result->priority = THIS->iter->slot->entry.priority;
        return result;
    }
    else {
        return NULL;
    }
}

BOOL
var_value_iterator_next(struct var_value_iterator_t *THIS)
{
    if (THIS->iter) {
        int const maxlevel = THIS->var->level;
        char const *value = NULL;
        while (!value) {
            do {
                THIS->iter = var_slot_iterator_next(THIS->iter);
            } while (THIS->iter && THIS->iter->level != maxlevel);
            if (THIS->iter) {
                value = var_slot_get_value(THIS->iter->slot, THIS->iter->var);
            }
            else {
                return FALSE;
            }
        }

        return TRUE;
    }
    else {
        return FALSE;
    }
}
