/*****************************************************************************
 *  @file var/props.c
 *  Functions for processing variable properties.
 *
 *  Copyright (c) 2000-2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: props.c 34688 2014-11-07 12:56:15Z florian $
 *****************************************************************************
 */

#include <libmkfli4l/defs.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <expression/expr_literal.h>
#include <stdlib.h>
#include <string.h>
#include "core.h"

#define MODULE "var::props"

struct var_properties_t *
var_properties_create(void)
{
    struct var_properties_t *result = (struct var_properties_t *)
        safe_malloc(sizeof(struct var_properties_t));

    result->comment = strsave("");
    result->persistent = TRUE;
    result->optional = FALSE;
    result->condition
        = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    result->def_value = NULL;
    return result;
}

struct var_properties_t *
var_properties_clone(struct var_properties_t const *THIS)
{
    struct var_properties_t *result = (struct var_properties_t *)
        safe_malloc(sizeof(struct var_properties_t));

    result->comment = strsave(THIS->comment);
    result->persistent = THIS->persistent;
    result->optional = THIS->optional;
    result->condition = expr_clone(THIS->condition);
    result->def_value = THIS->def_value ? strsave(THIS->def_value) : NULL;
    return result;
}

void
var_properties_set_comment(
    struct var_properties_t *THIS,
    char const *comment
)
{
    free(THIS->comment);
    THIS->comment = strsave(comment ? comment : "");
}

void
var_properties_set_default_value(
    struct var_properties_t *THIS,
    char const *def_value
)
{
    free(THIS->def_value);
    THIS->def_value = strsave(def_value);
}

void
var_properties_set_condition(
    struct var_properties_t *THIS,
    struct expr_t *condition
)
{
    expr_destroy(THIS->condition);
    THIS->condition = condition;
}

BOOL
var_properties_check(struct var_properties_t const *THIS, char const *name)
{
    if (!THIS->persistent) {
        if (THIS->optional) {
            log_error(
                MODULE ": Transient variable '%s' is optional, which is not supported.\n",
                name
            );
            return FALSE;
        }
    }
    else {
        if (THIS->optional && THIS->def_value) {
            log_error(
                MODULE ": Persistent variable '%s' is optional and has a default value, which is not supported.\n",
                name
            );
            return FALSE;
        }
    }

    return TRUE;
}

void
var_properties_destroy(struct var_properties_t *THIS)
{
    expr_destroy(THIS->condition);
    free(THIS->def_value);
    free(THIS->comment);
    free(THIS);
}
