/*****************************************************************************
 *  @file var/scope.c
 *  Functions for processing variable scopes.
 *
 *  Copyright (c) 2000-2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: scope.c 34688 2014-11-07 12:56:15Z florian $
 *****************************************************************************
 */

#include <libmkfli4l/defs.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>
#include "scope.h"
#include "core.h"

#define MODULE "var::scope"

static struct var_scope_t *global_scope = NULL;

/**
 * Destroys a variable.
 *
 * @param entry
 *  The variable to destroy. It is of type void* in order to be able to use
 *  this function as a hash map object destructor.
 */
static void
var_destroy_action(void *entry)
{
    struct var_t *const THIS = (struct var_t *) entry;
    var_destroy(THIS);
}

struct var_scope_t *
var_get_global_scope(void)
{
    return global_scope;
}

struct var_scope_t *
var_scope_create(struct location_t *location, struct var_scope_t *parent)
{
    struct var_scope_t *scope = (struct var_scope_t *)
        safe_malloc(sizeof(struct var_scope_t));
    scope->location = location;
    scope->parent = parent;
    scope->vars = hashmap_create(
        offsetof(struct var_t, name),
        1024,
        hash_ci_string,
        compare_ci_string_keys,
        var_destroy_action
    );
    return scope;
}

struct var_t *
var_scope_try_get(
    struct var_scope_t *THIS,
    char const *name,
    BOOL recursive
)
{
    struct var_t *var = (struct var_t *) hashmap_get(THIS->vars, &name);
    if (var) {
        return var;
    }
    else {
        return recursive && THIS->parent
            ? var_scope_try_get(THIS->parent, name, recursive) : NULL;
    }
}

struct var_t *
var_scope_get(
    struct var_scope_t *THIS,
    char const *name,
    BOOL recursive
)
{
    struct var_t *var = var_scope_try_get(THIS, name, recursive);
    if (var) {
        return var;
    }
    else {
        char *locstr = location_toString(THIS->location);
        log_error(
            MODULE ": Variable '%s' does not exist in scope starting at %s.\n",
            name,
            locstr
        );
        free(locstr);
        return NULL;
    }
}

BOOL
var_scope_add(struct var_scope_t *THIS, struct var_t *var)
{
    struct var_t *var_old = var_scope_try_get(THIS, var->name, FALSE);
    if (var_old) {
        char *locstr = location_toString(var->location);
        char *locstrOrig = location_toString(var_old->location);
        log_error(
            MODULE ": Redefinition of variable '%s' at %s is not allowed. The original definition is at %s.\n",
            var->name, locstr, locstrOrig
        );
        free(locstrOrig);
        free(locstr);
        var_destroy(var);
        return FALSE;
    }

    var_old = var_scope_try_get(THIS, var->name, TRUE);
    if (var_old) {
        char *locstr = location_toString(var->location);
        char *locstrOrig = location_toString(var_old->location);
        log_info(
            LOG_VAR,
            MODULE ": Warning: Definition of variable '%s' at %s overrides definition at %s.\n",
            var->name, locstr, locstrOrig
        );
        free(locstrOrig);
        free(locstr);
    }

    hashmap_put(THIS->vars, var);
    return TRUE;
}

BOOL
var_scope_foreach(
    struct var_scope_t *THIS,
    BOOL recursive,
    hashmap_visitor_t visitor,
    void *data
)
{
    if (!hashmap_foreach(THIS->vars, visitor, data)) {
        return FALSE;
    }
    else if (recursive && THIS->parent) {
        return var_scope_foreach(THIS->parent, recursive, visitor, data);
    }
    else {
        return TRUE;
    }
}

void
var_scope_destroy(struct var_scope_t *THIS)
{
    hashmap_destroy(THIS->vars);
    location_destroy(THIS->location);
    free(THIS);
}

void
var_scope_init(void)
{
    global_scope = var_scope_create(
        location_create_internal(PACKAGE_FILE_TYPE_CONFIG),
        NULL
    );
}

void
var_scope_fini(void)
{
    var_scope_destroy(global_scope);
    global_scope = NULL;
}
