/*****************************************************************************
 *  @file var/slot_check.c
 *  Functions for checking variable slots.
 *
 *  Copyright (c) 2000-2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: slot_check.c 34688 2014-11-07 12:56:15Z florian $
 *****************************************************************************
 */

#include <libmkfli4l/defs.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <container/hashmap.h>
#include <stdlib.h>
#include <string.h>
#include "core.h"
#include "slot.h"
#include "scope.h"
#include "expression/expr_literal.h"

#define MODULE "var::slot_check"

/**
 * Checks all values of a variable.
 *
 * @param entry
 *  The variable.
 * @param data
 *  A pointer to a boolean variable which is set to FALSE if at least one value
 *  of this variable is not type conformant. Set it to TRUE before calling this
 *  function.
 * @return
 *  Always TRUE in order to check all variables.
 */
static BOOL
var_check_variable(void *entry, void *data);

/**
 * Checks all values of a slot.
 *
 * @param THIS
 *  The slot.
 * @param inst
 *  The variable instance this slot belongs to.
 * @param level
 *  The slot level.
 * @return
 *  TRUE if all values of this slot have been successfully checked, FALSE
 *  otherwise.
 */
static BOOL
var_check_slot(union varslot_t *THIS, struct var_instance_t *inst, int level);

/**
 * Checks a single non-array slot.
 *
 * @param THIS
 *  The slot.
 * @param inst
 *  The variable instance this slot belongs to.
 * @return
 *  TRUE if the value of this slot has been successfully checked, FALSE
 *  otherwise.
 */
static BOOL
var_check_non_array_slot(union varslot_t *THIS, struct var_instance_t *inst)
{
    char *const inst_name = var_instance_to_string(inst);
    BOOL result;

    log_info(
        LOG_VAR,
        MODULE ": About to check value of variable '%s'.\n",
        inst_name
    );

    if (THIS->entry.checked) {
        log_info(
            LOG_VAR,
            MODULE ": Variable '%s' already checked, skipping.\n",
            inst_name
        );
        result = TRUE;
    }
    else {
        char const *value = var_slot_get_value(THIS, inst->var);
        if (!value) {
            if (!inst->var->props->optional) {
                log_error(
                    MODULE ": Value of mandatory variable '%s' has not been set.\n",
                    inst_name
                );
                result = FALSE;
            }
            else {
                log_info(
                    LOG_VAR,
                    MODULE ": Value of optional variable '%s' is missing, skipping type check.\n",
                    inst_name
                );
                result = TRUE;
            }
        }
        else {
            if (vartype_match(inst->var->type, value)) {
                log_info(
                    LOG_VAR,
                    MODULE ": Value '%s' of variable '%s' successfully checked.\n",
                    value,
                    inst_name
                );
                THIS->entry.checked = TRUE;
                result = TRUE;
            }
            else {
                log_error(
                    MODULE ": Value '%s' of variable '%s' is not properly typed.\n",
                    value,
                    inst_name
                );
                result = FALSE;
            }
        }
    }

    free(inst_name);
    return result;
}

/**
 * Checks an array of slots.
 *
 * @param THIS
 *  The slot.
 * @param inst
 *  The variable instance this slot belongs to.
 * @param level
 *  The slot level.
 * @return
 *  TRUE if the value of this slot has been successfully checked, FALSE
 *  otherwise.
 */
static BOOL
var_check_slot_array(
    union varslot_t *THIS,
    struct var_instance_t *inst,
    int level
)
{
    unsigned i;
    int num_bound = 0;
    unsigned num;
    BOOL result = TRUE;

    struct var_t *const var_n = inst->var->var_n[level];
    var_check_variable(var_n, &result);
    if (!result) {
        return FALSE;
    }
    else {
        struct var_instance_t *inst_n
            = var_instance_create(var_n, level, inst->indices);
        struct var_value_t *value = var_instance_get_value(inst_n);
        if (!value || !var_value_convert_to_integer(value, &num_bound)
                || num_bound < 0) {
            fatal_exit(
                MODULE ": BUG in var_check_slot(): Reading array bound variable '%s' failed.\n",
                var_instance_to_string(inst_n)
            );
        }
        else if ((unsigned) num_bound > THIS->array.num_slots) {
            if (!inst->var->props->optional) {
                char *const inst_name = var_instance_to_string(inst_n);
                log_error(
                    MODULE ": Array bound variable '%s' (%d) is larger than the real number of array entries (%d).\n",
                    inst_name,
                    num_bound,
                    THIS->array.num_slots
                );
                free(inst_name);
                result = FALSE;
            }

            num = THIS->array.num_slots;
        }
        else {
            num = (unsigned) num_bound;
        }
        var_value_destroy(value);
        var_instance_destroy(inst_n);

        for (i = 0; i < num; ++i) {
            inst->indices[level] = i;
            if (!var_check_slot(&THIS->array.slots[i], inst, level + 1)) {
                result = FALSE;
            }
        }

        return result;
    }
}

static BOOL
var_check_slot(union varslot_t *THIS, struct var_instance_t *inst, int level)
{
    if (level == inst->var->level) {
        return var_check_non_array_slot(THIS, inst);
    }
    else {
        return var_check_slot_array(THIS, inst, level);
    }
}

static BOOL
var_check_variable(void *entry, void *data)
{
    struct var_t *var = (struct var_t *) entry;

    if (!var_is_enabled(var)) {
        log_info(
            LOG_VAR,
            MODULE ": Variable '%s' not enabled, skipping check.\n",
            var->name
        );
    }
    else {
        struct var_instance_t *inst
            = var_instance_create(var, var->level, NULL);
        BOOL result = var_check_slot(&var->slot, inst, 0);
        var_instance_destroy(inst);
        if (!result) {
            *(BOOL *) data = FALSE;
        }
    }

    return TRUE;
}

BOOL
var_check_values(void)
{
    BOOL result = TRUE;

    log_info(
        LOG_VAR,
        MODULE ": About to check the values of all variables.\n"
    );

    var_scope_foreach(var_get_global_scope(), TRUE, &var_check_variable, &result);
    return result;
}
