/*****************************************************************************
 *  @file var/slot_read.c
 *  Functions for reading from variable slots.
 *
 *  Copyright (c) 2000-2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: slot_read.c 34688 2014-11-07 12:56:15Z florian $
 *****************************************************************************
 */

#include <libmkfli4l/defs.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>
#include <string.h>
#include "core.h"
#include "slot_read.h"

#define MODULE "var::slot_read"

union varslot_t *
var_instance_get_slot_for_reading(struct var_instance_t *THIS, BOOL warn)
{
    int level;

    union varslot_t *slot = &THIS->var->slot;
    unsigned const *index = THIS->indices;

    if (THIS->num_indices > THIS->var->level) {
        char *inst_name = var_instance_to_string(THIS);
        log_error(
            MODULE ": Excessive indices (%d instead of %d) while addressing variable '%s' for reading.\n",
            THIS->num_indices,
            THIS->var->level,
            inst_name
        );
        free(inst_name);
        return NULL;
    }

    for (level = 0; level < THIS->num_indices; ++level, ++index) {
        if (*index >= slot->array.num_slots) {
            if (warn) {
                char *const inst_name = var_instance_to_string(THIS);
                log_error(
                    MODULE ": Index '%u' of dimension %d out of bounds (maximum allowed index is %u) while reading variable '%s'.\n",
                    *index,
                    level + 1,
                    slot->array.num_slots,
                    inst_name
                );
                free(inst_name);
            }
            return NULL;
        }

        slot = &slot->array.slots[*index];
    }

    return slot;
}

/**
 * Returns the value of a variable slot.
 *
 * @param THIS
 *  The variable instance addressing the scalar slot. Excessive or missing
 *  indices are flagged as an error.
 * @param warn
 *  If TRUE, warn about indices which are out of bounds.
 * @return
 *  The value of the variable or NULL if the accessed variable slot has not
 *  been set. Use var_value_destroy() to destroy it if you don't need it
 *  anymore.
 */
static struct var_value_t *
var_instance_do_get_value(struct var_instance_t *THIS, BOOL warn)
{
    union varslot_t *slot;
    char const *value;
    struct var_value_t *result;
    char *inst_name;
    BOOL enabled;

    if (THIS->num_indices < THIS->var->level) {
        inst_name = var_instance_to_string(THIS);
        log_error(
            MODULE ": Missing indices (%d instead of %d) while reading variable '%s'.\n",
            THIS->num_indices,
            THIS->var->level,
            inst_name
        );
        free(inst_name);
        return NULL;
    }

    slot = var_instance_get_slot_for_reading (THIS, warn);
    if (!slot) {
        return NULL;
    }

    inst_name = var_instance_to_string(THIS);
    if (slot->entry.read_active) {
        log_error(
            MODULE ": Dependency cycle detected while reading variable '%s'.\n",
            inst_name
        );
        free(inst_name);
        return NULL;
    }

    slot->entry.read_active = TRUE;
    enabled = var_is_enabled(THIS->var);
    slot->entry.read_active = FALSE;

    if (!enabled) {
        log_info(
            LOG_VAR,
            MODULE ": Reading disabled variable '%s'.\n",
            inst_name
        );
        free(inst_name);
        return NULL;
    }

    value = var_slot_get_value(slot, THIS->var);
    if (!value) {
        if (warn) {
            log_error(MODULE ": Reading unset variable '%s'.\n", inst_name);
        }
        else {
            log_info(
                LOG_VAR,
                MODULE ": Reading unset variable '%s'.\n",
                inst_name
            );
        }
        free(inst_name);
        return NULL;
    }

    log_info(
        LOG_VAR,
        MODULE ": Reading variable '%s' yields '%s' (priority %d, %s).\n",
        inst_name,
        value,
        slot->entry.priority,
        slot->entry.checked ? "checked" : "unchecked"
    );
    free(inst_name);

    result = (struct var_value_t *) safe_malloc(sizeof(struct var_value_t));
    result->value = value;
    result->priority = slot->entry.priority;
    return result;
}

struct var_value_t *
var_instance_get_value(struct var_instance_t *THIS)
{
    return var_instance_do_get_value(THIS, TRUE);
}

struct var_value_t *
var_instance_try_get_value(struct var_instance_t *THIS)
{
    return var_instance_do_get_value(THIS, FALSE);
}
