/*****************************************************************************
 *  @file var/slot_write.c
 *  Functions for writing into variable slots.
 *
 *  Copyright (c) 2000-2001 - Frank Meyer
 *  Copyright (c) 2001-2014 - fli4l-Team <team@fli4l.de>
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: slot_write.c 34688 2014-11-07 12:56:15Z florian $
 *****************************************************************************
 */

#include <libmkfli4l/defs.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>
#include "core.h"
#include "slot_write.h"

#define MODULE "var::slot_write"

/**
 * Resizes an array slot. The caller must ensure that this function is only
 * called on array slots, i.e. on slots with a level smaller than the
 * variable's level.
 *
 * @param THIS
 *  The slot to resize.
 * @param level
 *  The slot level.
 * @param maxlevel
 *  The variable level.
 * @param tot_num_slots
 *  The new total number of entries in this slot.
 */
static void
var_slot_resize(
    union varslot_t *THIS,
    int level,
    int maxlevel,
    unsigned tot_num_slots
)
{
    int const next_level = level + 1;
    unsigned const add_num_slots = tot_num_slots - THIS->array.num_slots;

    union varslot_t *new_slot;
    unsigned i;

    THIS->array.slots = (union varslot_t *) safe_realloc(
        THIS->array.slots, tot_num_slots * sizeof(union varslot_t)
    );

    for (new_slot = THIS->array.slots + THIS->array.num_slots, i = 0;
            i < add_num_slots; ++i, ++new_slot) {
        var_slot_init(new_slot, next_level, maxlevel);
    }

    THIS->array.num_slots = tot_num_slots;
}

/**
 * Returns the priority of a slot. For an entry slot, the priority returned
 * corresponds to the priority stored in the entry. For an array slot, the
 * priority returned is the maximum priority of all subslots.
 *
 * @param THIS
 *  The slot.
 * @param level
 *  The slot level.
 * @param maxlevel
 *  The variable level.
 * @return
 *  The priority of the slot.
 */
static int
var_slot_get_priority(union varslot_t const *THIS, int level, int maxlevel)
{
    if (level == maxlevel) {
        return THIS->entry.priority;
    }
    else {
        int const next_level = level + 1;
        int priority = 0;

        unsigned i;
        for (i = 0; i < THIS->array.num_slots; ++i) {
            int const subprio = var_slot_get_priority(
                &THIS->array.slots[i], next_level, maxlevel
            );
            if (subprio > priority) {
                priority = subprio;
            }
        }

        return priority;
    }
}

/**
 * Updates array bounds if necessary when accessing a slot.
 *
 * @param THIS
 *  The variable instance addressing the array entry.
 * @return
 *  TRUE if the update was successful, FALSE otherwise.
 */
static BOOL
var_update_array_bound(struct var_instance_t *THIS)
{
    int old_bound = 0;
    int const lvl_index = THIS->num_indices - 1;
    int const new_bound = THIS->indices[lvl_index] + 1;
    struct var_t *const var_n = THIS->var->var_n[lvl_index];

    struct var_instance_t *inst_n
        = var_instance_create(var_n, var_n->level, THIS->indices);
    struct var_value_t *value = var_instance_try_get_value(inst_n);
    if (value) {
        if (!var_value_convert_to_integer(value, &old_bound)) {
            char *const inst_name_var_n = var_instance_to_string(inst_n);
            char *const inst_name = var_instance_to_string(THIS);
            log_error(
                MODULE ": Failed to convert value '%s' of array bound variable '%s' to an integer value while accessing variable '%s'.\n",
                value->value,
                inst_name_var_n,
                inst_name
            );
            free(inst_name);
            free(inst_name_var_n);
            var_instance_destroy(inst_n);
            var_value_destroy(value);
            return FALSE;
        }
        else {
            var_value_destroy(value);
        }
    }

    if (old_bound < new_bound) {
        char *str = safe_sprintf("%d", new_bound);
        if (!var_instance_set_and_check_value(inst_n, str, VAR_ACCESS_TENTATIVE,
                VAR_PRIORITY_ARRAY_BOUND)) {
            fatal_exit(
                MODULE ": BUG in var_update_array_bound(): Failed to update array bound '%s' with '%s' while accessing variable '%s'.\n",
                var_instance_to_string(inst_n),
                str,
                var_instance_to_string(THIS)
            );
        }
        free(str);
    }

    var_instance_destroy(inst_n);
    return TRUE;
}

union varslot_t *
var_instance_get_slot_for_writing(
    struct var_instance_t *THIS,
    int mode,
    int priority,
    BOOL *access_failed
)
{
    int level;
    int slot_priority;
    char const *mode_str;
    BOOL ok;
    BOOL resized = FALSE;

    union varslot_t *slot = &THIS->var->slot;
    unsigned const *index = THIS->indices;
    *access_failed = FALSE;

    if (THIS->num_indices > THIS->var->level) {
        char *inst_name = var_instance_to_string(THIS);
        log_error(
            MODULE ": Excessive indices (%d instead of %d) while addressing variable '%s' for writing.\n",
            THIS->num_indices,
            THIS->var->level,
            inst_name
        );
        free(inst_name);
        return NULL;
    }

    for (level = 0; level < THIS->num_indices; ++level, ++index) {
        if (*index >= slot->array.num_slots) {
            var_slot_resize(slot, level, THIS->var->level, (*index) + 1);
            resized = TRUE;
        }

        slot = &slot->array.slots[*index];
    }

    if (resized && !var_update_array_bound(THIS)) {
        return NULL;
    }

    slot_priority
        = var_slot_get_priority(slot, THIS->num_indices, THIS->var->level);

    switch (mode) {
    case VAR_ACCESS_STRONG :
        mode_str = "strong";
        ok = priority > slot_priority;
        break;
    case VAR_ACCESS_WEAK :
        mode_str = "weak";
        ok = priority >= slot_priority;
        break;
    case VAR_ACCESS_TENTATIVE :
        mode_str = "tentative";
        ok = priority >= slot_priority;
        break;
    default :
        fatal_exit(
            MODULE ": BUG in var_instance_get_slot_for_writing(): Unknown access mode %d.\n",
            mode
        );
    }

    if (ok) {
        return slot;
    }
    else {
        char *const inst_name = var_instance_to_string(THIS);
        if (mode != VAR_ACCESS_TENTATIVE) {
            log_error(
                MODULE ": Cannot write to '%s' (priority %d) with priority %d by %s access.\n",
                inst_name, slot_priority, priority, mode_str
            );
        }
        else {
            log_info(
                LOG_VAR,
                MODULE ": Ignoring write request to '%s' (priority %d) with priority %d by %s access.\n",
                inst_name, slot_priority, priority, mode_str
            );
        }
        free(inst_name);
        *access_failed = TRUE;
        return NULL;
    }
}

/**
 * Sets the value of a variable slot.
 *
 * @param THIS
 *  The variable instance addressing the scalar slot. Excessive or missing
 *  indices are flagged as an error.
 * @param value
 *  The value.
 * @param mode
 *  The access mode, expressed by some VAR_ACCESS_* constant.
 * @param priority
 *  The priority of the value being set. This is some arbitrary value that is
 *  stored together with the variable slots and that, together with the access
 *  mode, controls whether a slot can be changed or not.
 * @param check
 *  If TRUE, the value is checked against the variable type. FALSE is used when
 *  reading variable values from the configuration as the type system is not
 *  fully initialized at this point; checking these variable values is done at
 *  a later point.
 */
static BOOL
var_instance_do_set_value(
    struct var_instance_t *THIS,
    char const *value,
    int mode,
    int priority,
    BOOL check
)
{
    union varslot_t *slot;
    char *inst_name;
    BOOL access_failed = FALSE;

    if (!value) {
        inst_name = var_instance_to_string(THIS);
        log_error(
            MODULE ": Cannot set variable '%s' to an undefined value.\n",
            inst_name
        );
        free(inst_name);
        return FALSE;
    }

    if (THIS->num_indices < THIS->var->level) {
        inst_name = var_instance_to_string(THIS);
        log_error(
            MODULE ": Missing indices (%d instead of %d) while setting variable '%s' to '%s'.\n",
            THIS->num_indices,
            THIS->var->level,
            inst_name,
            value
        );
        free(inst_name);
        return FALSE;
    }

    slot = var_instance_get_slot_for_writing(THIS, mode, priority, &access_failed);
    if (!slot) {
        return access_failed && mode == VAR_ACCESS_TENTATIVE;
    }

    inst_name = var_instance_to_string(THIS);
    if (check) {
        log_info(
            LOG_VAR,
            MODULE ": About to check value '%s' against type of variable '%s'.\n",
            value,
            inst_name
        );

        if (vartype_match(THIS->var->type, value)) {
            log_info(
                LOG_VAR,
                MODULE ": Value '%s' conforms to type of variable '%s'.\n",
                value,
                inst_name
            );
        }
        else {
            log_error(
                MODULE ": Failed to set variable '%s' to '%s' as new value is not properly typed.\n",
                inst_name,
                value
            );
            free(inst_name);
            return FALSE;
        }
    }

    log_info(
        LOG_VAR,
        MODULE ": Writing %s '%s' (priority %d) to variable '%s'.\n",
        check ? "checked" : "unchecked",
        value,
        priority,
        inst_name
    );
    free(inst_name);

    free(slot->entry.value);
    slot->entry.value = strsave(value);
    slot->entry.priority = priority;
    slot->entry.checked = check;
    return TRUE;
}

BOOL
var_instance_set_and_check_value(
    struct var_instance_t *THIS,
    char const *value,
    int mode,
    int priority
)
{
    return var_instance_do_set_value(THIS, value, mode, priority, TRUE);
}

BOOL
var_instance_set_value(
    struct var_instance_t *THIS,
    char const *value,
    int mode,
    int priority
)
{
    return var_instance_do_set_value(THIS, value, mode, priority, FALSE);
}

BOOL
var_instance_clear(
    struct var_instance_t *THIS,
    int mode,
    int priority
)
{
    union varslot_t *slot;
    BOOL access_failed = FALSE;

    if (THIS->num_indices == THIS->var->level) {
        char *inst_name = var_instance_to_string(THIS);
        log_error(
            MODULE ": Cannot clear non-array variable '%s'.\n",
            inst_name
        );
        free(inst_name);
        return FALSE;
    }

    slot = var_instance_get_slot_for_writing(THIS, mode, priority, &access_failed);
    if (!slot) {
        return access_failed && mode == VAR_ACCESS_TENTATIVE;
    }
    else {
        struct var_t *var_n;
        struct var_instance_t *inst_n;

        var_slot_destroy(slot, THIS->num_indices, THIS->var->level);
        var_n = THIS->var->var_n[THIS->num_indices];
        inst_n = var_instance_create(var_n, THIS->num_indices, THIS->indices);
        if (!var_instance_set_and_check_value(inst_n, "0", VAR_ACCESS_TENTATIVE,
                VAR_PRIORITY_ARRAY_BOUND)) {
            fatal_exit(
                MODULE ": BUG in var_clear_array(): Failed to update array bound '%s' with 0 while accessing variable '%s'.\n",
                var_instance_to_string(inst_n),
                var_instance_to_string(THIS)
            );
        }

        var_instance_destroy(inst_n);
        return TRUE;
    }
}
