/*****************************************************************************
 *  @file test_var_condition.c
 *  Functions for testing variable conditions.
 *
 *  Copyright (c) 2012 The fli4l team
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: test_var_condition.c 24261 2012-11-05 19:38:01Z kristov $
 *****************************************************************************
 */

#include <testing/tests.h>
#include <var.h>
#include <expression/expr_literal.h>
#include <expression/expr_logical_or.h>
#include <expression/expr_equal.h>
#include <expression/expr_variable.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>
#include <string.h>

#define MU_TESTSUITE "var.condition"

/**
 * Tests adding a simple non-array variable with a condition.
 */
mu_test_begin(test_var_cond_simple1)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_properties_t const *propsNew;
    struct var_t *var;
    struct expr_t *cond;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    var_properties_set_condition(props, expr_literal_create_boolean(var_get_global_scope(), FALSE));
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), props);
    mu_assert(var);
    mu_assert_eq_str("V", var_get_name(var));
    mu_assert(var_get_type(var) == type);
    mu_assert_eq_int(0, var_get_level(var));

    propsNew = var_get_properties(var);
    mu_assert(!propsNew->def_value);

    cond = var_get_condition(var);
    mu_assert(expr_equal(props->condition, cond));
    expr_destroy(cond);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding a one-dimensional array with a condition.
 */
mu_test_begin(test_var_cond_array1)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_properties_t const *propsNew;
    struct var_t *var;
    struct var_t *var_n;
    struct expr_t *cond;
    struct expr_t *cond_exp;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    var_properties_set_condition(props, expr_literal_create_boolean(var_get_global_scope(), FALSE));
    var = var_add(var_get_global_scope(), "V_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), props);
    mu_assert(var);
    mu_assert_eq_str("V_%", var_get_name(var));
    mu_assert(var_get_type(var) == type);
    mu_assert_eq_int(1, var_get_level(var));

    propsNew = var_get_properties(var);
    mu_assert(!propsNew->def_value);

    cond = var_get_condition(var);
    mu_assert(expr_equal(props->condition, cond));
    expr_destroy(cond);

    var_n = var_get("V_N");
    mu_assert(var_n);
    propsNew = var_get_properties(var_n);
    mu_assert_eq_str("0", propsNew->def_value);

    cond = var_get_condition(var_n);
    cond_exp = var_get_condition(var);
    mu_assert(expr_equal(cond_exp, cond));
    expr_destroy(cond_exp);
    expr_destroy(cond);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding two one-dimensional arrays depending on the same array bound
 * but with different conditions (one/one) (1).
 */
mu_test_begin(test_var_cond_array2a)
    struct vartype_t *type;
    struct var_properties_t *props1;
    struct var_properties_t *props2;
    struct var_t *var_n;
    struct expr_t *cond_n;
    struct expr_t *cond_exp;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props1 = var_properties_create();
    var_properties_set_condition(props1, expr_literal_create_boolean(var_get_global_scope(), TRUE));
    var_add(var_get_global_scope(), "V_%_A", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), props1);
    props2 = var_properties_create();
    var_properties_set_condition(props2, expr_literal_create_boolean(var_get_global_scope(), FALSE));
    var_add(var_get_global_scope(), "V_%_B", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 24), props2);

    var_n = var_get("V_N");
    mu_assert(var_n);
    cond_n = var_get_condition(var_n);
    cond_exp = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    mu_assert(expr_equal(cond_exp, cond_n));
    expr_destroy(cond_exp);
    expr_destroy(cond_n);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding two one-dimensional arrays depending on the same array bound
 * but with different conditions (one/one) (2).
 */
mu_test_begin(test_var_cond_array2b)
    struct vartype_t *type;
    struct var_properties_t *props1;
    struct var_properties_t *props2;
    struct var_t *var_n;
    struct expr_t *cond_n;
    struct expr_t *cond_exp;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props1 = var_properties_create();
    var_properties_set_condition(props1, expr_literal_create_boolean(var_get_global_scope(), FALSE));
    var_add(var_get_global_scope(), "V_%_A", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), props1);
    props2 = var_properties_create();
    var_properties_set_condition(props2, expr_literal_create_boolean(var_get_global_scope(), FALSE));
    var_add(var_get_global_scope(), "V_%_B", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 24), props2);

    var_n = var_get("V_N");
    mu_assert(var_n);
    cond_n = var_get_condition(var_n);
    cond_exp = expr_literal_create_boolean(var_get_global_scope(), FALSE);
    mu_assert(expr_equal(cond_exp, cond_n));
    expr_destroy(cond_exp);
    expr_destroy(cond_n);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding two one-dimensional arrays depending on the same array bound
 * but with different conditions (one/one) (3).
 */
mu_test_begin(test_var_cond_array2c)
    struct vartype_t *type;
    struct var_properties_t *props1;
    struct var_properties_t *props2;
    struct var_t *var_n;
    struct expr_t *cond1;
    struct expr_t *cond2;
    struct expr_t *cond_n;
    struct expr_t *cond_exp;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props1 = var_properties_create();
    cond1 = expr_equal_create(
        var_get_global_scope(),
        expr_literal_create_boolean(var_get_global_scope(), FALSE),
        expr_literal_create_boolean(var_get_global_scope(), FALSE)
    );
    var_properties_set_condition(props1, cond1);
    var_add(var_get_global_scope(), "V_%_A", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), props1);
    props2 = var_properties_create();
    cond2 = expr_equal_create(
        var_get_global_scope(),
        expr_literal_create_boolean(var_get_global_scope(), TRUE),
        expr_literal_create_boolean(var_get_global_scope(), FALSE)
    );
    var_properties_set_condition(props2, cond2);
    var_add(var_get_global_scope(), "V_%_B", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 24), props2);

    var_n = var_get("V_N");
    mu_assert(var_n);
    cond_n = var_get_condition(var_n);
    cond_exp = expr_logical_or_create(var_get_global_scope(),
        expr_clone(cond1),
        expr_clone(cond2)
    );
    mu_assert(expr_equal(cond_exp, cond_n));
    expr_destroy(cond_exp);
    expr_destroy(cond_n);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding two one-dimensional arrays depending on the same array bound
 * but with different conditions (none/one).
 */
mu_test_begin(test_var_cond_array3)
    struct vartype_t *type;
    struct var_properties_t *props1;
    struct var_properties_t *props2;
    struct var_t *var_n;
    struct expr_t *cond_n;
    struct expr_t *cond_exp;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props1 = var_properties_create();
    var_add(var_get_global_scope(), "V_%_A", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), props1);
    props2 = var_properties_create();
    var_properties_set_condition(props2, expr_literal_create_boolean(var_get_global_scope(), FALSE));
    var_add(var_get_global_scope(), "V_%_B", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 24), props2);

    var_n = var_get("V_N");
    mu_assert(var_n);
    cond_n = var_get_condition(var_n);
    cond_exp = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    mu_assert(expr_equal(cond_exp, cond_n));
    expr_destroy(cond_exp);
    expr_destroy(cond_n);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding two one-dimensional arrays depending on the same array bound
 * but with different conditions (one/none).
 */
mu_test_begin(test_var_cond_array4)
    struct vartype_t *type;
    struct var_properties_t *props1;
    struct var_properties_t *props2;
    struct var_t *var_n;
    struct expr_t *cond_n;
    struct expr_t *cond_exp;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props1 = var_properties_create();
    var_properties_set_condition(props1, expr_literal_create_boolean(var_get_global_scope(), FALSE));
    var_add(var_get_global_scope(), "V_%_A", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), props1);
    props2 = var_properties_create();
    var_add(var_get_global_scope(), "V_%_B", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 24), props2);

    var_n = var_get("V_N");
    mu_assert(var_n);
    cond_n = var_get_condition(var_n);
    cond_exp = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    mu_assert(expr_equal(cond_exp, cond_n));
    expr_destroy(cond_exp);
    expr_destroy(cond_n);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding two simple variables which depend on each other. Reading those
 * variables must fail as we require that those variables are disabled in such
 * a case of mutual dependency in order to stop the recursion.
 */
mu_test_begin(test_var_cond_mutual)
    struct vartype_t *type;
    struct var_properties_t *props1;
    struct var_properties_t *props2;
    struct var_scope_t *scope;
    struct var_t *var1;
    struct var_t *var2;
    struct var_instance_t *varinst1;
    struct var_instance_t *varinst2;
    struct expr_t *cond1;
    struct expr_t *cond2;
    struct var_value_t *value;

    package_init_module();
    vartype_init_module();
    var_init_module();
    vartype_finalize_type_system();

    type = vartype_get(VARTYPE_PREDEFINED_BOOLEAN);
    scope = var_get_global_scope();

    props1 = var_properties_create();
    var1 = var_add(scope, "V1", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), props1);
    mu_assert(var1);
    props2 = var_properties_create();
    var2 = var_add(scope, "V2", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), props2);
    mu_assert(var2);

    varinst1 = var_instance_create(var1, 0, NULL);
    mu_assert(varinst1);
    varinst2 = var_instance_create(var2, 0, NULL);
    mu_assert(varinst2);

    cond1 = expr_variable_create(scope, varinst2);
    mu_assert(cond1);
    cond2 = expr_variable_create(scope, varinst1);
    mu_assert(cond2);

    props1 = var_properties_clone(var_get_properties(var1));
    var_properties_set_condition(props1, cond1);
    mu_assert(var_set_properties(var1, props1));

    props2 = var_properties_clone(var_get_properties(var2));
    var_properties_set_condition(props2, cond2);
    mu_assert(var_set_properties(var2, props2));

    mu_assert(var_instance_set_and_check_value(varinst1, "yes", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    mu_assert(var_instance_set_and_check_value(varinst2, "yes", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));

    mu_assert(!var_is_enabled(var1));
    value = var_instance_get_value(varinst1);
    mu_assert(!value);

    mu_assert(!var_is_enabled(var2));
    value = var_instance_get_value(varinst2);
    mu_assert(!value);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

void
test_var_condition(void)
{
    set_log_level(LOG_VAR);
    mu_run_test(test_var_cond_simple1);
    mu_run_test(test_var_cond_array1);
    mu_run_test(test_var_cond_array2a);
    mu_run_test(test_var_cond_array2b);
    mu_run_test(test_var_cond_array2c);
    mu_run_test(test_var_cond_array3);
    mu_run_test(test_var_cond_array4);
    mu_run_test(test_var_cond_mutual);
}
