/*****************************************************************************
 *  @file test_var_core.c
 *  Functions for testing variables.
 *
 *  Copyright (c) 2012 The fli4l team
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: test_var_core.c 24261 2012-11-05 19:38:01Z kristov $
 *****************************************************************************
 */

#include <testing/tests.h>
#include <var.h>
#include <expression/expr_literal.h>
#include <expression/expr_logical_or.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>
#include <string.h>

#define MU_TESTSUITE "var.core"

/**
 * Tests adding a simple non-array variable.
 */
mu_test_begin(test_var_add_simple1)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_properties_t const *propsNew;
    struct var_t *var;
    struct location_t *location;
    struct expr_t *expr_true
        = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    struct expr_t *cond;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    location = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    var = var_add(var_get_global_scope(), "V", type, location, props);
    mu_assert(var);
    mu_assert_eq_str("V", var_get_name(var));
    mu_assert(var_get_type(var) == type);
    mu_assert_eq_int(0, var_get_level(var));
    mu_assert(location == var_get_location(var));

    propsNew = var_get_properties(var);
    mu_assert(!propsNew->def_value);
    cond = var_get_condition(var);
    mu_assert(expr_equal(expr_true, cond));
    expr_destroy(cond);

    expr_destroy(expr_true);
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding a simple non-array variable with a default value.
 */
mu_test_begin(test_var_add_simple2)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_properties_t const *propsNew;
    struct var_t *var;
    struct location_t *location;
    struct expr_t *expr_true
        = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    struct expr_t *cond;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    var_properties_set_default_value(props, "0");
    location = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    var = var_add(var_get_global_scope(), "V", type, location, props);
    mu_assert(var);
    mu_assert_eq_str("V", var_get_name(var));
    mu_assert(var_get_type(var) == type);
    mu_assert_eq_int(0, var_get_level(var));
    mu_assert(location == var_get_location(var));

    propsNew = var_get_properties(var);
    mu_assert_eq_str("0", propsNew->def_value);
    cond = var_get_condition(var);
    mu_assert(expr_equal(expr_true, cond));
    expr_destroy(cond);

    expr_destroy(expr_true);
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding a simple non-array variable with invalid properties.
 */
mu_test_begin(test_var_add_simple3)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct location_t *loc;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    props->optional = TRUE;
    var_properties_set_default_value(props, "0");
    loc = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    mu_assert(!var_add(var_get_global_scope(), "V", type, loc, props));
    location_destroy(loc);
    var_properties_destroy(props);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding a one-dimensional array.
 */
mu_test_begin(test_var_add_array1)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_properties_t const *propsNew;
    struct var_t *var;
    struct var_t *var_n;
    struct location_t *location;
    struct expr_t *expr_true
        = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    struct expr_t *cond;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    location = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    var = var_add(var_get_global_scope(), "V_%", type, location, props);
    mu_assert(var);
    mu_assert_eq_str("V_%", var_get_name(var));
    mu_assert(var_get_type(var) == type);
    mu_assert_eq_int(1, var_get_level(var));
    mu_assert(location == var_get_location(var));

    propsNew = var_get_properties(var);
    mu_assert(!propsNew->def_value);
    cond = var_get_condition(var);
    mu_assert(expr_equal(expr_true, cond));
    expr_destroy(cond);

    var_n = var_get("V_N");
    mu_assert(var_n);
    propsNew = var_get_properties(var_n);
    mu_assert_eq_str("0", propsNew->def_value);
    cond = var_get_condition(var_n);
    mu_assert(expr_equal(expr_true, cond));
    expr_destroy(cond);

    expr_destroy(expr_true);
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding a one-dimensional array with a default value.
 */
mu_test_begin(test_var_add_array1_def)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_properties_t const *propsNew;
    struct var_t *var;
    struct var_t *var_n;
    struct location_t *location;
    struct expr_t *expr_true
        = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    struct expr_t *cond;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    var_properties_set_default_value(props, "3");
    location = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    var = var_add(var_get_global_scope(), "V_%", type, location, props);
    mu_assert(var);
    mu_assert_eq_str("V_%", var_get_name(var));
    mu_assert(var_get_type(var) == type);
    mu_assert_eq_int(1, var_get_level(var));
    mu_assert(location == var_get_location(var));

    propsNew = var_get_properties(var);
    mu_assert_eq_str("3", propsNew->def_value);
    cond = var_get_condition(var);
    mu_assert(expr_equal(expr_true, cond));
    expr_destroy(cond);

    var_n = var_get("V_N");
    mu_assert(var_n);
    propsNew = var_get_properties(var_n);
    mu_assert_eq_str("0", propsNew->def_value);
    cond = var_get_condition(var_n);
    mu_assert(expr_equal(expr_true, cond));
    expr_destroy(cond);

    expr_destroy(expr_true);
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding a one-dimensional array with a wrong name.
 */
mu_test_begin(test_var_add_array1_wrong)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct location_t *loc;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    loc = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    var = var_add(var_get_global_scope(), "V(%", type, loc, props);
    mu_assert(!var);
    location_destroy(loc);
    var_properties_destroy(props);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding a three-dimensional array.
 */
mu_test_begin(test_var_add_array2)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_properties_t const *propsNew;
    struct var_t *var;
    struct var_t *var_n;
    struct location_t *location;
    struct expr_t *expr_true
        = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    struct expr_t *cond;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    location = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    var = var_add(var_get_global_scope(), "V_%_A_%_B_%_C", type, location, props);
    mu_assert(var);
    mu_assert_eq_str("V_%_A_%_B_%_C", var_get_name(var));
    mu_assert(var_get_type(var) == type);
    mu_assert_eq_int(3, var_get_level(var));
    mu_assert(location == var_get_location(var));

    propsNew = var_get_properties(var);
    mu_assert(!propsNew->def_value);
    cond = var_get_condition(var);
    mu_assert(expr_equal(expr_true, cond));
    expr_destroy(cond);

    var_n = var_get("V_N");
    mu_assert(var_n);
    propsNew = var_get_properties(var_n);
    mu_assert_eq_str("0", propsNew->def_value);
    cond = var_get_condition(var_n);
    mu_assert(expr_equal(expr_true, cond));
    expr_destroy(cond);

    var_n = var_get("V_%_A_N");
    mu_assert(var_n);
    propsNew = var_get_properties(var_n);
    mu_assert_eq_str("0", propsNew->def_value);
    cond = var_get_condition(var_n);
    mu_assert(expr_equal(expr_true, cond));
    expr_destroy(cond);

    var_n = var_get("V_%_A_%_B_N");
    mu_assert(var_n);
    propsNew = var_get_properties(var_n);
    mu_assert_eq_str("0", propsNew->def_value);
    cond = var_get_condition(var_n);
    mu_assert(expr_equal(expr_true, cond));
    expr_destroy(cond);

    expr_destroy(expr_true);
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding two one-dimensional arrays depending on the same array bound.
 */
mu_test_begin(test_var_add_array3)
    struct vartype_t *type;
    struct var_properties_t *props1;
    struct var_properties_t *props2;
    struct var_t *var_n;
    struct expr_t *expr_true
        = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    struct expr_t *cond;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props1 = var_properties_create();
    var_add(var_get_global_scope(), "V_%_A", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), props1);
    props2 = var_properties_create();
    var_add(var_get_global_scope(), "V_%_B", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 24), props2);

    var_n = var_get("V_N");
    mu_assert(var_n);
    cond = var_get_condition(var_n);
    mu_assert(expr_equal(expr_true, cond));
    expr_destroy(cond);

    expr_destroy(expr_true);
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests adding a one-dimensional array with an incompatible array bound.
 */
mu_test_begin(test_var_add_array_incompatible)
    struct vartype_t *type_none;
    struct vartype_t *type;
    struct var_properties_t *props1;
    struct var_properties_t *props2;
    struct var_t *var;
    struct location_t *loc;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type_none = vartype_get(VARTYPE_PREDEFINED_ANYTHING);
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props1 = var_properties_create();
    var_add(var_get_global_scope(), "V_N", type_none, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 22), props1);
    props2 = var_properties_create();
    loc = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    var = var_add(var_get_global_scope(), "V_%", type, loc, props2);
    mu_assert(!var);
    location_destroy(loc);
    var_properties_destroy(props2);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests var_add(var_get_global_scope(), ) using wrong names.
 */
mu_test_begin(test_var_add_wrong)
    struct vartype_t *type;
    struct location_t *loc;
    struct var_properties_t *props;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    loc = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    props = var_properties_create();
    mu_assert(!var_add(var_get_global_scope(), "V_._%", type, loc, props));
    mu_assert(!var_add(var_get_global_scope(), "V_()_%", type, loc, props));
    mu_assert(!var_add(var_get_global_scope(), "V_[A]_%", type, loc, props));
    mu_assert(!var_add(var_get_global_scope(), "V_A{4}_%", type, loc, props));
    mu_assert(!var_add(var_get_global_scope(), "V_?_%", type, loc, props));
    mu_assert(!var_add(var_get_global_scope(), "V_x*_%", type, loc, props));
    mu_assert(!var_add(var_get_global_scope(), "V_y+_%", type, loc, props));
    mu_assert(!var_add(var_get_global_scope(), "V_^^_%", type, loc, props));
    mu_assert(!var_add(var_get_global_scope(), "V_$$_%", type, loc, props));
    var_properties_destroy(props);
    location_destroy(loc);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests redefinition of a simple non-array variable.
 */
mu_test_begin(test_var_add_redef1)
    struct vartype_t *type;
    struct var_t *var1;
    struct var_t *var2;
    struct location_t *loc;
    struct var_properties_t *props;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    var1 = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    mu_assert(var1);
    loc = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 24);
    props = var_properties_create();
    var2 = var_add(var_get_global_scope(), "V", type, loc, props);
    mu_assert(!var2);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests redefinition of a one-dimensionale array variable.
 */
mu_test_begin(test_var_add_redef2)
    struct vartype_t *type;
    struct var_t *var1;
    struct var_t *var2;
    struct location_t *loc;
    struct var_properties_t *props;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    var1 = var_add(var_get_global_scope(), "V_%_%_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    mu_assert(var1);
    loc = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 24);
    props = var_properties_create();
    var2 = var_add(var_get_global_scope(), "V_%_%_%", type, loc, props);
    mu_assert(!var2);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests retrieving a non-existing variable.
 */
mu_test_begin(test_var_get_non_existing)
    package_init_module();
    vartype_init_module();
    var_init_module();
    vartype_finalize_type_system();
    mu_assert(!var_get("V"));
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests changing variable properties (1).
 */
mu_test_begin(test_var_change_props1)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_t *var_n;
    struct location_t *location;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    location = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    var = var_add(var_get_global_scope(), "V_%", type, location, props);
    mu_assert(var);
    var_n = var_get("V_N");
    mu_assert(var_n);

    props = var_properties_clone(var_get_properties(var));
    props->persistent = !props->persistent;
    mu_assert(!var_set_properties(var, props));
    var_properties_destroy(props);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests changing variable properties (2).
 */
mu_test_begin(test_var_change_props2)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_t *var_n;
    struct location_t *location;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    location = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    var = var_add(var_get_global_scope(), "V_%", type, location, props);
    mu_assert(var);
    var_n = var_get("V_N");
    mu_assert(var_n);

    props = var_properties_clone(var_get_properties(var));
    props->optional = !props->optional;
    mu_assert(!var_set_properties(var, props));
    var_properties_destroy(props);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests changing variable properties (3).
 */
mu_test_begin(test_var_change_props3)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_t *var_n;
    struct location_t *location;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    location = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    var = var_add(var_get_global_scope(), "V_%", type, location, props);
    mu_assert(var);
    var_n = var_get("V_N");
    mu_assert(var_n);

    props = var_properties_clone(var_get_properties(var));
    var_properties_set_default_value(props, "2");
    mu_assert(var_set_properties(var, props));

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests changing variable properties (4).
 */
mu_test_begin(test_var_change_props4)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_t *var_n;
    struct expr_t *expr_true
        = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    struct expr_t *expr_false
        = expr_literal_create_boolean(var_get_global_scope(), FALSE);
    struct expr_t *cond;
    struct location_t *location;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    props = var_properties_create();
    location = test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23);
    var = var_add(var_get_global_scope(), "V_%", type, location, props);
    mu_assert(var);
    var_n = var_get("V_N");
    mu_assert(var_n);

    cond = var_get_condition(var_n);
    mu_assert(expr_equal(expr_true, cond));
    expr_destroy(cond);

    props = var_properties_clone(var_get_properties(var));
    var_properties_set_condition(props, expr_clone(expr_false));
    mu_assert(var_set_properties(var, props));

    cond = var_get_condition(var_n);
    mu_assert(expr_equal(expr_false, cond));
    expr_destroy(cond);

    expr_destroy(expr_true);
    expr_destroy(expr_false);
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

void
test_var_core(void)
{
    set_log_level(LOG_VAR);
    mu_run_test(test_var_add_simple1);
    mu_run_test(test_var_add_simple2);
    mu_run_test(test_var_add_simple3);
    mu_run_test(test_var_add_array1);
    mu_run_test(test_var_add_array1_def);
    mu_run_test(test_var_add_array1_wrong);
    mu_run_test(test_var_add_array2);
    mu_run_test(test_var_add_array3);
    mu_run_test(test_var_add_array_incompatible);
    mu_run_test(test_var_add_wrong);
    mu_run_test(test_var_add_redef1);
    mu_run_test(test_var_add_redef2);
    mu_run_test(test_var_get_non_existing);
    mu_run_test(test_var_change_props1);
    mu_run_test(test_var_change_props2);
    mu_run_test(test_var_change_props3);
    mu_run_test(test_var_change_props4);
}
