/*****************************************************************************
 *  @file test_var_iter.c
 *  Functions for testing iteration of arrays.
 *
 *  Copyright (c) 2012 The fli4l team
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: test_var_iter.c 24192 2012-10-29 07:31:29Z kristov $
 *****************************************************************************
 */

#include <testing/tests.h>
#include <var.h>
#include <expression/expr_literal.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>
#include <string.h>

#define MU_TESTSUITE "var.iter"

/**
 * Tests iterating over a one-dimensional array without holes.
 */
mu_test_begin(test_var_iter_onedim_no_holes)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_value_t *value;
    struct var_value_iterator_t *iter;
    unsigned *ind;
    struct var_instance_t *inst;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
    vartype_finalize_type_system();
    props = var_properties_create();
    var = var_add(var_get_global_scope(), "V_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 42), props);

    inst = var_instance_create(var, 1, NULL);
    inst->indices[0] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "1", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 1;
    mu_assert(var_instance_set_and_check_value(inst, "2", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 2;
    mu_assert(var_instance_set_and_check_value(inst, "3", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 3;
    mu_assert(var_instance_set_and_check_value(inst, "4", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    var_instance_destroy(inst);

    inst = var_instance_create(var, 0, NULL);
    iter = var_value_iterator_create(inst);
    var_instance_destroy(inst);
    mu_assert(iter);
    mu_assert(var_value_iterator_get_var(iter) == var);

    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(0, ind[0]);
    free(ind);
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "1");
    var_value_destroy(value);
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "1");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(0, ind[0]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "2");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(1, ind[0]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "3");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(2, ind[0]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "4");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(3, ind[0]);
    free(ind);

    mu_assert(!var_value_iterator_next(iter));
    mu_assert(!var_value_iterator_get(iter));
    var_value_iterator_destroy(iter);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests iterating over a one-dimensional array with holes.
 */
mu_test_begin(test_var_iter_onedim_holes)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_value_t *value;
    struct var_value_iterator_t *iter;
    unsigned *ind;
    struct var_instance_t *inst;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
    vartype_finalize_type_system();
    props = var_properties_create();
    var = var_add(var_get_global_scope(), "V_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 42), props);

    inst = var_instance_create(var, 1, NULL);
    inst->indices[0] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "1", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 2;
    mu_assert(var_instance_set_and_check_value(inst, "3", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 4;
    mu_assert(var_instance_set_and_check_value(inst, "5", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    var_instance_destroy(inst);

    inst = var_instance_create(var, 0, NULL);
    iter = var_value_iterator_create(inst);
    var_instance_destroy(inst);
    mu_assert(iter);
    mu_assert(var_value_iterator_get_var(iter) == var);

    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "1");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(0, ind[0]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "3");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(2, ind[0]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "5");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(4, ind[0]);
    free(ind);

    mu_assert(!var_value_iterator_next(iter));
    mu_assert(!var_value_iterator_get(iter));
    var_value_iterator_destroy(iter);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests iterating over a two-dimensional array without holes.
 */
mu_test_begin(test_var_iter_twodim_no_holes)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_value_t *value;
    struct var_value_iterator_t *iter;
    unsigned *ind;
    struct var_instance_t *inst;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
    vartype_finalize_type_system();
    props = var_properties_create();
    var = var_add(var_get_global_scope(), "V_%_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 42), props);

    inst = var_instance_create(var, 2, NULL);
    inst->indices[0] = 0;
    inst->indices[1] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "0", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 0;
    inst->indices[1] = 1;
    mu_assert(var_instance_set_and_check_value(inst, "1", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 1;
    inst->indices[1] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "2", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 1;
    inst->indices[1] = 1;
    mu_assert(var_instance_set_and_check_value(inst, "3", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 1;
    inst->indices[1] = 2;
    mu_assert(var_instance_set_and_check_value(inst, "4", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 2;
    inst->indices[1] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "5", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    var_instance_destroy(inst);

    inst = var_instance_create(var, 0, NULL);
    iter = var_value_iterator_create(inst);
    var_instance_destroy(inst);
    mu_assert(iter);
    mu_assert(var_value_iterator_get_var(iter) == var);

    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "0");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(0, ind[0]);
    mu_assert_eq_int(0, ind[1]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "1");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(0, ind[0]);
    mu_assert_eq_int(1, ind[1]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "2");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(1, ind[0]);
    mu_assert_eq_int(0, ind[1]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "3");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(1, ind[0]);
    mu_assert_eq_int(1, ind[1]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "4");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(1, ind[0]);
    mu_assert_eq_int(2, ind[1]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "5");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(2, ind[0]);
    mu_assert_eq_int(0, ind[1]);
    free(ind);

    mu_assert(!var_value_iterator_next(iter));
    mu_assert(!var_value_iterator_get(iter));
    var_value_iterator_destroy(iter);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests iterating over a two-dimensional array with holes.
 */
mu_test_begin(test_var_iter_twodim_holes)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_value_t *value;
    struct var_value_iterator_t *iter;
    unsigned *ind;
    struct var_instance_t *inst;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
    vartype_finalize_type_system();
    props = var_properties_create();
    var = var_add(var_get_global_scope(), "V_%_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 42), props);

    inst = var_instance_create(var, 2, NULL);
    inst->indices[0] = 0;
    inst->indices[1] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "0", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 2;
    inst->indices[1] = 2;
    mu_assert(var_instance_set_and_check_value(inst, "2", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 3;
    inst->indices[1] = 3;
    mu_assert(var_instance_set_and_check_value(inst, "3", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    var_instance_destroy(inst);

    inst = var_instance_create(var, 0, NULL);
    iter = var_value_iterator_create(inst);
    var_instance_destroy(inst);
    mu_assert(iter);
    mu_assert(var_value_iterator_get_var(iter) == var);

    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "0");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(0, ind[0]);
    mu_assert_eq_int(0, ind[1]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "2");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(2, ind[0]);
    mu_assert_eq_int(2, ind[1]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "3");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(3, ind[0]);
    mu_assert_eq_int(3, ind[1]);
    free(ind);

    mu_assert(!var_value_iterator_next(iter));
    mu_assert(!var_value_iterator_get(iter));
    var_value_iterator_destroy(iter);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests iterating over a one-dimensional array which is modified during the
 * iteration.
 */
mu_test_begin(test_var_iter_onedim_modified)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_value_t *value;
    struct var_value_iterator_t *iter;
    unsigned *ind;
    struct var_instance_t *inst;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
    vartype_finalize_type_system();
    props = var_properties_create();
    var = var_add(var_get_global_scope(), "V_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 42), props);

    inst = var_instance_create(var, 1, NULL);
    inst->indices[0] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "1", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 1;
    mu_assert(var_instance_set_and_check_value(inst, "2", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 2;
    mu_assert(var_instance_set_and_check_value(inst, "3", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 3;
    mu_assert(var_instance_set_and_check_value(inst, "4", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    var_instance_destroy(inst);

    inst = var_instance_create(var, 0, NULL);
    iter = var_value_iterator_create(inst);
    var_instance_destroy(inst);
    mu_assert(iter);
    mu_assert(var_value_iterator_get_var(iter) == var);

    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "1");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(0, ind[0]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "2");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(1, ind[0]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "3");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(2, ind[0]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "4");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(3, ind[0]);
    free(ind);

    inst = var_instance_create(var, 1, NULL);
    inst->indices[0] = 5;
    mu_assert(var_instance_set_and_check_value(inst, "6", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    var_instance_destroy(inst);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "6");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(5, ind[0]);
    free(ind);

    mu_assert(!var_value_iterator_next(iter));
    mu_assert(!var_value_iterator_get(iter));
    var_value_iterator_destroy(iter);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests iterating over an empty one-dimensional array.
 */
mu_test_begin(test_var_iter_onedim_empty)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_value_iterator_t *iter;
    struct var_instance_t *inst;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
    vartype_finalize_type_system();
    props = var_properties_create();
    var = var_add(var_get_global_scope(), "V_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 42), props);

    inst = var_instance_create(var, 0, NULL);
    iter = var_value_iterator_create(inst);
    var_instance_destroy(inst);
    mu_assert(iter);
    mu_assert(var_value_iterator_get_var(iter) == var);

    mu_assert(!var_value_iterator_get_var_indices(iter));
    mu_assert(!var_value_iterator_get(iter));
    mu_assert(!var_value_iterator_next(iter));
    var_value_iterator_destroy(iter);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests iterating over an existing subarray of a two-dimensional array.
 */
mu_test_begin(test_var_iter_subarray1)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_value_t *value;
    struct var_value_iterator_t *iter;
    unsigned *ind;
    struct var_instance_t *inst;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
    vartype_finalize_type_system();
    props = var_properties_create();
    var = var_add(var_get_global_scope(), "V_%_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 42), props);

    inst = var_instance_create(var, 2, NULL);
    inst->indices[0] = 0;
    inst->indices[1] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "0", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 0;
    inst->indices[1] = 1;
    mu_assert(var_instance_set_and_check_value(inst, "1", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 1;
    inst->indices[1] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "2", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 1;
    inst->indices[1] = 1;
    mu_assert(var_instance_set_and_check_value(inst, "3", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 1;
    inst->indices[1] = 2;
    mu_assert(var_instance_set_and_check_value(inst, "4", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 2;
    inst->indices[1] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "5", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    var_instance_destroy(inst);

    inst = var_instance_create(var, 1, NULL);
    inst->indices[0] = 1;
    iter = var_value_iterator_create(inst);
    var_instance_destroy(inst);
    mu_assert(iter);
    mu_assert(var_value_iterator_get_var(iter) == var);

    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "2");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(0, ind[0]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "3");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(1, ind[0]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "4");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(2, ind[0]);
    free(ind);

    mu_assert(!var_value_iterator_next(iter));
    mu_assert(!var_value_iterator_get(iter));
    var_value_iterator_destroy(iter);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests iterating over a non-existing subarray of a two-dimensional array.
 */
mu_test_begin(test_var_iter_subarray2)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_value_iterator_t *iter;
    struct var_instance_t *inst;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
    vartype_finalize_type_system();
    props = var_properties_create();
    var = var_add(var_get_global_scope(), "V_%_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 42), props);

    inst = var_instance_create(var, 2, NULL);
    inst->indices[0] = 0;
    inst->indices[1] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "0", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 0;
    inst->indices[1] = 1;
    mu_assert(var_instance_set_and_check_value(inst, "1", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 1;
    inst->indices[1] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "2", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 1;
    inst->indices[1] = 1;
    mu_assert(var_instance_set_and_check_value(inst, "3", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 1;
    inst->indices[1] = 2;
    mu_assert(var_instance_set_and_check_value(inst, "4", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 2;
    inst->indices[1] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "5", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    var_instance_destroy(inst);

    inst = var_instance_create(var, 1, NULL);
    inst->indices[0] = 3;
    iter = var_value_iterator_create(inst);
    var_instance_destroy(inst);
    mu_assert(iter);
    mu_assert(var_value_iterator_get_var(iter) == var);

    mu_assert(!var_value_iterator_get_var_indices(iter));
    mu_assert(!var_value_iterator_get(iter));
    mu_assert(!var_value_iterator_next(iter));
    var_value_iterator_destroy(iter);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests iterating over a non-array variable.
 */
mu_test_begin(test_var_iter_non_array)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_value_iterator_t *iter;
    struct var_instance_t *inst;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
    vartype_finalize_type_system();
    props = var_properties_create();
    var = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 42), props);

    inst = var_instance_create(var, 0, NULL);
    mu_assert(var_instance_set_and_check_value(inst, "1", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    iter = var_value_iterator_create(inst);
    var_instance_destroy(inst);
    mu_assert(iter);
    mu_assert(var_value_iterator_get_var(iter) == var);

    mu_assert(!var_value_iterator_get_var_indices(iter));
    mu_assert(!var_value_iterator_get(iter));
    mu_assert(!var_value_iterator_next(iter));

    var_value_iterator_destroy(iter);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests destroying an unfinished iterator.
 */
mu_test_begin(test_var_iter_destroy)
    struct vartype_t *type;
    struct var_properties_t *props;
    struct var_t *var;
    struct var_value_t *value;
    struct var_value_iterator_t *iter;
    unsigned *ind;
    struct var_instance_t *inst;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_get(VARTYPE_PREDEFINED_UNSIGNED_INTEGER);
    vartype_finalize_type_system();
    props = var_properties_create();
    var = var_add(var_get_global_scope(), "V_%", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 42), props);

    inst = var_instance_create(var, 1, NULL);
    inst->indices[0] = 0;
    mu_assert(var_instance_set_and_check_value(inst, "1", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 1;
    mu_assert(var_instance_set_and_check_value(inst, "2", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 2;
    mu_assert(var_instance_set_and_check_value(inst, "3", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    inst->indices[0] = 3;
    mu_assert(var_instance_set_and_check_value(inst, "4", VAR_ACCESS_STRONG, VAR_PRIORITY_CONFIGURATION));
    var_instance_destroy(inst);

    inst = var_instance_create(var, 0, NULL);
    iter = var_value_iterator_create(inst);
    var_instance_destroy(inst);
    mu_assert(iter);
    mu_assert(var_value_iterator_get_var(iter) == var);

    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "1");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(0, ind[0]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    value = var_value_iterator_get(iter);
    mu_assert(value);
    mu_assert_eq_str(value->value, "2");
    var_value_destroy(value);
    ind = var_value_iterator_get_var_indices(iter);
    mu_assert(ind);
    mu_assert_eq_int(1, ind[0]);
    free(ind);

    mu_assert(var_value_iterator_next(iter));
    var_value_iterator_destroy(iter);
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

void
test_var_iter(void)
{
    set_log_level(LOG_VAR);
    mu_run_test(test_var_iter_onedim_no_holes);
    mu_run_test(test_var_iter_onedim_holes);
    mu_run_test(test_var_iter_twodim_no_holes);
    mu_run_test(test_var_iter_twodim_holes);
    mu_run_test(test_var_iter_onedim_modified);
    mu_run_test(test_var_iter_onedim_empty);
    mu_run_test(test_var_iter_subarray1);
    mu_run_test(test_var_iter_subarray2);
    mu_run_test(test_var_iter_non_array);
    mu_run_test(test_var_iter_destroy);
}
