/*****************************************************************************
 *  @file test_var_props.c
 *  Functions for testing variable property sets.
 *
 *  Copyright (c) 2012 The fli4l team
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: test_var_props.c 24254 2012-11-04 17:47:48Z kristov $
 *****************************************************************************
 */

#include <testing/tests.h>
#include <var.h>
#include <expression/expr_literal.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>
#include <string.h>

#define MU_TESTSUITE "var.props"

/**
 * Tests the default property set.
 */
mu_test_begin(test_var_props_default)
    struct var_properties_t *props = var_properties_create();
    struct expr_t *expr_true
        = expr_literal_create_boolean(var_get_global_scope(), TRUE);

    mu_assert(!*props->comment);
    mu_assert(props->persistent);
    mu_assert(!props->optional);
    mu_assert(!props->def_value);
    mu_assert(expr_equal(expr_true, props->condition));

    expr_destroy(expr_true);
    var_properties_destroy(props);
mu_test_end()

/**
 * Tests cloning a property set.
 */
mu_test_begin(test_var_props_clone)
    struct var_properties_t *props1 = var_properties_create();
    struct var_properties_t *props2;

    props1->persistent = FALSE;
    var_properties_set_comment(props1, "Blabla");
    var_properties_set_default_value(props1, "Hallo");
    var_properties_set_condition(props1, expr_literal_create_boolean(var_get_global_scope(), FALSE));

    props2 = var_properties_clone(props1);
    mu_assert(props2);
    mu_assert_eq_str(props1->comment, props2->comment);
    mu_assert_eq_int(props1->persistent, props2->persistent);
    mu_assert_eq_int(props1->optional, props2->optional);
    mu_assert(props1->def_value != props2->def_value);
    mu_assert_eq_str(props1->def_value, props2->def_value);
    mu_assert(props1->condition != props2->condition);
    mu_assert(expr_equal(props1->condition, props2->condition));

    var_properties_destroy(props2);
    var_properties_destroy(props1);
mu_test_end()

/**
 * Tests changing the comment in a property set.
 */
mu_test_begin(test_var_props_set_comment)
    struct var_properties_t *props = var_properties_create();

    mu_assert(!*props->comment);
    var_properties_set_comment(props, "Blabla");
    mu_assert_eq_str("Blabla", props->comment);

    var_properties_destroy(props);
mu_test_end()

/**
 * Tests changing the default value in a property set.
 */
mu_test_begin(test_var_props_set_default_value)
    struct var_properties_t *props = var_properties_create();

    mu_assert(!props->def_value);
    var_properties_set_default_value(props, "Hallo");
    mu_assert_eq_str("Hallo", props->def_value);

    var_properties_destroy(props);
mu_test_end()

/**
 * Tests changing the condition in a property set.
 */
mu_test_begin(test_var_props_set_condition)
    struct var_properties_t *props = var_properties_create();
    struct expr_t *expr_true
        = expr_literal_create_boolean(var_get_global_scope(), TRUE);
    struct expr_t *expr_false
        = expr_literal_create_boolean(var_get_global_scope(), FALSE);
    struct expr_t *cond;

    mu_assert(expr_equal(expr_true, props->condition));
    var_properties_set_condition(props, expr_false);
    cond = expr_literal_create_boolean(var_get_global_scope(), FALSE);
    mu_assert(expr_equal(cond, props->condition));
    expr_destroy(cond);

    expr_destroy(expr_true);
    var_properties_destroy(props);
mu_test_end()

/**
 * Tests checking default properties.
 */
mu_test_begin(test_var_props_check_valid1)
    struct var_properties_t *props = var_properties_create();
    mu_assert(var_properties_check(props, "V"));
    var_properties_destroy(props);
mu_test_end()

/**
 * Tests checking valid properties (1).
 */
mu_test_begin(test_var_props_check_valid2)
    struct var_properties_t *props = var_properties_create();
    props->persistent = FALSE;
    mu_assert(var_properties_check(props, "V"));
    var_properties_destroy(props);
mu_test_end()

/**
 * Tests checking valid properties (2).
 */
mu_test_begin(test_var_props_check_valid3)
    struct var_properties_t *props = var_properties_create();
    props->optional = TRUE;
    mu_assert(var_properties_check(props, "V"));
    var_properties_destroy(props);
mu_test_end()

/**
 * Tests checking valid properties (3).
 */
mu_test_begin(test_var_props_check_valid4)
    struct var_properties_t *props = var_properties_create();
    var_properties_set_default_value(props, "Hallo");
    mu_assert(var_properties_check(props, "V"));
    var_properties_destroy(props);
mu_test_end()

/**
 * Tests checking valid properties (4).
 */
mu_test_begin(test_var_props_check_valid5)
    struct var_properties_t *props = var_properties_create();
    props->persistent = FALSE;
    var_properties_set_default_value(props, "Hallo");
    mu_assert(var_properties_check(props, "V"));
    var_properties_destroy(props);
mu_test_end()

/**
 * Tests checking invalid properties: Transient variables may not be optional.
 */
mu_test_begin(test_var_props_check_invalid1)
    struct var_properties_t *props = var_properties_create();
    props->persistent = FALSE;
    props->optional = TRUE;
    mu_assert(!var_properties_check(props, "V"));
    var_properties_destroy(props);
mu_test_end()

/**
 * Tests checking invalid properties: Persistent optional variables may not
 * have a default value.
 */
mu_test_begin(test_var_props_check_invalid2)
    struct var_properties_t *props = var_properties_create();
    props->optional = TRUE;
    var_properties_set_default_value(props, "Hallo");
    mu_assert(!var_properties_check(props, "V"));
    var_properties_destroy(props);
mu_test_end()

void
test_var_props(void)
{
    set_log_level(LOG_VAR);
    mu_run_test(test_var_props_default);
    mu_run_test(test_var_props_clone);
    mu_run_test(test_var_props_set_comment);
    mu_run_test(test_var_props_set_default_value);
    mu_run_test(test_var_props_set_condition);
    mu_run_test(test_var_props_check_valid1);
    mu_run_test(test_var_props_check_valid2);
    mu_run_test(test_var_props_check_valid3);
    mu_run_test(test_var_props_check_valid4);
    mu_run_test(test_var_props_check_valid5);
    mu_run_test(test_var_props_check_invalid1);
    mu_run_test(test_var_props_check_invalid2);
}
