/*****************************************************************************
 *  @file test_var_scope.c
 *  Functions for testing variable scopes.
 *
 *  Copyright (c) 2012 The fli4l team
 *
 *  This file is part of fli4l.
 *
 *  fli4l is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fli4l is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fli4l.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Last Update: $Id: test_var_scope.c 24254 2012-11-04 17:47:48Z kristov $
 *****************************************************************************
 */

#include <testing/tests.h>
#include <var.h>
#include <libmkfli4l/str.h>
#include <libmkfli4l/log.h>
#include <stdlib.h>
#include <string.h>

#define MU_TESTSUITE "var.scope"

/**
 * Tests the global scope.
 */
mu_test_begin(test_var_scope_global)
    struct vartype_t *type;
    struct var_t *var_global;
    struct var_t *var;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    var_global = var_add(var_get_global_scope(), "V1", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    mu_assert(var_global);

    var = var_scope_get(var_get_global_scope(), "V1", FALSE);
    mu_assert(var == var_global);
    var = var_scope_get(var_get_global_scope(), "V1", TRUE);
    mu_assert(var == var_global);

    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests a nested scope.
 */
mu_test_begin(test_var_scope_nested)
    struct vartype_t *type;
    struct var_scope_t *scope;
    struct var_t *var_global;
    struct var_t *var_local;
    struct var_t *var;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    scope = var_scope_create(location_create_internal(PACKAGE_FILE_TYPE_EXTCHECK), var_get_global_scope());
    var_global = var_add(var_get_global_scope(), "V1", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    mu_assert(var_global);
    var_local = var_add(scope, "V2", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    mu_assert(var_local);

    var = var_scope_get(scope, "V1", FALSE);
    mu_assert(!var);
    var = var_scope_get(scope, "V1", TRUE);
    mu_assert(var == var_global);
    var = var_scope_get(var_get_global_scope(), "V1", FALSE);
    mu_assert(var == var_global);
    var = var_scope_get(var_get_global_scope(), "V1", TRUE);
    mu_assert(var == var_global);

    var = var_scope_get(scope, "V2", FALSE);
    mu_assert(var == var_local);
    var = var_scope_get(scope, "V2", TRUE);
    mu_assert(var == var_local);
    var = var_scope_get(var_get_global_scope(), "V2", FALSE);
    mu_assert(!var);
    var = var_scope_get(var_get_global_scope(), "V2", TRUE);
    mu_assert(!var);

    var_scope_destroy(scope);
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

/**
 * Tests redefinition in a nested scope.
 */
mu_test_begin(test_var_scope_redef)
    struct vartype_t *type;
    struct var_scope_t *scope;
    struct var_t *var_global;
    struct var_t *var_local;
    struct var_t *var;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    scope = var_scope_create(location_create_internal(PACKAGE_FILE_TYPE_EXTCHECK), var_get_global_scope());
    var_global = var_add(var_get_global_scope(), "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    mu_assert(var_global);
    var_local = var_add(scope, "V", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    mu_assert(var_local);

    var = var_scope_get(scope, "V", FALSE);
    mu_assert(var == var_local);
    var = var_scope_get(scope, "V", TRUE);
    mu_assert(var == var_local);
    var = var_scope_get(var_get_global_scope(), "V", FALSE);
    mu_assert(var == var_global);
    var = var_scope_get(var_get_global_scope(), "V", TRUE);
    mu_assert(var == var_global);

    var_scope_destroy(scope);
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

static BOOL
test_var_scope_foreach_visitor1(void *entry, void *data)
{
    ++*(int *) data;
    return TRUE;
    (void) entry;
}

static BOOL
test_var_scope_foreach_visitor2(void *entry, void *data)
{
    return FALSE;
    (void) data;
    (void) entry;
}

/**
 * Tests var_scope_foreach().
 */
mu_test_begin(test_var_scope_foreach)
    struct vartype_t *type;
    struct var_scope_t *scope;
    struct var_t *var_global;
    struct var_t *var_local;
    int i;

    package_init_module();
    vartype_init_module();
    var_init_module();
    type = vartype_add("MyType", "(0|1|2|3)", "0 to 3 expected", NULL, test_location_create(PACKAGE_FILE_TYPE_VARTYPE, 42));
    vartype_finalize_type_system();

    scope = var_scope_create(location_create_internal(PACKAGE_FILE_TYPE_EXTCHECK), var_get_global_scope());
    var_global = var_add(var_get_global_scope(), "V1", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    mu_assert(var_global);
    var_local = var_add(scope, "V2", type, test_location_create(PACKAGE_FILE_TYPE_VARDEF, 23), var_properties_create());
    mu_assert(var_local);

    i = 0;
    var_scope_foreach(scope, TRUE, test_var_scope_foreach_visitor1, &i);
    mu_assert_eq_int(2, i);
    i = 0;
    var_scope_foreach(scope, FALSE, test_var_scope_foreach_visitor1, &i);
    mu_assert_eq_int(1, i);
    i = 0;
    var_scope_foreach(scope, TRUE, test_var_scope_foreach_visitor2, &i);
    mu_assert_eq_int(0, i);
    var_scope_foreach(scope, FALSE, test_var_scope_foreach_visitor2, &i);
    mu_assert_eq_int(0, i);

    i = 0;
    var_scope_foreach(var_get_global_scope(), TRUE, test_var_scope_foreach_visitor1, &i);
    mu_assert_eq_int(1, i);
    i = 0;
    var_scope_foreach(var_get_global_scope(), FALSE, test_var_scope_foreach_visitor1, &i);
    mu_assert_eq_int(1, i);
    i = 0;
    var_scope_foreach(var_get_global_scope(), TRUE, test_var_scope_foreach_visitor2, &i);
    mu_assert_eq_int(0, i);
    var_scope_foreach(var_get_global_scope(), FALSE, test_var_scope_foreach_visitor2, &i);
    mu_assert_eq_int(0, i);

    var_scope_destroy(scope);
    var_fini_module();
    vartype_fini_module();
    package_fini_module();
mu_test_end()

void
test_var_scope(void)
{
    set_log_level(LOG_VAR);
    mu_run_test(test_var_scope_global);
    mu_run_test(test_var_scope_nested);
    mu_run_test(test_var_scope_redef);
    mu_run_test(test_var_scope_foreach);
}
